# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
# pylint: disable=simplifiable-if-statement
u"""
Interfaces for the NAG Mark 28.7 `omp` Chapter.

``omp`` - OpenMP Utilities

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module contains utilities for controlling the OpenMP environment
for your program.
They are based on OpenMP runtime library routines, although their
functionality varies slightly.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/x06/x06intro.html
"""

# NAG Copyright 2017-2022.

import logging as _logging
import sys as _sys

from ..base import utils

def set_num_threads(num):
    u"""
    Sets the number of threads for OpenMP parallel regions.

    ``set_num_threads`` sets the number of threads requested for
    subsequent OpenMP parallel regions.

    For full information please refer to the NAG Library document for
    x06aa

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/x06/x06aaf.html

    Parameters
    ----------
    num : int
        The number of threads requested for subsequent OpenMP parallel
        regions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 1.

    Notes
    -----
    ``set_num_threads``, for multithreaded implementations, sets the
    number of threads to be requested for subsequent parallel regions to
    `num`. The first element of the list held by the OpenMP Internal
    Control Variable (ICV) used in determining the number of threads is
    set.

    The number of threads used in parallel regions will be equal to, or
    less than, the value of the ICV.
    The actual number of threads used is dependent on several factors,
    such as the presence of a ``num_threads`` clause on the ``parallel``
    directive or the number of threads already in use by the program.
    Please refer to [References] for a full description of how the
    number of threads is chosen for a particular parallel region.

    In serial implementations of the NAG Library this function has no
    effect. See [the X06 Introduction] for a discussion of the behaviour
    of these functions when called in serial.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.set_num_threads'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, num.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _num,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: num >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def get_num_threads():
    u"""
    The number of OpenMP threads in the current team.

    ``get_num_threads`` returns the number of OpenMP threads in the
    current team.

    For full information please refer to the NAG Library document for
    x06ab

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/x06/x06abf.html

    Returns
    -------
    nt : int
        The returned number of OpenMP threads in the current team.

    Notes
    -----
    ``get_num_threads``, for multithreaded implementations, returns the
    number of OpenMP threads in the current team.
    If the number of threads is deemed critical then you are advised to
    use ``get_num_threads`` to retrieve this value as it may be less
    than that requested with either a call to ``set_num_threads``, your
    OMP_NUM_THREADS environment variable value or by using a
    ``num_threads`` clause on an OpenMP ``parallel`` directive.

    The number of threads actually in use in a parallel region is
    dependent on several factors.
    Please refer to [References] for a full description of how the
    number of threads is chosen for a particular parallel region.

    If this function is called from a sequential part of a multithreaded
    program then it will return the value 1.

    In serial implementations of the NAG Library this function will
    always return 1.
    See [the X06 Introduction] for a discussion of the behaviour of
    these functions when called in serial.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.get_num_threads'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, nt.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nt = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nt,
    )
    utils._handle_cb_exception(excs)
    _nt_py = utils._EngineIntScalarType.to_py(_nt)
    return_data = _nt_py
    return return_data

def get_max_threads():
    u"""
    An upper bound on the number of threads in the next parallel region.

    ``get_max_threads`` returns an upper bound on the number of threads
    used for subsequent OpenMP parallel regions.

    For full information please refer to the NAG Library document for
    x06ac

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/x06/x06acf.html

    Returns
    -------
    ntmax : int
        The returned upper bound on the number of threads used for
        subsequent OpenMP parallel regions.

    Notes
    -----
    ``get_max_threads``, for multithreaded implementations, returns the
    number of the threads to be requested for subsequent parallel
    regions. The value is the first element of the list held by the
    OpenMP Internal Control Variable (ICV) used in determining the
    number of threads.

    The number of threads used in parallel regions will be equal to, or
    less than, the first value of the ICV.
    The actual number of threads used is dependent on several factors,
    such as the presence of a ``num_threads`` clause on the ``parallel``
    directive or the number of threads already in use by the program.
    Please refer to [References] for a full description of how the
    number of threads is chosen for a particular parallel region.

    In serial implementations of the NAG Library this function will
    always return 1. See [the X06 Introduction] for a discussion of the
    behaviour of these functions when called in serial.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.get_max_threads'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, ntmax.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ntmax = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ntmax,
    )
    utils._handle_cb_exception(excs)
    _ntmax_py = utils._EngineIntScalarType.to_py(_ntmax)
    return_data = _ntmax_py
    return return_data

def get_thread_num():
    u"""
    The OpenMP thread number of the calling thread.

    ``get_thread_num`` returns the OpenMP thread number of the calling
    thread.

    For full information please refer to the NAG Library document for
    x06ad

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/x06/x06adf.html

    Returns
    -------
    tn : int
        The returned OpenMP thread number of the calling thread.

    Notes
    -----
    ``get_thread_num``, for multithreaded implementations, returns the
    calling OpenMP thread's unique thread number within the current
    team.
    The master thread will always return 0.
    The remaining threads will return a value between 1 and the value
    returned by ``get_num_threads`` less 1.

    If this function is called from a sequential part of a multithreaded
    program then it will return the value 0.

    In serial implementations of the NAG Library this function will
    always return 0.
    See [the X06 Introduction] for a discussion of the behaviour of
    these functions when called in serial.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.get_thread_num'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, tn.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _tn = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _tn,
    )
    utils._handle_cb_exception(excs)
    _tn_py = utils._EngineIntScalarType.to_py(_tn)
    return_data = _tn_py
    return return_data

def in_parallel():
    u"""
    Tests for an active OpenMP parallel region.

    ``in_parallel`` returns True if called from within an active OpenMP
    parallel region, and False otherwise.

    For full information please refer to the NAG Library document for
    x06af

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/x06/x06aff.html

    Returns
    -------
    inpar : bool
        True if called from within an active OpenMP parallel region, and
        False otherwise.

    Notes
    -----
    ``in_parallel``, for multithreaded implementations, returns True if
    called from within an active OpenMP parallel region, and False
    otherwise.

    In serial implementations of the NAG Library this function will
    always return False.
    See [the X06 Introduction] for a discussion of the behaviour of
    these functions when called in serial.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.in_parallel'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06afft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, inpar.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _inpar = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _inpar,
    )
    utils._handle_cb_exception(excs)
    _inpar_py = utils._EngineIntScalarType.to_py(_inpar)
    if _inpar_py == 0:
        _inpar_py = False
    elif _inpar_py == 1:
        _inpar_py = True
    else:
        raise utils.NagException(
            {'fun_name': fun_name},
            'Unexpected value for internal inpar.',
        )
    return_data = _inpar_py
    return return_data

@utils._withdrawal(
    'naginterfaces.base.omp.set_nested',
    mkwdrn='31.3',
    advice=[
        'Please use set_max_active_levels instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def set_nested(nesting):
    u"""
    Enables or disables nested OpenMP parallelism.

    ``set_nested`` enables or disables nested OpenMP parallelism.

    Note: ``set_nested`` will be removed in naginterfaces 31.3.0.0.
    Please use ``set_max_active_levels`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    x06ag

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/x06/x06agf.html

    Parameters
    ----------
    nesting : int
        If `nesting` = 0, nesting of OpenMP parallel regions is
        disabled, otherwise it is enabled.

    Notes
    -----
    ``set_nested``, for multithreaded implementations, enables or
    disables the nesting of OpenMP parallel regions by setting an OpenMP
    Internal Control Variable (ICV) and any vendor library specific
    options where that is possible.

    Nesting is disabled by default in OpenMP.

    In serial implementations of the NAG Library ``set_nested`` has no
    effect.
    See the submodule ``omp`` for a discussion of the behaviour of these
    functions when called in serial.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.set_nested'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06agft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nesting.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nesting = utils._EngineIntScalarType(
        nesting, {'fun_name': fun_name, 'entity': 'nesting'},
    ).to_ctypes()
    engine_ctypes(
        _en_data,
        _nesting,
    )
    utils._handle_cb_exception(excs)
    return_data = None
    return return_data

@utils._withdrawal(
    'naginterfaces.base.omp.get_nested',
    mkwdrn='31.3',
    advice=[
        'Please use get_max_active_levels instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def get_nested():
    u"""
    Tests the status of nested OpenMP parallelism.

    ``get_nested`` returns True if nested OpenMP parallelism has been
    enabled, and False otherwise.

    Note: ``get_nested`` will be removed in naginterfaces 31.3.0.0.
    Please use ``get_max_active_levels`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    x06ah

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/x06/x06ahf.html

    Returns
    -------
    nesting : bool
        True if nested OpenMP parallelism has been enabled, and False
        otherwise.

    Notes
    -----
    ``get_nested``, for multithreaded implementations, tests whether the
    nesting of OpenMP parallel regions has been enabled as if with a
    call to ``set_nested``.
    The return value will be True if nesting is enabled or False if it
    is disabled.

    In serial implementations of the NAG Library ``get_nested`` will
    always return False.
    See [the X06 Introduction] for a discussion of the behaviour of
    these functions when called in serial.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.get_nested'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06ahft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, nesting.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nesting = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nesting,
    )
    utils._handle_cb_exception(excs)
    _nesting_py = utils._EngineIntScalarType.to_py(_nesting)
    if _nesting_py == 0:
        _nesting_py = False
    elif _nesting_py == 1:
        _nesting_py = True
    else:
        raise utils.NagException(
            {'fun_name': fun_name},
            'Unexpected value for internal nesting.',
        )
    return_data = _nesting_py
    return return_data

def set_max_active_levels(num):
    u"""
    Limit the number of active nested parallel regions.

    ``set_max_active_levels`` sets the number of nested active parallel
    regions.

    For full information please refer to the NAG Library document for
    x06aj

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/x06/x06ajf.html

    Parameters
    ----------
    num : int
        n, the number of active parallel regions allowed.

        If `num` = 0

            parallelism is disabled.

        If `num` = 1

            the outermost parallel region is active and nested regions
            are inactive (i.e., no new threads are created when entering
            the nested parallel region and it is executed in serial.)

        If `num` >= 2

            `num` parallel regions can be active and `num`-1 nested
            regions can be active.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 0.

    Notes
    -----
    ``set_max_active_levels``, for multithreaded implementations, sets
    the maximum number of nested active parallel regions to `num`.

    When `num` is set to zero parallelism is disabled, if set to one
    then only the outermost parallel region is active and nested regions
    are inactive (i.e., no new threads are created when entering the
    nested parallel region and it is executed in serial.) If `num` is
    set to n, where n >= 2 then n parallel regions can be active and n-1
    nested regions can be active.

    In serial implementations of the NAG Library this function has no
    effect.
    See [the X06 Introduction] for a discussion of the behaviour of
    these functions when called in serial.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.set_max_active_levels'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06ajft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, num.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _num,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: num >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def get_max_active_levels():
    u"""
    Returns the maximum number of active nested parallel regions
    allowed.

    ``get_max_active_levels`` returns the maximum number of nested
    active parallel regions allowed.

    For full information please refer to the NAG Library document for
    x06ak

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/x06/x06akf.html

    Returns
    -------
    nl : int
        The returned maximum number of active nested parallel regions
        allowed.

    Notes
    -----
    ``get_max_active_levels``, for multithreaded implementations,
    retrieves the maximum number of nested active parallel regions
    allowed.
    This has either been set by the user or is the default for the
    implementation of OpenMP being used.

    In serial implementations of the NAG Library this function will
    always return 0.
    See [the X06 Introduction] for a discussion of the behaviour of
    these functions when called in serial.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.get_max_active_levels'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06akft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, nl.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nl = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nl,
    )
    utils._handle_cb_exception(excs)
    _nl_py = utils._EngineIntScalarType.to_py(_nl)
    return_data = _nl_py
    return return_data

def using_threaded_impl():
    u"""
    Tests whether a threaded NAG Library is being used.

    ``using_threaded_impl`` returns True when using a multithreaded
    implementation of the NAG Library and False otherwise.

    For full information please refer to the NAG Library document for
    x06xa

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/x06/x06xaf.html

    Returns
    -------
    smplib : bool
        True if using a multithreaded implementation of the NAG Library
        and False otherwise.

    Notes
    -----
    ``using_threaded_impl`` can be used to determine whether you are
    using a serial or multithreaded implementation of the NAG Library.
    In particular, with this information the behaviour of the other
    routines in this module can be predicted, as described in [the X06
    Introduction].

    The function returns True when a multithreaded implementation of the
    NAG Library is being used and False if a serial implementation is
    being used.
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.using_threaded_impl'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06xaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, smplib.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _smplib = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _smplib,
    )
    utils._handle_cb_exception(excs)
    _smplib_py = utils._EngineIntScalarType.to_py(_smplib)
    if _smplib_py == 0:
        _smplib_py = False
    elif _smplib_py == 1:
        _smplib_py = True
    else:
        raise utils.NagException(
            {'fun_name': fun_name},
            'Unexpected value for internal smplib.',
        )
    return_data = _smplib_py
    return return_data
