# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 28.7 `lapacklin` Chapter.

``lapacklin`` - Linear Equations (LAPACK)

This module provides functions for the solution of systems of
simultaneous linear equations, and associated computations.
It provides functions for

    matrix factorizations;

    solution of linear equations;

    estimating matrix condition numbers;

    computing error bounds for the solution of linear equations;

    matrix inversion;

    computing scaling factors to equilibrate a matrix.

Functions are provided for both `real` and `complex` data.

For a general introduction to the solution of systems of linear
equations, you should turn first to [the F04 Introduction].
The decision trees, in [the F04 Introduction], direct you to the most
appropriate functions in submodule ``linsys`` or submodule ``lapacklin``
for solving your particular problem.
In particular, submodule ``linsys`` and submodule ``lapacklin`` contain
`Black Box` (or `driver`) functions which enable some standard types of
problem to be solved by a call to a single function.
Where possible, functions in submodule ``linsys`` call submodule
``lapacklin`` functions to perform the necessary computational tasks.

There are two types of driver functions in this module: `simple drivers`
which just return the solution to the linear equations; and `expert
drivers` which also return condition and error estimates and, in many
cases, also allow equilibration.
The simple drivers for real matrices have names of the form and for
complex matrices have names of the form The expert drivers for real
matrices have names of the form and for complex matrices have names of
the form

The functions in this module (submodule ``lapacklin``) handle only
`dense` and `band` matrices (not matrices with more specialised
structures, or general sparse matrices).

The functions in this module have all been derived from the LAPACK
project (see Anderson `et al.` (1999)).
They have been designed to be efficient on a wide range of
high-performance computers, without compromising efficiency on
conventional serial machines.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07intro.html

See Also
--------
``naginterfaces.library.examples.lapacklin`` :
    This subpackage contains examples for the ``lapacklin`` module.
    See also the :ref:`library_lapacklin_ex` subsection.
"""

# NAG Copyright 2017-2022.

import collections as _collections

from ..base import lapacklin as _b_lapacklin
from ..base import utils as _b_utils
from ..library import _utils

DgesvReturnData = _collections.namedtuple(
    'DgesvReturnData',
    ['a', 'ipiv', 'b'],
)

def dgesv(a, b):
    u"""
    Computes the solution to a real system of linear equations.

    ``dgesv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n*n matrix and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07aa

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07aaf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n*n coefficient matrix A.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    Notes
    -----
    ``dgesv`` uses the LU decomposition with partial pivoting and row
    interchanges to factor A as

        A = PLU,

    where P is a permutation matrix, L is unit lower triangular, and U
    is upper triangular.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    See Also
    --------
    :meth:`naginterfaces.library.examples.lapacklin.dgesv_ex.main`
    """

    fun_name = 'naginterfaces.library.lapacklin.dgesv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dgesv(
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = DgesvReturnData(
        _a_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DgesvxReturnData = _collections.namedtuple(
    'DgesvxReturnData',
    ['a', 'af', 'ipiv', 'equed', 'r', 'c', 'b', 'x', 'rcond', 'ferr', 'berr', 'recip_growth_factor'],
)

def dgesvx(fact, trans, n, nrhs, a, equed, b, af=None, ipiv=None, r=None, c=None):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a real system of linear equations.

    ``dgesvx`` uses the LU factorization to compute the solution to a
    real system of linear equations

        AX = B or A^TX = B,

    where A is an n*n matrix and X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07ab

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07abf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `af` and `ipiv` contain the factorized form of A. If `equed`
            != 'N', the matrix A has been equilibrated with scaling
            factors given by `r` and `c`. `a`, `af` and `ipiv` are not
            modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `af` and factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T' or 'C'

            A^TX = B (Transpose).

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : float, array-like, shape (`n`, `n`)
        The n*n matrix A.

        If `fact` = 'F' and `equed` != 'N', `a` must have been
        equilibrated by the scaling factors in `r` and/or `c`.

    equed : None or str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'R', row equilibration, i.e., A has been
            premultiplied by D_R;

            if `equed` = 'C', column equilibration, i.e., A has been
            postmultiplied by D_C;

            if `equed` = 'B', both row and column equilibration, i.e., A
            has been replaced by D_R AD_C.

    b : float, array-like, shape (`n`, `nrhs`)
        The n*r right-hand side matrix B.

    af : None or float, array-like, shape (`n`, `n`), optional
        If `fact` = 'F', `af` contains the factors L and U from the
        factorization A = PLU as computed by ``dgetrf``. If `equed` !=
        'N', `af` is the factorized form of the equilibrated matrix A.

        If `fact` = 'N' or 'E', `af` need not be set.

    ipiv : None or int, array-like, shape (`n`), optional
        If `fact` = 'F', `ipiv` contains the pivot indices from the
        factorization A = PLU as computed by ``dgetrf``; at the ith step
        row i of the matrix was interchanged with row `ipiv`[i-1].
        `ipiv`[i-1] = i indicates a row interchange was not required.

        If `fact` = 'N' or 'E', `ipiv` need not be set.

    r : None or float, array-like, shape (`n`), optional
        If `fact` = 'N' or 'E', `r` need not be set.

        If `fact` = 'F' and `equed` = 'R' or 'B', `r` must contain the
        row scale factors for A, D_R; each element of `r` must be
        positive.

    c : None or float, array-like, shape (`n`), optional
        If `fact` = 'N' or 'E', `c` need not be set.

        If `fact` = 'F' and `equed` = 'C' or 'B', `c` must contain the
        column scale factors for A, D_C; each element of `c` must be
        positive.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `a` is not modified.

        If `fact` = 'E' or `equed` != 'N', A is scaled as follows:

            if `equed` = 'R', A = D_R A;

            if `equed` = 'C', A = AD_C;

            if `equed` = 'B', A = D_R AD_C.

    af : None or float, ndarray, shape (`n`, `n`)
        If `fact` = 'N', `af` returns the factors L and U from the
        factorization A = PLU of the original matrix A.

        If `fact` = 'E', `af` returns the factors L and U from the
        factorization A = PLU of the equilibrated matrix A (see the
        description of `a` for the form of the equilibrated matrix).

        If `fact` = 'F', `af` is unchanged from entry.

    ipiv : None or int, ndarray, shape (`n`)
        If `fact` = 'N', `ipiv` contains the pivot indices from the
        factorization A = PLU of the original matrix A.

        If `fact` = 'E', `ipiv` contains the pivot indices from the
        factorization A = PLU of the equilibrated matrix A.

        If `fact` = 'F', `ipiv` is unchanged from entry.

    equed : None or str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of equilibration that was performed as specified above.

    r : None or float, ndarray, shape (`n`)
        If `fact` = 'F', `r` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'R' or
        'B', `r` contains the row scale factors for A, D_R, such that A
        is multiplied on the left by D_R; each element of `r` is
        positive.

    c : None or float, ndarray, shape (`n`)
        If `fact` = 'F', `c` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'C' or
        'B', `c` contains the row scale factors for A, D_C; each element
        of `c` is positive.

    b : float, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `trans` = 'N' and `equed` = 'R' or 'B', `b` is overwritten by
        D_R B.

        If `trans` = 'T' or 'C' and `equed` = 'C' or 'B', `b` is
        overwritten by D_C B.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X to the original system of equations. Note that
        the arrays A and B are modified on exit if `equed` != 'N', and
        the solution to the equilibrated system is D_C^-1X if `trans` =
        'N' and `equed` = 'C' or 'B', or D_R^-1X if `trans` = 'T' or 'C'
        and `equed` = 'R' or 'B'.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    recip_growth_factor : float
        If no exception or warning is raised, the reciprocal pivot
        growth factor ||A||/||U||, where ||.|| denotes the maximum
        absolute element norm. If `recip_growth_factor`≪1, the stability
        of the LU factorization of (equilibrated) A could be poor. This
        also means that the solution `x`, condition estimate `rcond`,
        and forward error bound `ferr` could be unreliable. If the
        factorization fails with `errno` in 1 ... `n`, then
        `recip_growth_factor` contains the reciprocal pivot growth
        factor for the leading errno columns of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -10)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N', 'R', 'C' or 'B'.

        (`errno` -11)
            On entry, error in parameter `r`.

        (`errno` -12)
            On entry, error in parameter `c`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=`n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` `n`+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``dgesvx`` performs the following steps:

    (1) **Equilibration**

        The linear system to be solved may be badly scaled.
        However, the system can be equilibrated as a first stage by
        setting `fact` = 'E'.
        In this case, real scaling factors are computed and these
        factors then determine whether the system is to be equilibrated.
        Equilibrated forms of the systems AX = B and A^TX = B are

            (D_R AD_C)(D_C^-1X) = D_R B

        and

            (D_R AD_C)^T(D_R^-1X) = D_C B,

        respectively, where D_R and D_C are diagonal matrices, with
        positive diagonal elements, formed from the computed scaling
        factors.

        When equilibration is used, A will be overwritten by D_R AD_C
        and B will be overwritten by D_R B (or D_C B when the solution
        of A^TX = B is sought).

    (#) **Factorization**

        The matrix A, or its scaled form, is copied and factored using
        the LU decomposition

            A = PLU,

        where P is a permutation matrix, L is a unit lower triangular
        matrix, and U is upper triangular.

        This stage can be by-passed when a factored matrix (with scaled
        matrices and scaling factors) are supplied; for example, as
        provided by a previous call to ``dgesvx`` with the same matrix
        A.

    (#) **Condition Number Estimation**

        The LU factorization of A determines whether a solution to the
        linear system exists.
        If some diagonal element of U is zero, then U is exactly
        singular, no solution exists and the function returns with a
        failure.
        Otherwise the factorized form of A is used to estimate the
        condition number of the matrix A.
        If the reciprocal of the condition number is less than machine
        precision then a warning code is returned on final exit.

    (#) **Solution**

        The (equilibrated) system is solved for X (D_C^-1X or D_R^-1X)
        using the factored form of A (D_R AD_C).

    (#) **Iterative Refinement**

        Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for the computed solution.

    (#) **Construct Solution Matrix X**

        If equilibration was used, the matrix X is premultiplied by D_C
        (if `trans` = 'N') or D_R (if `trans` = 'T' or 'C') so that it
        solves the original system before equilibration.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dgesvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_af, 'af')),
    )
    _fact = _b_utils._EngineCharScalarType(
        fact, {'fun_name': fun_name, 'entity': 'fact'}, exp_chrlen=1,
    ).to_ctypes()
    _fact_py = _b_utils._EngineCharScalarType.to_py(_fact)
    if _fact_py in ('F', 'f'):
        if af is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'af'},
                'must not be None when fact in (\'F\', \'f\')',
            )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    if af is None:
        _af = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_n_py, _n_py),
            sorder=_m_sorder,
        )
    if _fact_py in ('F', 'f'):
        if ipiv is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'ipiv'},
                'must not be None when fact in (\'F\', \'f\')',
            )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
        allow_none=True,
    )
    _equed = _b_utils._EngineCharScalarType(
        equed, {'fun_name': fun_name, 'entity': 'equed'}, exp_chrlen=1,
        allow_none=True,
    ).to_ctypes()
    _equed_py = _b_utils._EngineCharScalarType.to_py(_equed)
    if (_fact_py in ('F', 'f')) and _equed_py in ('R', 'r', 'B', 'b'):
        if r is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'r'},
                'must not be None when (fact in (\'F\', \'f\')) and equed in (\'R\', \'r\', \'B\', \'b\')',
            )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    if (_fact_py in ('F', 'f')) and _equed_py in ('C', 'c', 'B', 'b'):
        if c is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'c'},
                'must not be None when (fact in (\'F\', \'f\')) and equed in (\'C\', \'c\', \'B\', \'b\')',
            )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dgesvx(
            fact,
            trans,
            _n,
            _nrhs,
            _a,
            _af,
            _ipiv,
            equed,
            _r,
            _c,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    if _fact_py not in ('F', 'f'):
        _equed_py = _return_scalars.equed
    else:
        _equed_py = None
    _b_py = _b
    if _fact_py not in ('F', 'f'):
        _af_py = _af
    else:
        _af_py = None
    if _fact_py not in ('F', 'f'):
        _ipiv_py = _ipiv
    else:
        _ipiv_py = None
    if _fact_py not in ('F', 'f') and _equed_py in ('R', 'r', 'B', 'b'):
        _r_py = _r
    else:
        _r_py = None
    if _fact_py not in ('F', 'f') and _equed_py in ('C', 'c', 'B', 'b'):
        _c_py = _c
    else:
        _c_py = None
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _recip_growth_factor_py = _return_scalars.recip_growth_factor
    _return_data = DgesvxReturnData(
        _a_py,
        _af_py,
        _ipiv_py,
        _equed_py,
        _r_py,
        _c_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
        _recip_growth_factor_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DsgesvReturnData = _collections.namedtuple(
    'DsgesvReturnData',
    ['a', 'ipiv', 'x', 'itera'],
)

def dsgesv(a, b):
    u"""
    Computes the solution to a real system of linear equations using
    mixed precision arithmetic.

    ``dsgesv`` computes the solution to a real system of linear
    equations

        AX = B,

    where A is an n*n matrix and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07ac

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07acf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n*n coefficient matrix A.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        If iterative refinement has been successfully used (i.e., if no
        exception or warning is raised and `itera` >= 0), then A is
        unchanged. If double precision factorization has been used (when
        no exception or warning is raised and `itera` < 0), A contains
        the factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required. `ipiv` corresponds either to
        the single precision factorization (if no exception or warning
        is raised and `itera` >= 0) or to the double precision
        factorization (if no exception or warning is raised and `itera`
        < 0).

    x : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    itera : int
        If `itera` > 0, iterative refinement has been successfully used
        and `itera` is the number of iterations carried out.

        If `itera` < 0, iterative refinement has failed for one of the
        reasons given below and double precision factorization has been
        carried out instead.

        `itera` = -1

            Taking into account machine parameters, and the values of n
            and nrhs, it is not worth working in single precision.

        `itera` = -2

            Overflow of an entry occurred when moving from double to
            single precision.

        `itera` = -3

            An intermediate single precision factorization failed.

        `itera` = -31

            The maximum permitted number of iterations was exceeded.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    Notes
    -----
    ``dsgesv`` first attempts to factorize the matrix in single
    precision and use this factorization within an iterative refinement
    procedure to produce a solution with full double precision accuracy.
    If the approach fails the method switches to a double precision
    factorization and solve.

    The iterative refinement process is stopped if

        `itera` > itermax,

    where `itera` is the number of iterations carried out thus far and
    itermax is the maximum number of iterations allowed, which is fixed
    at 30 iterations.
    The process is also stopped if for all right-hand sides we have

        ||resid|| < sqrt(n)||x||||A||epsilon,

    where ||resid|| is the infinity-norm of the residual, ||x|| is the
    infinity-norm of the solution, ||A|| is the infinity-operator-norm
    of the matrix A and epsilon is the machine precision returned by
    ``machine.precision``.

    The iterative refinement strategy used by ``dsgesv`` can be more
    efficient than the corresponding direct full precision algorithm.
    Since this strategy must perform iterative refinement on each
    right-hand side, any efficiency gains will reduce as the number of
    right-hand sides increases.
    Conversely, as the matrix size increases the cost of these iterative
    refinements become less significant relative to the cost of
    factorization.
    Thus, any efficiency gains will be greatest for a very small number
    of right-hand sides and for large matrix sizes.
    The cut-off values for the number of right-hand sides and matrix
    size, for which the iterative refinement strategy performs better,
    depends on the relative performance of the reduced and full
    precision factorization and back-substitution.
    For now, ``dsgesv`` always attempts the iterative refinement
    strategy first; you are advised to compare the performance of
    ``dsgesv`` with that of its full precision counterpart ``dgesv`` to
    determine whether this strategy is worthwhile for your particular
    problem dimensions.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Buttari, A, Dongarra, J, Langou, J, Langou, J, Luszczek, P and
    Kurzak, J, 2007, `Mixed precision iterative refinement techniques
    for the solution of dense linear systems`, International Journal of
    High Performance Computing Applications

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsgesv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalar = _b_lapacklin.dsgesv(
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
            _x,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _ipiv_py = _ipiv
    _x_py = _x
    _itera_py = _return_scalar
    _return_data = DsgesvReturnData(
        _a_py,
        _ipiv_py,
        _x_py,
        _itera_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DgetrfReturnData = _collections.namedtuple(
    'DgetrfReturnData',
    ['a', 'ipiv'],
)

def dgetrf(a):
    u"""
    LU factorization of real m*n matrix.

    ``dgetrf`` computes the LU factorization of a real m*n matrix.

    For full information please refer to the NAG Library document for
    f07ad

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07adf.html

    Parameters
    ----------
    a : float, array-like, shape (m, n)
        The m*n matrix A.

    Returns
    -------
    a : float, ndarray, shape (m, n)
        The factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (min(m, n))
        The pivot indices that define the permutation matrix. At the ith
        step, if `ipiv`[i-1] > i then row i of the matrix A was
        interchanged with row `ipiv`[i-1], for i = 1,2,...,min(m, n).
        `ipiv`[i-1] <= i indicates that, at the ith step, a row
        interchange was not required.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    Notes
    -----
    ``dgetrf`` forms the LU factorization of a real m*n matrix A as A =
    PLU, where P is a permutation matrix, L is lower triangular with
    unit diagonal elements (lower trapezoidal if m > n) and U is upper
    triangular (upper trapezoidal if m < n).
    Usually A is square (m = n), and both L and U are triangular.
    The function uses partial pivoting, with row interchanges.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgetrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (min(_m_py, _n_py),),
    )
    try:
        _b_lapacklin.dgetrf(
            _m,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _ipiv_py = _ipiv
    _return_data = DgetrfReturnData(
        _a_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgetrs(trans, a, ipiv, b):
    u"""
    Solution of real system of linear equations, multiple right-hand
    sides, matrix already factorized by ``dgetrf``.

    ``dgetrs`` solves a real system of linear equations with multiple
    right-hand sides,

        AX = B or A^TX = B,

    where A has been factorized by ``dgetrf``.

    For full information please refer to the NAG Library document for
    f07ae

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07aef.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            AX = B is solved for X.

        `trans` = 'T' or 'C'

            A^TX = B is solved for X.

    a : float, array-like, shape (n, n)
        The LU factorization of A, as returned by ``dgetrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``dgetrf``.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dgetrs`` is used to solve a real system of linear equations AX = B
    or A^TX = B, the function must be preceded by a call to ``dgetrf``
    which computes the LU factorization of A as A = PLU.
    The solution is computed by forward and backward substitution.

    If `trans` = 'N', the solution is computed by solving PLY = B and
    then UX = Y.

    If `trans` = 'T' or 'C', the solution is computed by solving U^TY =
    B and then L^TP^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgetrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.dgetrs(
        trans,
        _n,
        _nrhs,
        _a,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

DgeequReturnData = _collections.namedtuple(
    'DgeequReturnData',
    ['r', 'c', 'rowcnd', 'colcnd', 'amax'],
)

def dgeequ(a):
    u"""
    Computes row and column scalings intended to equilibrate a general
    real matrix and reduce its condition number.

    ``dgeequ`` computes diagonal scaling matrices D_R and D_C intended
    to equilibrate a real m*n matrix A and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07af

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07aff.html

    Parameters
    ----------
    a : float, array-like, shape (m, n)
        The matrix A whose scaling factors are to be computed.

    Returns
    -------
    r : float, ndarray, shape (m)
        If no exception or warning is raised or `errno` > `m`, `r`
        contains the row scale factors, the diagonal elements of D_R.
        The elements of `r` will be positive.

    c : float, ndarray, shape (n)
        If no exception or warning is raised, `c` contains the column
        scale factors, the diagonal elements of D_C. The elements of `c`
        will be positive.

    rowcnd : float
        If no exception or warning is raised or `errno` > `m`, `rowcnd`
        contains the ratio of the smallest value of `r`[i-1] to the
        largest value of `r`[i-1]. If `rowcnd` >= 0.1 and `amax` is
        neither too large nor too small, it is not worth scaling by D_R.

    colcnd : float
        If no exception or warning is raised, `colcnd` contains the
        ratio of the smallest value of `c`[i-1] to the largest value of
        `c`[i-1].

        If `colcnd` >= 0.1, it is not worth scaling by D_C.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=m))
            Row *<value>* of A is exactly zero.

        (`errno` i > m)
            Column *<value>* of A is exactly zero.

    Notes
    -----
    ``dgeequ`` computes the diagonal scaling matrices.
    The diagonal scaling matrices are chosen to try to make the elements
    of largest absolute value in each row and column of the matrix B
    given by

        B = D_R AD_C

    have absolute value 1.
    The diagonal elements of D_R and D_C are restricted to lie in the
    safe range (delta,1/delta), where delta is the value returned by
    function ``machine.real_safe``.
    Use of these scaling factors is not guaranteed to reduce the
    condition number of A but works well in practice.
    """

    fun_name = 'naginterfaces.library.lapacklin.dgeequ'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dgeequ(
            _m,
            _n,
            _a,
            _r,
            _c,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _r_py = _r
    _c_py = _c
    _rowcnd_py = _return_scalars.rowcnd
    _colcnd_py = _return_scalars.colcnd
    _amax_py = _return_scalars.amax
    _return_data = DgeequReturnData(
        _r_py,
        _c_py,
        _rowcnd_py,
        _colcnd_py,
        _amax_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgecon(norm, n, a, anorm):
    u"""
    Estimate condition number of real matrix, matrix already factorized
    by ``dgetrf``.

    ``dgecon`` estimates the condition number of a real matrix A, where
    A has been factorized by ``dgetrf``.

    For full information please refer to the NAG Library document for
    f07ag

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07agf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The LU factorization of A, as returned by ``dgetrf``.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.dlange`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``dgetrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``dgecon`` estimates the condition number of a real matrix A, in
    either the 1-norm or the infinity-norm:

        kappa_1(A) = ||A||_1||A^-1||_1 or kappa_infinity (A) =
        ||A||_infinity||A^-1||_infinity.

    Note that kappa_infinity (A) = kappa_1(A^T).

    Because the condition number is infinite if A is singular, the
    function actually returns an estimate of the **reciprocal** of the
    condition number.

    The function should be preceded by a call to ``blas.dlange`` to
    compute ||A||_1 or ||A||_infinity, and a call to ``dgetrf`` to
    compute the LU factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1 or ||A^-1||_infinity.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dgecon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dgecon(
        norm,
        n,
        _a,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DgerfsReturnData = _collections.namedtuple(
    'DgerfsReturnData',
    ['x', 'ferr', 'berr'],
)

def dgerfs(trans, n, a, af, ipiv, b, x):
    u"""
    Refined solution with error bounds of real system of linear
    equations, multiple right-hand sides.

    ``dgerfs`` returns error bounds for the solution of a real system of
    linear equations with multiple right-hand sides, AX = B or A^TX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07ah

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07ahf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the linear equations for which X is the
        computed solution.

        `trans` = 'N'

            The linear equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The linear equations are of the form A^TX = B.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n*n original matrix A as supplied to ``dgetrf``.

    af : float, array-like, shape (`n`, `n`)
        The LU factorization of A, as returned by ``dgetrf``.

    ipiv : int, array-like, shape (`n`)
        The pivot indices, as returned by ``dgetrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``dgetrs``.

    Returns
    -------
    x : float, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dgerfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a real system of linear equations
    with multiple right-hand sides AX = B or A^TX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``dgerfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgerfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dgerfs(
        trans,
        n,
        _nrhs,
        _a,
        _af,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DgerfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

DgetriReturnData = _collections.namedtuple(
    'DgetriReturnData',
    ['a'],
)

def dgetri(a, ipiv):
    u"""
    Inverse of real matrix, matrix already factorized by ``dgetrf``.

    ``dgetri`` computes the inverse of a real matrix A, where A has been
    factorized by ``dgetrf``.

    For full information please refer to the NAG Library document for
    f07aj

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07ajf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The LU factorization of A, as returned by ``dgetrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``dgetrf``.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The factorization is overwritten by the n*n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is zero. U is singular,
            and the inverse of A cannot be computed.

    Notes
    -----
    ``dgetri`` is used to compute the inverse of a real matrix A, the
    function must be preceded by a call to ``dgetrf``, which computes
    the LU factorization of A as A = PLU.
    The inverse of A is computed by forming U^-1 and then solving the
    equation XPL = U^-1 for X.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dgetri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.dgetri(
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = DgetriReturnData(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

ZgesvReturnData = _collections.namedtuple(
    'ZgesvReturnData',
    ['a', 'ipiv', 'b'],
)

def zgesv(a, b):
    u"""
    Computes the solution to a complex system of linear equations.

    ``zgesv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n matrix and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07an

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07anf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n*n coefficient matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    Notes
    -----
    ``zgesv`` uses the LU decomposition with partial pivoting and row
    interchanges to factor A as

        A = PLU,

    where P is a permutation matrix, L is unit lower triangular, and U
    is upper triangular.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgesv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zgesv(
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = ZgesvReturnData(
        _a_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZgesvxReturnData = _collections.namedtuple(
    'ZgesvxReturnData',
    ['a', 'af', 'ipiv', 'equed', 'r', 'c', 'b', 'x', 'rcond', 'ferr', 'berr', 'recip_growth_factor'],
)

def zgesvx(fact, trans, nrhs, a, af, ipiv, equed, r, c, b):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a complex system of linear equations.

    ``zgesvx`` uses the LU factorization to compute the solution to a
    complex system of linear equations

        AX = B or A^TX = B or A^HX = B,

    where A is an n*n matrix and X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07ap

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07apf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `af` and `ipiv` contain the factorized form of A. If `equed`
            != 'N', the matrix A has been equilibrated with scaling
            factors given by `r` and `c`. `a`, `af` and `ipiv` are not
            modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `af` and factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T'

            A^TX = B (Transpose).

        `trans` = 'C'

            A^HX = B (Conjugate transpose).

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : complex, array-like, shape (n, n)
        The n*n matrix A.

        If `fact` = 'F' and `equed` != 'N', `a` must have been
        equilibrated by the scaling factors in `r` and/or `c`.

    af : complex, array-like, shape (n, n)
        If `fact` = 'F', `af` contains the factors L and U from the
        factorization A = PLU as computed by ``zgetrf``. If `equed` !=
        'N', `af` is the factorized form of the equilibrated matrix A.

        If `fact` = 'N' or 'E', `af` need not be set.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains the pivot indices from the
        factorization A = PLU as computed by ``zgetrf``; at the ith step
        row i of the matrix was interchanged with row `ipiv`[i-1].
        `ipiv`[i-1] = i indicates a row interchange was not required.

        If `fact` = 'N' or 'E', `ipiv` need not be set.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'R', row equilibration, i.e., A has been
            premultiplied by D_R;

            if `equed` = 'C', column equilibration, i.e., A has been
            postmultiplied by D_C;

            if `equed` = 'B', both row and column equilibration, i.e., A
            has been replaced by D_R AD_C.

    r : float, array-like, shape (n)
        If `fact` = 'N' or 'E', `r` need not be set.

        If `fact` = 'F' and `equed` = 'R' or 'B', `r` must contain the
        row scale factors for A, D_R; each element of `r` must be
        positive.

    c : float, array-like, shape (n)
        If `fact` = 'N' or 'E', `c` need not be set.

        If `fact` = 'F' and `equed` = 'C' or 'B', `c` must contain the
        column scale factors for A, D_C; each element of `c` must be
        positive.

    b : complex, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `a` is not modified.

        If `fact` = 'E' or `equed` != 'N', A is scaled as follows:

            if `equed` = 'R', A = D_R A;

            if `equed` = 'C', A = AD_C;

            if `equed` = 'B', A = D_R AD_C.

    af : complex, ndarray, shape (n, n)
        If `fact` = 'N', `af` returns the factors L and U from the
        factorization A = PLU of the original matrix A.

        If `fact` = 'E', `af` returns the factors L and U from the
        factorization A = PLU of the equilibrated matrix A (see the
        description of `a` for the form of the equilibrated matrix).

        If `fact` = 'F', `af` is unchanged from entry.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains the pivot indices from the
        factorization A = PLU of the original matrix A.

        If `fact` = 'E', `ipiv` contains the pivot indices from the
        factorization A = PLU of the equilibrated matrix A.

        If `fact` = 'F', `ipiv` is unchanged from entry.

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of equilibration that was performed as specified above.

    r : float, ndarray, shape (n)
        If `fact` = 'F', `r` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'R' or
        'B', `r` contains the row scale factors for A, D_R, such that A
        is multiplied on the left by D_R; each element of `r` is
        positive.

    c : float, ndarray, shape (n)
        If `fact` = 'F', `c` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'C' or
        'B', `c` contains the row scale factors for A, D_C; each element
        of `c` is positive.

    b : complex, ndarray, shape (n, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `trans` = 'N' and `equed` = 'R' or 'B', `b` is overwritten by
        D_R B.

        If `trans` = 'T' or 'C' and `equed` = 'C' or 'B', `b` is
        overwritten by D_C B.

    x : complex, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X to the original system of equations. Note that
        the arrays A and B are modified on exit if `equed` != 'N', and
        the solution to the equilibrated system is D_C^-1X if `trans` =
        'N' and `equed` = 'C' or 'B', or D_R^-1X if `trans` = 'T' or 'C'
        and `equed` = 'R' or 'B'.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    recip_growth_factor : float
        If no exception or warning is raised, the reciprocal pivot
        growth factor ||A||/||U||, where ||.|| denotes the maximum
        absolute element norm. If `recip_growth_factor`≪1, the stability
        of the LU factorization of (equilibrated) A could be poor. This
        also means that the solution `x`, condition estimate `rcond`,
        and forward error bound `ferr` could be unreliable. If the
        factorization fails with `errno` in 1 ... `n`, then
        `recip_growth_factor` contains the reciprocal pivot growth
        factor for the leading errno columns of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -10)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N', 'R', 'C' or 'B'.

        (`errno` -11)
            On entry, error in parameter `r`.

        (`errno` -12)
            On entry, error in parameter `c`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``zgesvx`` performs the following steps:

    (1) **Equilibration**

        The linear system to be solved may be badly scaled.
        However, the system can be equilibrated as a first stage by
        setting `fact` = 'E'.
        In this case, real scaling factors are computed and these
        factors then determine whether the system is to be equilibrated.
        Equilibrated forms of the systems AX = B, A^TX = B and A^HX = B
        are

            (D_R AD_C)(D_C^-1X) = D_R B,

            (D_R AD_C)^T(D_R^-1X) = D_C B,

        and

            (D_R AD_C)^H(D_R^-1X) = D_C B,

        respectively, where D_R and D_C are diagonal matrices, with
        positive diagonal elements, formed from the computed scaling
        factors.

        When equilibration is used, A will be overwritten by D_R AD_C
        and B will be overwritten by D_R B (or D_C B when the solution
        of A^TX = B or A^HX = B is sought).

    (#) **Factorization**

        The matrix A, or its scaled form, is copied and factored using
        the LU decomposition

            A = PLU,

        where P is a permutation matrix, L is a unit lower triangular
        matrix, and U is upper triangular.

        This stage can be by-passed when a factored matrix (with scaled
        matrices and scaling factors) are supplied; for example, as
        provided by a previous call to ``zgesvx`` with the same matrix
        A.

    (#) **Condition Number Estimation**

        The LU factorization of A determines whether a solution to the
        linear system exists.
        If some diagonal element of U is zero, then U is exactly
        singular, no solution exists and the function returns with a
        failure.
        Otherwise the factorized form of A is used to estimate the
        condition number of the matrix A.
        If the reciprocal of the condition number is less than machine
        precision then a warning code is returned on final exit.

    (#) **Solution**

        The (equilibrated) system is solved for X (D_C^-1X or D_R^-1X)
        using the factored form of A (D_R AD_C).

    (#) **Iterative Refinement**

        Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for the computed solution.

    (#) **Construct Solution Matrix X**

        If equilibration was used, the matrix X is premultiplied by D_C
        (if `trans` = 'N') or D_R (if `trans` = 'T' or 'C') so that it
        solves the original system before equilibration.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zgesvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zgesvx(
            fact,
            trans,
            _n,
            _nrhs,
            _a,
            _af,
            _ipiv,
            equed,
            _r,
            _c,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _af_py = _af
    _ipiv_py = _ipiv
    _equed_py = _return_scalars.equed
    _r_py = _r
    _c_py = _c
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _recip_growth_factor_py = _return_scalars.recip_growth_factor
    _return_data = ZgesvxReturnData(
        _a_py,
        _af_py,
        _ipiv_py,
        _equed_py,
        _r_py,
        _c_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
        _recip_growth_factor_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZcgesvReturnData = _collections.namedtuple(
    'ZcgesvReturnData',
    ['a', 'ipiv', 'x', 'itera'],
)

def zcgesv(a, b):
    u"""
    Computes the solution to a complex system of linear equations using
    mixed precision arithmetic.

    ``zcgesv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n matrix and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07aq

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07aqf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n*n coefficient matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If iterative refinement has been successfully used (i.e., if no
        exception or warning is raised and `itera` >= 0), then A is
        unchanged. If double precision factorization has been used (when
        no exception or warning is raised and `itera` < 0), A contains
        the factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required. `ipiv` corresponds either to
        the single precision factorization (if no exception or warning
        is raised and `itera` >= 0) or to the double precision
        factorization (if no exception or warning is raised and `itera`
        < 0).

    x : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    itera : int
        If `itera` > 0, iterative refinement has been successfully used
        and `itera` is the number of iterations carried out.

        If `itera` < 0, iterative refinement has failed for one of the
        reasons given below and double precision factorization has been
        carried out instead.

        `itera` = -1

            Taking into account machine parameters, and the values of n
            and nrhs, it is not worth working in single precision.

        `itera` = -2

            Overflow of an entry occurred when moving from double to
            single precision.

        `itera` = -3

            An intermediate single precision factorization failed.

        `itera` = -31

            The maximum permitted number of iterations was exceeded.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    Notes
    -----
    ``zcgesv`` first attempts to factorize the matrix in single
    precision and use this factorization within an iterative refinement
    procedure to produce a solution with double precision accuracy.
    If the approach fails the method switches to a double precision
    factorization and solve.

    The iterative refinement process is stopped if

        `itera` > itermax,

    where `itera` is the number of iterations carried out thus far and
    itermax is the maximum number of iterations allowed, which is fixed
    at 30 iterations.
    The process is also stopped if for all right-hand sides we have

        ||resid|| < sqrt(n)||x||||A||epsilon,

    where ||resid|| is the infinity-norm of the residual, ||x|| is the
    infinity-norm of the solution, ||A|| is the infinity-operator-norm
    of the matrix A and epsilon is the machine precision returned by
    ``machine.precision``.

    The iterative refinement strategy used by ``zcgesv`` can be more
    efficient than the corresponding direct full precision algorithm.
    Since this strategy must perform iterative refinement on each
    right-hand side, any efficiency gains will reduce as the number of
    right-hand sides increases.
    Conversely, as the matrix size increases the cost of these iterative
    refinements become less significant relative to the cost of
    factorization.
    Thus, any efficiency gains will be greatest for a very small number
    of right-hand sides and for large matrix sizes.
    The cut-off values for the number of right-hand sides and matrix
    size, for which the iterative refinement strategy performs better,
    depends on the relative performance of the reduced and full
    precision factorization and back-substitution.
    For now, ``zcgesv`` always attempts the iterative refinement
    strategy first; you are advised to compare the performance of
    ``zcgesv`` with that of its full precision counterpart ``zgesv`` to
    determine whether this strategy is worthwhile for your particular
    problem dimensions.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Buttari, A, Dongarra, J, Langou, J, Langou, J, Luszczek, P and
    Kurzak, J, 2007, `Mixed precision iterative refinement techniques
    for the solution of dense linear systems`, International Journal of
    High Performance Computing Applications

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zcgesv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalar = _b_lapacklin.zcgesv(
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
            _x,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _ipiv_py = _ipiv
    _x_py = _x
    _itera_py = _return_scalar
    _return_data = ZcgesvReturnData(
        _a_py,
        _ipiv_py,
        _x_py,
        _itera_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZgetrfReturnData = _collections.namedtuple(
    'ZgetrfReturnData',
    ['a', 'ipiv'],
)

def zgetrf(a):
    u"""
    LU factorization of complex m*n matrix.

    ``zgetrf`` computes the LU factorization of a complex m*n matrix.

    For full information please refer to the NAG Library document for
    f07ar

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07arf.html

    Parameters
    ----------
    a : complex, array-like, shape (m, n)
        The m*n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (m, n)
        The factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (min(m, n))
        The pivot indices that define the permutation matrix. At the ith
        step, if `ipiv`[i-1] > i then row i of the matrix A was
        interchanged with row `ipiv`[i-1], for i = 1,2,...,min(m, n).
        `ipiv`[i-1] <= i indicates that, at the ith step, a row
        interchange was not required.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    Notes
    -----
    ``zgetrf`` forms the LU factorization of a complex m*n matrix A as A
    = PLU, where P is a permutation matrix, L is lower triangular with
    unit diagonal elements (lower trapezoidal if m > n) and U is upper
    triangular (upper trapezoidal if m < n).
    Usually A is square (m = n), and both L and U are triangular.
    The function uses partial pivoting, with row interchanges.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgetrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _m_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (min(_m_py, _n_py),),
    )
    try:
        _b_lapacklin.zgetrf(
            _m,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _ipiv_py = _ipiv
    _return_data = ZgetrfReturnData(
        _a_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgetrs(trans, a, ipiv, b):
    u"""
    Solution of complex system of linear equations, multiple right-hand
    sides, matrix already factorized by ``zgetrf``.

    ``zgetrs`` solves a complex system of linear equations with multiple
    right-hand sides,

        AX = B, A^TX = B or A^HX = B,

    where A has been factorized by ``zgetrf``.

    For full information please refer to the NAG Library document for
    f07as

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07asf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            AX = B is solved for X.

        `trans` = 'T'

            A^TX = B is solved for X.

        `trans` = 'C'

            A^HX = B is solved for X.

    a : complex, array-like, shape (n, n)
        The LU factorization of A, as returned by ``zgetrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``zgetrf``.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zgetrs`` is used to solve a complex system of linear equations AX
    = B, A^TX = B or A^HX = B, the function must be preceded by a call
    to ``zgetrf`` which computes the LU factorization of A as A = PLU.
    The solution is computed by forward and backward substitution.

    If `trans` = 'N', the solution is computed by solving PLY = B and
    then UX = Y.

    If `trans` = 'T', the solution is computed by solving U^TY = B and
    then L^TP^TX = Y.

    If `trans` = 'C', the solution is computed by solving U^HY = B and
    then L^HP^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgetrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zgetrs(
        trans,
        _n,
        _nrhs,
        _a,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

ZgeequReturnData = _collections.namedtuple(
    'ZgeequReturnData',
    ['r', 'c', 'rowcnd', 'colcnd', 'amax'],
)

def zgeequ(a):
    u"""
    Computes row and column scalings intended to equilibrate a general
    complex matrix and reduce its condition number.

    ``zgeequ`` computes real diagonal scaling matrices D_R and D_C
    intended to equilibrate a complex m*n matrix A and reduce its
    condition number.

    For full information please refer to the NAG Library document for
    f07at

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07atf.html

    Parameters
    ----------
    a : complex, array-like, shape (m, n)
        The matrix A whose scaling factors are to be computed.

    Returns
    -------
    r : float, ndarray, shape (m)
        If no exception or warning is raised or `errno` > `m`, `r`
        contains the row scale factors, the diagonal elements of D_R.
        The elements of `r` will be positive.

    c : float, ndarray, shape (n)
        If no exception or warning is raised, `c` contains the column
        scale factors, the diagonal elements of D_C. The elements of `c`
        will be positive.

    rowcnd : float
        If no exception or warning is raised or `errno` > `m`, `rowcnd`
        contains the ratio of the smallest value of `r`[i-1] to the
        largest value of `r`[i-1]. If `rowcnd` >= 0.1 and `amax` is
        neither too large nor too small, it is not worth scaling by D_R.

    colcnd : float
        If no exception or warning is raised, `colcnd` contains the
        ratio of the smallest value of `c`[i-1] to the largest value of
        `c`[i-1].

        If `colcnd` >= 0.1, it is not worth scaling by D_C.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=m))
            Row *<value>* of A is exactly zero.

        (`errno` i > m)
            Column *<value>* of A is exactly zero.

    Notes
    -----
    ``zgeequ`` computes the diagonal scaling matrices.
    The diagonal scaling matrices are chosen to try to make the elements
    of largest absolute value in each row and column of the matrix B
    given by

        B = D_R AD_C

    have absolute value 1.
    The diagonal elements of D_R and D_C are restricted to lie in the
    safe range (delta,1/delta), where delta is the value returned by
    function ``machine.real_safe``.
    Use of these scaling factors is not guaranteed to reduce the
    condition number of A but works well in practice.
    """

    fun_name = 'naginterfaces.library.lapacklin.zgeequ'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _m_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zgeequ(
            _m,
            _n,
            _a,
            _r,
            _c,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _r_py = _r
    _c_py = _c
    _rowcnd_py = _return_scalars.rowcnd
    _colcnd_py = _return_scalars.colcnd
    _amax_py = _return_scalars.amax
    _return_data = ZgeequReturnData(
        _r_py,
        _c_py,
        _rowcnd_py,
        _colcnd_py,
        _amax_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgecon(norm, a, anorm):
    u"""
    Estimate condition number of complex matrix, matrix already
    factorized by ``zgetrf``.

    ``zgecon`` estimates the condition number of a complex matrix A,
    where A has been factorized by ``zgetrf``.

    For full information please refer to the NAG Library document for
    f07au

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07auf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    a : complex, array-like, shape (n, n)
        The LU factorization of A, as returned by ``zgetrf``.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.zlange`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``zgetrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``zgecon`` estimates the condition number of a complex matrix A, in
    either the 1-norm or the infinity-norm:

        kappa_1(A) = ||A||_1||A^-1||_1 or kappa_infinity (A) =
        ||A||_infinity||A^-1||_infinity.

    Note that kappa_infinity (A) = kappa_1(A^H).

    Because the condition number is infinite if A is singular, the
    function actually returns an estimate of the **reciprocal** of the
    condition number.

    The function should be preceded by a call to ``blas.zlange`` to
    compute ||A||_1 or ||A||_infinity, and a call to ``zgetrf`` to
    compute the LU factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1 or ||A^-1||_infinity.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zgecon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_lapacklin.zgecon(
        norm,
        _n,
        _a,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZgerfsReturnData = _collections.namedtuple(
    'ZgerfsReturnData',
    ['x', 'ferr', 'berr'],
)

def zgerfs(trans, n, a, af, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex system of linear
    equations, multiple right-hand sides.

    ``zgerfs`` returns error bounds for the solution of a complex system
    of linear equations with multiple right-hand sides, AX = B, A^TX = B
    or A^HX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07av

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07avf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the linear equations for which X is the
        computed solution as follows:

        `trans` = 'N'

            The linear equations are of the form AX = B.

        `trans` = 'T'

            The linear equations are of the form A^TX = B.

        `trans` = 'C'

            The linear equations are of the form A^HX = B.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n*n original matrix A as supplied to ``zgetrf``.

    af : complex, array-like, shape (`n`, `n`)
        The LU factorization of A, as returned by ``zgetrf``.

    ipiv : int, array-like, shape (`n`)
        The pivot indices, as returned by ``zgetrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``zgetrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zgerfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex system of linear
    equations with multiple right-hand sides AX = B, A^TX = B or A^HX =
    B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``zgerfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgerfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zgerfs(
        trans,
        n,
        _nrhs,
        _a,
        _af,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZgerfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZgetriReturnData = _collections.namedtuple(
    'ZgetriReturnData',
    ['a'],
)

def zgetri(a, ipiv):
    u"""
    Inverse of complex matrix, matrix already factorized by ``zgetrf``.

    ``zgetri`` computes the inverse of a complex matrix A, where A has
    been factorized by ``zgetrf``.

    For full information please refer to the NAG Library document for
    f07aw

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07awf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The LU factorization of A, as returned by ``zgetrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``zgetrf``.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The factorization is overwritten by the n*n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is zero. U is singular,
            and the inverse of A cannot be computed.

    Notes
    -----
    ``zgetri`` is used to compute the inverse of a complex matrix A, the
    function must be preceded by a call to ``zgetrf``, which computes
    the LU factorization of A as A = PLU.
    The inverse of A is computed by forming U^-1 and then solving the
    equation XPL = U^-1 for X.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zgetri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.zgetri(
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = ZgetriReturnData(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

DgbsvReturnData = _collections.namedtuple(
    'DgbsvReturnData',
    ['ab', 'ipiv', 'b'],
)

def dgbsv(kl, ku, ab, b):
    u"""
    Computes the solution to a real banded system of linear equations.

    ``dgbsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n*n band matrix, with k_l subdiagonals and k_u
    superdiagonals, and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07ba

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07baf.html

    Parameters
    ----------
    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : float, array-like, shape (2*`kl`+`ku`+1, n)
        The n*n coefficient matrix A.

        See [Further Comments] for further details.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    ab : float, ndarray, shape (2*`kl`+`ku`+1, n)
        If `errno` >= 0, `ab` is overwritten by details of the
        factorization.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -3)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    Notes
    -----
    ``dgbsv`` uses the LU decomposition with partial pivoting and row
    interchanges to factor A as A = PLU, where P is a permutation
    matrix, L is a product of permutation and unit lower triangular
    matrices with k_l subdiagonals, and U is upper triangular with
    (k_l+k_u) superdiagonals.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dgbsv(
            _n,
            kl,
            ku,
            _nrhs,
            _ab,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ab_py = _ab
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = DgbsvReturnData(
        _ab_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DgbsvxReturnData = _collections.namedtuple(
    'DgbsvxReturnData',
    ['ab', 'afb', 'ipiv', 'equed', 'r', 'c', 'b', 'x', 'rcond', 'ferr', 'berr', 'recip_growth_factor'],
)

def dgbsvx(fact, trans, n, kl, ku, nrhs, ab, afb, ipiv, equed, r, c, b):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a real banded system of linear equations.

    ``dgbsvx`` uses the LU factorization to compute the solution to a
    real system of linear equations

        AX = B or A^TX = B,

    where A is an n*n band matrix with k_l subdiagonals and k_u
    superdiagonals, and X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07bb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07bbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afb` and `ipiv` contain the factorized form of A. If
            `equed` != 'N', the matrix A has been equilibrated with
            scaling factors given by `r` and `c`. `ab`, `afb` and `ipiv`
            are not modified.

        `fact` = 'N'

            The matrix A will be copied to `afb` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afb` and factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T' or 'C'

            A^TX = B (Transpose).

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ab : float, array-like, shape (`kl`+`ku`+1, `n`)
        The n*n coefficient matrix A.

        See [Further Comments] for further details.

        If `fact` = 'F' and `equed` != 'N', A must have been
        equilibrated by the scaling factors in `r` and/or `c`.

    afb : float, array-like, shape (2*`kl`+`ku`+1, `n`)
        If `fact` = 'N' or 'E', `afb` need not be set.

        If `fact` = 'F', details of the LU factorization of the n*n band
        matrix A, as computed by ``dgbtrf``.

        If `equed` != 'N', `afb` is the factorized form of the
        equilibrated matrix A.

    ipiv : int, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `ipiv` need not be set.

        If `fact` = 'F', `ipiv` contains the pivot indices from the
        factorization A = LU, as computed by ``dgbtrf``; row i of the
        matrix was interchanged with row `ipiv`[i-1].

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'R', row equilibration, i.e., A has been
            premultiplied by D_R;

            if `equed` = 'C', column equilibration, i.e., A has been
            postmultiplied by D_C;

            if `equed` = 'B', both row and column equilibration, i.e., A
            has been replaced by D_R AD_C.

    r : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `r` need not be set.

        If `fact` = 'F' and `equed` = 'R' or 'B', `r` must contain the
        row scale factors for A, D_R; each element of `r` must be
        positive.

    c : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `c` need not be set.

        If `fact` = 'F' and `equed` = 'C' or 'B', `c` must contain the
        column scale factors for A, D_C; each element of `c` must be
        positive.

    b : float, array-like, shape (`n`, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    ab : float, ndarray, shape (`kl`+`ku`+1, `n`)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `ab` is not modified.

        If `equed` != 'N' then, if no constraints are violated, A is
        scaled as follows:

            if `equed` = 'R', A = D_r A;

            if `equed` = 'C', A = AD_c;

            if `equed` = 'B', A = D_r AD_c.

    afb : float, ndarray, shape (2*`kl`+`ku`+1, `n`)
        If `fact` = 'F', `afb` is unchanged from entry.

        Otherwise, if no constraints are violated, then if `fact` = 'N',
        `afb` returns details of the LU factorization of the band matrix
        A, and if `fact` = 'E', `afb` returns details of the LU
        factorization of the equilibrated band matrix A (see the
        description of `ab` for the form of the equilibrated matrix).

    ipiv : int, ndarray, shape (`n`)
        If `fact` = 'F', `ipiv` is unchanged from entry.

        Otherwise, if no constraints are violated, `ipiv` contains the
        pivot indices that define the permutation matrix P; at the ith
        step row i of the matrix was interchanged with row `ipiv`[i-1].
        `ipiv`[i-1] = i indicates a row interchange was not required.

        If `fact` = 'N', the pivot indices are those corresponding to
        the factorization A = LU of the original matrix A.

        If `fact` = 'E', the pivot indices are those corresponding to
        the factorization of A = LU of the equilibrated matrix A.

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of equilibration that was performed as specified above.

    r : float, ndarray, shape (`n`)
        If `fact` = 'F', `r` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'R' or
        'B', `r` contains the row scale factors for A, D_R, such that A
        is multiplied on the left by D_R; each element of `r` is
        positive.

    c : float, ndarray, shape (`n`)
        If `fact` = 'F', `c` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'C' or
        'B', `c` contains the row scale factors for A, D_C; each element
        of `c` is positive.

    b : float, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `trans` = 'N' and `equed` = 'R' or 'B', `b` is overwritten by
        D_R B.

        If `trans` = 'T' or 'C' and `equed` = 'C' or 'B', `b` is
        overwritten by D_C B.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X to the original system of equations. Note that
        the arrays A and B are modified on exit if `equed` != 'N', and
        the solution to the equilibrated system is D_C^-1X if `trans` =
        'N' and `equed` = 'C' or 'B', or D_R^-1X if `trans` = 'T' or 'C'
        and `equed` = 'R' or 'B'.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    recip_growth_factor : float
        If no exception or warning is raised, the reciprocal pivot
        growth factor ||A||/||U||, where ||.|| denotes the maximum
        absolute element norm. If `recip_growth_factor`≪1, the stability
        of the LU factorization of (equilibrated) A could be poor. This
        also means that the solution `x`, condition estimate `rcond`,
        and forward error bound `ferr` could be unreliable. If the
        factorization fails with `errno` in 1 ... `n`, then
        `recip_growth_factor` contains the reciprocal pivot growth
        factor for the leading errno columns of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -5)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -6)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -12)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N', 'R', 'C' or 'B'.

        (`errno` -13)
            On entry, error in parameter `r`.

        (`errno` -14)
            On entry, error in parameter `c`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=`n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` `n`+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``dgbsvx`` performs the following steps:

    (1) **Equilibration**

        The linear system to be solved may be badly scaled.
        However, the system can be equilibrated as a first stage by
        setting `fact` = 'E'.
        In this case, real scaling factors are computed and these
        factors then determine whether the system is to be equilibrated.
        Equilibrated forms of the systems AX = B and A^TX = B are

            (D_R AD_C)(D_C^-1X) = D_R B

        and

            (D_R AD_C)^T(D_R^-1X) = D_C B,

        respectively, where D_R and D_C are diagonal matrices, with
        positive diagonal elements, formed from the computed scaling
        factors.

        When equilibration is used, A will be overwritten by D_R AD_C
        and B will be overwritten by D_R B (or D_C B when the solution
        of A^TX = B is sought).

    (#) **Factorization**

        The matrix A, or its scaled form, is copied and factored using
        the LU decomposition

            A = PLU,

        where P is a permutation matrix, L is a unit lower triangular
        matrix, and U is upper triangular.

        This stage can be by-passed when a factored matrix (with scaled
        matrices and scaling factors) are supplied; for example, as
        provided by a previous call to ``dgbsvx`` with the same matrix
        A.

    (#) **Condition Number Estimation**

        The LU factorization of A determines whether a solution to the
        linear system exists.
        If some diagonal element of U is zero, then U is exactly
        singular, no solution exists and the function returns with a
        failure.
        Otherwise the factorized form of A is used to estimate the
        condition number of the matrix A.
        If the reciprocal of the condition number is less than machine
        precision then a warning code is returned on final exit.

    (#) **Solution**

        The (equilibrated) system is solved for X (D_C^-1X or D_R^-1X)
        using the factored form of A (D_R AD_C).

    (#) **Iterative Refinement**

        Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for the computed solution.

    (#) **Construct Solution Matrix X**

        If equilibration was used, the matrix X is premultiplied by D_C
        (if `trans` = 'N') or D_R (if `trans` = 'T' or 'C') so that it
        solves the original system before equilibration.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbsvx'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_ab, 'ab'), (_afb, 'afb'), (_b, 'b')),
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dgbsvx(
            fact,
            trans,
            _n,
            kl,
            ku,
            _nrhs,
            _ab,
            _afb,
            _ipiv,
            equed,
            _r,
            _c,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ab_py = _ab
    _afb_py = _afb
    _ipiv_py = _ipiv
    _equed_py = _return_scalars.equed
    _r_py = _r
    _c_py = _c
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _recip_growth_factor_py = _return_scalars.recip_growth_factor
    _return_data = DgbsvxReturnData(
        _ab_py,
        _afb_py,
        _ipiv_py,
        _equed_py,
        _r_py,
        _c_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
        _recip_growth_factor_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DgbtrfReturnData = _collections.namedtuple(
    'DgbtrfReturnData',
    ['ab', 'ipiv'],
)

def dgbtrf(m, kl, ku, ab):
    u"""
    LU factorization of real m*n band matrix.

    ``dgbtrf`` computes the LU factorization of a real m*n band matrix.

    For full information please refer to the NAG Library document for
    f07bd

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07bdf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : float, array-like, shape (2*`kl`+`ku`+1, n)
        The m*n matrix A.

        See [Further Comments] for further details.

    Returns
    -------
    ab : float, ndarray, shape (2*`kl`+`ku`+1, n)
        If `errno` >= 0, `ab` is overwritten by details of the
        factorization.

    ipiv : int, ndarray, shape (min(`m`, n))
        The pivot indices that define the permutation matrix. At the ith
        step, if `ipiv`[i-1] > i then row i of the matrix A was
        interchanged with row `ipiv`[i-1], for i = 1,2,...,min(m, n).
        `ipiv`[i-1] <= i indicates that, at the ith step, a row
        interchange was not required.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    Notes
    -----
    ``dgbtrf`` forms the LU factorization of a real m*n band matrix A
    using partial pivoting, with row interchanges.
    Usually m = n, and then, if A has k_l nonzero subdiagonals and k_u
    nonzero superdiagonals, the factorization has the form A = PLU,
    where P is a permutation matrix, L is a lower triangular matrix with
    unit diagonal elements and at most k_l nonzero elements in each
    column, and U is an upper triangular band matrix with k_l+k_u
    superdiagonals.

    Note that L is not a band matrix, but the nonzero elements of L can
    be stored in the same space as the subdiagonal elements of A. U is a
    band matrix but with k_l additional superdiagonals compared with A.
    These additional superdiagonals are created by the row interchanges.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbtrf'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (min(_m_py, _n_py),),
    )
    try:
        _b_lapacklin.dgbtrf(
            _m,
            _n,
            kl,
            ku,
            _ab,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ab_py = _ab
    _ipiv_py = _ipiv
    _return_data = DgbtrfReturnData(
        _ab_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgbtrs(trans, kl, ku, ab, ipiv, b):
    u"""
    Solution of real band system of linear equations, multiple
    right-hand sides, matrix already factorized by ``dgbtrf``.

    ``dgbtrs`` solves a real band system of linear equations with
    multiple right-hand sides,

        AX = B or A^TX = B,

    where A has been factorized by ``dgbtrf``.

    For full information please refer to the NAG Library document for
    f07be

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07bef.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            AX = B is solved for X.

        `trans` = 'T' or 'C'

            A^TX = B is solved for X.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : float, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``dgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``dgbtrf``.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dgbtrs`` is used to solve a real band system of linear equations
    AX = B or A^TX = B, the function must be preceded by a call to
    ``dgbtrf`` which computes the LU factorization of A as A = PLU.
    The solution is computed by forward and backward substitution.

    If `trans` = 'N', the solution is computed by solving PLY = B and
    then UX = Y.

    If `trans` = 'T' or 'C', the solution is computed by solving U^TY =
    B and then L^TP^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.dgbtrs(
        trans,
        _n,
        kl,
        ku,
        _nrhs,
        _ab,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

DgbequReturnData = _collections.namedtuple(
    'DgbequReturnData',
    ['r', 'c', 'rowcnd', 'colcnd', 'amax'],
)

def dgbequ(m, kl, ku, ab):
    u"""
    Computes row and column scalings intended to equilibrate a real
    banded matrix and reduce its condition number.

    ``dgbequ`` computes diagonal scaling matrices D_R and D_C intended
    to equilibrate a real m*n band matrix A of band width (k_l+k_u+1),
    and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07bf

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07bff.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    kl : int
        k_l, the number of subdiagonals of the matrix A.

    ku : int
        k_u, the number of superdiagonals of the matrix A.

    ab : float, array-like, shape (`kl`+`ku`+1, n)
        The m*n band matrix A whose scaling factors are to be computed.

        See [Further Comments] for further details.

    Returns
    -------
    r : float, ndarray, shape (`m`)
        If no exception or warning is raised or `errno` > `m`, `r`
        contains the row scale factors, the diagonal elements of D_R.
        The elements of `r` will be positive.

    c : float, ndarray, shape (n)
        If no exception or warning is raised, `c` contains the column
        scale factors, the diagonal elements of D_C. The elements of `c`
        will be positive.

    rowcnd : float
        If no exception or warning is raised or `errno` > `m`, `rowcnd`
        contains the ratio of the smallest value of `r`[i-1] to the
        largest value of `r`[i-1]. If `rowcnd` >= 0.1 and `amax` is
        neither too large nor too small, it is not worth scaling by D_R.

    colcnd : float
        If no exception or warning is raised, `colcnd` contains the
        ratio of the smallest value of `c`[i-1] to the largest value of
        `c`[i-1].

        If `colcnd` >= 0.1, it is not worth scaling by D_C.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=`m`))
            Row *<value>* of A is exactly zero.

        (`errno` i > `m`)
            Column *<value>* of A is exactly zero.

    Notes
    -----
    ``dgbequ`` computes the diagonal scaling matrices.
    The diagonal scaling matrices are chosen to try to make the elements
    of largest absolute value in each row and column of the matrix B
    given by

        B = D_R AD_C

    have absolute value 1.
    The diagonal elements of D_R and D_C are restricted to lie in the
    safe range (delta,1/delta), where delta is the value returned by
    function ``machine.real_safe``.
    Use of these scaling factors is not guaranteed to reduce the
    condition number of A but works well in practice.
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbequ'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dgbequ(
            _m,
            _n,
            kl,
            ku,
            _ab,
            _r,
            _c,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _r_py = _r
    _c_py = _c
    _rowcnd_py = _return_scalars.rowcnd
    _colcnd_py = _return_scalars.colcnd
    _amax_py = _return_scalars.amax
    _return_data = DgbequReturnData(
        _r_py,
        _c_py,
        _rowcnd_py,
        _colcnd_py,
        _amax_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgbcon(norm, kl, ku, ab, ipiv, anorm):
    u"""
    Estimate condition number of real band matrix, matrix already
    factorized by ``dgbtrf``.

    ``dgbcon`` estimates the condition number of a real band matrix A,
    where A has been factorized by ``dgbtrf``.

    For full information please refer to the NAG Library document for
    f07bg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07bgf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : float, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``dgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``dgbtrf``.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.dlangb`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``dgbtrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -8)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``dgbcon`` estimates the condition number of a real band matrix A,
    in either the 1-norm or the infinity-norm:

        kappa_1(A) = ||A||_1||A^-1||_1 or kappa_infinity (A) =
        ||A||_infinity||A^-1||_infinity.

    Note that kappa_infinity (A) = kappa_1(A^T).

    Because the condition number is infinite if A is singular, the
    function actually returns an estimate of the **reciprocal** of the
    condition number.

    The function should be preceded by a call to ``blas.dlangb`` to
    compute ||A||_1 or ||A||_infinity, and a call to ``dgbtrf`` to
    compute the LU factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1 or ||A^-1||_infinity.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.dgbcon(
        norm,
        _n,
        kl,
        ku,
        _ab,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DgbrfsReturnData = _collections.namedtuple(
    'DgbrfsReturnData',
    ['x', 'ferr', 'berr'],
)

def dgbrfs(trans, kl, ku, nrhs, ab, afb, ipiv, b, x):
    u"""
    Refined solution with error bounds of real band system of linear
    equations, multiple right-hand sides.

    ``dgbrfs`` returns error bounds for the solution of a real band
    system of linear equations with multiple right-hand sides, AX = B or
    A^TX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07bh

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07bhf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the linear equations for which X is the
        computed solution.

        `trans` = 'N'

            The linear equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The linear equations are of the form A^TX = B.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides.

    ab : float, array-like, shape (`kl`+`ku`+1, n)
        The original n*n band matrix A as supplied to ``dgbtrf`` but
        with reduced requirements since the matrix is not factorized.

        See [Further Comments] for further details.

    afb : float, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``dgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``dgbtrf``.

    b : float, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    x : float, array-like, shape (n, `nrhs`)
        The n*r solution matrix X, as returned by ``dgbtrs``.

    Returns
    -------
    x : float, ndarray, shape (n, `nrhs`)
        The improved solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -5)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Notes
    -----
    ``dgbrfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a real band system of linear
    equations with multiple right-hand sides AX = B or A^TX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``dgbrfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dgbrfs(
        trans,
        _n,
        kl,
        ku,
        _nrhs,
        _ab,
        _afb,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DgbrfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZgbsvReturnData = _collections.namedtuple(
    'ZgbsvReturnData',
    ['ab', 'ipiv', 'b'],
)

def zgbsv(kl, ku, ab, b):
    u"""
    Computes the solution to a complex banded system of linear
    equations.

    ``zgbsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n band matrix, with k_l subdiagonals and k_u
    superdiagonals, and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07bn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07bnf.html

    Parameters
    ----------
    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : complex, array-like, shape (2*`kl`+`ku`+1, n)
        The n*n coefficient matrix A.

        See [Further Comments] for further details.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    ab : complex, ndarray, shape (2*`kl`+`ku`+1, n)
        If `errno` >= 0, `ab` is overwritten by details of the
        factorization.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -3)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    Notes
    -----
    ``zgbsv`` uses the LU decomposition with partial pivoting and row
    interchanges to factor A as A = PLU, where P is a permutation
    matrix, L is a product of permutation and unit lower triangular
    matrices with k_l subdiagonals, and U is upper triangular with
    (k_l+k_u) superdiagonals.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zgbsv(
            _n,
            kl,
            ku,
            _nrhs,
            _ab,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ab_py = _ab
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = ZgbsvReturnData(
        _ab_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZgbsvxReturnData = _collections.namedtuple(
    'ZgbsvxReturnData',
    ['ab', 'afb', 'ipiv', 'equed', 'r', 'c', 'b', 'x', 'rcond', 'ferr', 'berr', 'recip_growth_factor'],
)

def zgbsvx(fact, trans, kl, ku, nrhs, ab, afb, ipiv, equed, r, c, b):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a complex banded system of linear equations.

    ``zgbsvx`` uses the LU factorization to compute the solution to a
    complex system of linear equations

        AX = B, A^TX = B or A^HX = B,

    where A is an n*n band matrix with k_l subdiagonals and k_u
    superdiagonals, and X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07bp

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07bpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afb` and `ipiv` contain the factorized form of A. If
            `equed` != 'N', the matrix A has been equilibrated with
            scaling factors given by `r` and `c`. `ab`, `afb` and `ipiv`
            are not modified.

        `fact` = 'N'

            The matrix A will be copied to `afb` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afb` and factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T'

            A^TX = B (Transpose).

        `trans` = 'C'

            A^HX = B (Conjugate transpose).

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ab : complex, array-like, shape (`kl`+`ku`+1, n)
        The n*n coefficient matrix A.

        See [Further Comments] for further details.

        If `fact` = 'F' and `equed` != 'N', A must have been
        equilibrated by the scaling factors in `r` and/or `c`.

    afb : complex, array-like, shape (2*`kl`+`ku`+1, n)
        If `fact` = 'N' or 'E', `afb` need not be set.

        If `fact` = 'F', details of the LU factorization of the n*n band
        matrix A, as computed by ``zgbtrf``.

        If `equed` != 'N', `afb` is the factorized form of the
        equilibrated matrix A.

    ipiv : int, array-like, shape (n)
        If `fact` = 'N' or 'E', `ipiv` need not be set.

        If `fact` = 'F', `ipiv` contains the pivot indices from the
        factorization A = LU, as computed by ``dgbtrf``; row i of the
        matrix was interchanged with row `ipiv`[i-1].

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'R', row equilibration, i.e., A has been
            premultiplied by D_R;

            if `equed` = 'C', column equilibration, i.e., A has been
            postmultiplied by D_C;

            if `equed` = 'B', both row and column equilibration, i.e., A
            has been replaced by D_R AD_C.

    r : float, array-like, shape (n)
        If `fact` = 'N' or 'E', `r` need not be set.

        If `fact` = 'F' and `equed` = 'R' or 'B', `r` must contain the
        row scale factors for A, D_R; each element of `r` must be
        positive.

    c : float, array-like, shape (n)
        If `fact` = 'N' or 'E', `c` need not be set.

        If `fact` = 'F' and `equed` = 'C' or 'B', `c` must contain the
        column scale factors for A, D_C; each element of `c` must be
        positive.

    b : complex, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    ab : complex, ndarray, shape (`kl`+`ku`+1, n)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `ab` is not modified.

        If `equed` != 'N' then, if no constraints are violated, A is
        scaled as follows:

            if `equed` = 'R', A = D_r A;

            if `equed` = 'C', A = AD_c;

            if `equed` = 'B', A = D_r AD_c.

    afb : complex, ndarray, shape (2*`kl`+`ku`+1, n)
        If `fact` = 'F', `afb` is unchanged from entry.

        Otherwise, if no constraints are violated, then if `fact` = 'N',
        `afb` returns details of the LU factorization of the band matrix
        A, and if `fact` = 'E', `afb` returns details of the LU
        factorization of the equilibrated band matrix A (see the
        description of `ab` for the form of the equilibrated matrix).

    ipiv : int, ndarray, shape (n)
        If `fact` = 'F', `ipiv` is unchanged from entry.

        Otherwise, if no constraints are violated, `ipiv` contains the
        pivot indices that define the permutation matrix P; at the ith
        step row i of the matrix was interchanged with row `ipiv`[i-1].
        `ipiv`[i-1] = i indicates a row interchange was not required.

        If `fact` = 'N', the pivot indices are those corresponding to
        the factorization A = LU of the original matrix A.

        If `fact` = 'E', the pivot indices are those corresponding to
        the factorization of A = LU of the equilibrated matrix A.

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of equilibration that was performed as specified above.

    r : float, ndarray, shape (n)
        If `fact` = 'F', `r` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'R' or
        'B', `r` contains the row scale factors for A, D_R, such that A
        is multiplied on the left by D_R; each element of `r` is
        positive.

    c : float, ndarray, shape (n)
        If `fact` = 'F', `c` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'C' or
        'B', `c` contains the row scale factors for A, D_C; each element
        of `c` is positive.

    b : complex, ndarray, shape (n, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `trans` = 'N' and `equed` = 'R' or 'B', `b` is overwritten by
        D_R B.

        If `trans` = 'T' or 'C' and `equed` = 'C' or 'B', `b` is
        overwritten by D_C B.

    x : complex, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X to the original system of equations. Note that
        the arrays A and B are modified on exit if `equed` != 'N', and
        the solution to the equilibrated system is D_C^-1X if `trans` =
        'N' and `equed` = 'C' or 'B', or D_R^-1X if `trans` = 'T' or 'C'
        and `equed` = 'R' or 'B'.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    recip_growth_factor : float
        If no exception or warning is raised, the reciprocal pivot
        growth factor ||A||/||U||, where ||.|| denotes the maximum
        absolute element norm. If `recip_growth_factor`≪1, the stability
        of the LU factorization of (equilibrated) A could be poor. This
        also means that the solution `x`, condition estimate `rcond`,
        and forward error bound `ferr` could be unreliable. If the
        factorization fails with `errno` in 1 ... `n`, then
        `recip_growth_factor` contains the reciprocal pivot growth
        factor for the leading errno columns of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -5)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -6)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -12)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N', 'R', 'C' or 'B'.

        (`errno` -13)
            On entry, error in parameter `r`.

        (`errno` -14)
            On entry, error in parameter `c`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``zgbsvx`` performs the following steps:

    (1) **Equilibration**

        The linear system to be solved may be badly scaled.
        However, the system can be equilibrated as a first stage by
        setting `fact` = 'E'.
        In this case, real scaling factors are computed and these
        factors then determine whether the system is to be equilibrated.
        Equilibrated forms of the systems AX = B, A^TX = B and A^HX = B
        are

            (D_R AD_C)(D_C^-1X) = D_R B,

            (D_R AD_C)^T(D_R^-1X) = D_C B,

        and

            (D_R AD_C)^H(D_R^-1X) = D_C B,

        respectively, where D_R and D_C are diagonal matrices, with
        positive diagonal elements, formed from the computed scaling
        factors.

        When equilibration is used, A will be overwritten by D_R AD_C
        and B will be overwritten by D_R B (or D_C B when the solution
        of A^TX = B or A^HX = B is sought).

    (#) **Factorization**

        The matrix A, or its scaled form, is copied and factored using
        the LU decomposition

            A = PLU,

        where P is a permutation matrix, L is a unit lower triangular
        matrix, and U is upper triangular.

        This stage can be by-passed when a factored matrix (with scaled
        matrices and scaling factors) are supplied; for example, as
        provided by a previous call to ``zgbsvx`` with the same matrix
        A.

    (#) **Condition Number Estimation**

        The LU factorization of A determines whether a solution to the
        linear system exists.
        If some diagonal element of U is zero, then U is exactly
        singular, no solution exists and the function returns with a
        failure.
        Otherwise the factorized form of A is used to estimate the
        condition number of the matrix A.
        If the reciprocal of the condition number is less than machine
        precision then a warning code is returned on final exit.

    (#) **Solution**

        The (equilibrated) system is solved for X (D_C^-1X or D_R^-1X)
        using the factored form of A (D_R AD_C).

    (#) **Iterative Refinement**

        Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for the computed solution.

    (#) **Construct Solution Matrix X**

        If equilibration was used, the matrix X is premultiplied by D_C
        (if `trans` = 'N') or D_R (if `trans` = 'T' or 'C') so that it
        solves the original system before equilibration.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbsvx'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_ab, 'ab'), (_afb, 'afb'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zgbsvx(
            fact,
            trans,
            _n,
            kl,
            ku,
            _nrhs,
            _ab,
            _afb,
            _ipiv,
            equed,
            _r,
            _c,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ab_py = _ab
    _afb_py = _afb
    _ipiv_py = _ipiv
    _equed_py = _return_scalars.equed
    _r_py = _r
    _c_py = _c
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _recip_growth_factor_py = _return_scalars.recip_growth_factor
    _return_data = ZgbsvxReturnData(
        _ab_py,
        _afb_py,
        _ipiv_py,
        _equed_py,
        _r_py,
        _c_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
        _recip_growth_factor_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZgbtrfReturnData = _collections.namedtuple(
    'ZgbtrfReturnData',
    ['ab', 'ipiv'],
)

def zgbtrf(m, kl, ku, ab):
    u"""
    LU factorization of complex m*n band matrix.

    ``zgbtrf`` computes the LU factorization of a complex m*n band
    matrix.

    For full information please refer to the NAG Library document for
    f07br

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07brf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : complex, array-like, shape (2*`kl`+`ku`+1, n)
        The m*n matrix A.

        See [Further Comments] for further details.

    Returns
    -------
    ab : complex, ndarray, shape (2*`kl`+`ku`+1, n)
        If `errno` >= 0, `ab` is overwritten by details of the
        factorization.

    ipiv : int, ndarray, shape (min(`m`, n))
        The pivot indices that define the permutation matrix. At the ith
        step, if `ipiv`[i-1] > i then row i of the matrix A was
        interchanged with row `ipiv`[i-1], for i = 1,2,...,min(m, n).
        `ipiv`[i-1] <= i indicates that, at the ith step, a row
        interchange was not required.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    Notes
    -----
    ``zgbtrf`` forms the LU factorization of a complex m*n band matrix A
    using partial pivoting, with row interchanges.
    Usually m = n, and then, if A has k_l nonzero subdiagonals and k_u
    nonzero superdiagonals, the factorization has the form A = PLU,
    where P is a permutation matrix, L is a lower triangular matrix with
    unit diagonal elements and at most k_l nonzero elements in each
    column, and U is an upper triangular band matrix with k_l+k_u
    superdiagonals.

    Note that L is not a band matrix, but the nonzero elements of L can
    be stored in the same space as the subdiagonal elements of A. U is a
    band matrix but with k_l additional superdiagonals compared with A.
    These additional superdiagonals are created by the row interchanges.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbtrf'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (min(_m_py, _n_py),),
    )
    try:
        _b_lapacklin.zgbtrf(
            _m,
            _n,
            kl,
            ku,
            _ab,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ab_py = _ab
    _ipiv_py = _ipiv
    _return_data = ZgbtrfReturnData(
        _ab_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgbtrs(trans, kl, ku, ab, ipiv, b):
    u"""
    Solution of complex band system of linear equations, multiple
    right-hand sides, matrix already factorized by ``zgbtrf``.

    ``zgbtrs`` solves a complex band system of linear equations with
    multiple right-hand sides,

        AX = B, A^TX = B or A^HX = B,

    where A has been factorized by ``zgbtrf``.

    For full information please refer to the NAG Library document for
    f07bs

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07bsf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            AX = B is solved for X.

        `trans` = 'T'

            A^TX = B is solved for X.

        `trans` = 'C'

            A^HX = B is solved for X.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : complex, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``zgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``zgbtrf``.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zgbtrs`` is used to solve a complex band system of linear
    equations AX = B, A^TX = B or A^HX = B, the function must be
    preceded by a call to ``zgbtrf`` which computes the LU factorization
    of A as A = PLU.
    The solution is computed by forward and backward substitution.

    If `trans` = 'N', the solution is computed by solving PLY = B and
    then UX = Y.

    If `trans` = 'T', the solution is computed by solving U^TY = B and
    then L^TP^TX = Y.

    If `trans` = 'C', the solution is computed by solving U^HY = B and
    then L^HP^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zgbtrs(
        trans,
        _n,
        kl,
        ku,
        _nrhs,
        _ab,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

ZgbequReturnData = _collections.namedtuple(
    'ZgbequReturnData',
    ['r', 'c', 'rowcnd', 'colcnd', 'amax'],
)

def zgbequ(m, kl, ku, ab):
    u"""
    Computes row and column scalings intended to equilibrate a complex
    banded matrix and reduce its condition number.

    ``zgbequ`` computes diagonal scaling matrices D_R and D_C intended
    to equilibrate a complex m*n band matrix A of band width
    (k_l+k_u+1), and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07bt

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07btf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    kl : int
        k_l, the number of subdiagonals of the matrix A.

    ku : int
        k_u, the number of superdiagonals of the matrix A.

    ab : complex, array-like, shape (`kl`+`ku`+1, n)
        The m*n band matrix A whose scaling factors are to be computed.

        See [Further Comments] for further details.

    Returns
    -------
    r : float, ndarray, shape (`m`)
        If no exception or warning is raised or `errno` > `m`, `r`
        contains the row scale factors, the diagonal elements of D_R.
        The elements of `r` will be positive.

    c : float, ndarray, shape (n)
        If no exception or warning is raised, `c` contains the column
        scale factors, the diagonal elements of D_C. The elements of `c`
        will be positive.

    rowcnd : float
        If no exception or warning is raised or `errno` > `m`, `rowcnd`
        contains the ratio of the smallest value of `r`[i-1] to the
        largest value of `r`[i-1]. If `rowcnd` >= 0.1 and `amax` is
        neither too large nor too small, it is not worth scaling by D_R.

    colcnd : float
        If no exception or warning is raised, `colcnd` contains the
        ratio of the smallest value of `c`[i-1] to the largest value of
        `c`[i-1].

        If `colcnd` >= 0.1, it is not worth scaling by D_C.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=`m`))
            Row *<value>* of A is exactly zero.

        (`errno` i > `m`)
            Column *<value>* of A is exactly zero.

    Notes
    -----
    ``zgbequ`` computes the diagonal scaling matrices.
    The diagonal scaling matrices are chosen to try to make the elements
    of largest absolute value in each row and column of the matrix B
    given by

        B = D_R AD_C

    have absolute value 1.
    The diagonal elements of D_R and D_C are restricted to lie in the
    safe range (delta,1/delta), where delta is the value returned by
    function ``machine.real_safe``.
    Use of these scaling factors is not guaranteed to reduce the
    condition number of A but works well in practice.
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbequ'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zgbequ(
            _m,
            _n,
            kl,
            ku,
            _ab,
            _r,
            _c,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _r_py = _r
    _c_py = _c
    _rowcnd_py = _return_scalars.rowcnd
    _colcnd_py = _return_scalars.colcnd
    _amax_py = _return_scalars.amax
    _return_data = ZgbequReturnData(
        _r_py,
        _c_py,
        _rowcnd_py,
        _colcnd_py,
        _amax_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgbcon(norm, kl, ku, ab, ipiv, anorm):
    u"""
    Estimate condition number of complex band matrix, matrix already
    factorized by ``zgbtrf``.

    ``zgbcon`` estimates the condition number of a complex band matrix
    A, where A has been factorized by ``zgbtrf``.

    For full information please refer to the NAG Library document for
    f07bu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07buf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : complex, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``zgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``zgbtrf``.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.zlangb`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``zgbtrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -8)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``zgbcon`` estimates the condition number of a complex band matrix
    A, in either the 1-norm or the infinity-norm:

        kappa_1(A) = ||A||_1||A^-1||_1 or kappa_infinity (A) =
        ||A||_infinity||A^-1||_infinity.

    Note that kappa_infinity (A) = kappa_1(A^H).

    Because the condition number is infinite if A is singular, the
    function actually returns an estimate of the **reciprocal** of the
    condition number.

    The function should be preceded by a call to ``blas.zlangb`` to
    compute ||A||_1 or ||A||_infinity, and a call to ``zgbtrf`` to
    compute the LU factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1 or ||A^-1||_infinity.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zgbcon(
        norm,
        _n,
        kl,
        ku,
        _ab,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZgbrfsReturnData = _collections.namedtuple(
    'ZgbrfsReturnData',
    ['x', 'ferr', 'berr'],
)

def zgbrfs(trans, kl, ku, nrhs, ab, afb, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex band system of linear
    equations, multiple right-hand sides.

    ``zgbrfs`` returns error bounds for the solution of a complex band
    system of linear equations with multiple right-hand sides, AX = B,
    A^TX = B or A^HX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07bv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07bvf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the linear equations for which X is the
        computed solution as follows:

        `trans` = 'N'

            The linear equations are of the form AX = B.

        `trans` = 'T'

            The linear equations are of the form A^TX = B.

        `trans` = 'C'

            The linear equations are of the form A^HX = B.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides.

    ab : complex, array-like, shape (`kl`+`ku`+1, n)
        The original n*n band matrix A as supplied to ``zgbtrf`` but
        with reduced requirements since the matrix is not factorized.

        See [Further Comments] for further details.

    afb : complex, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``zgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``zgbtrf``.

    b : complex, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (n, `nrhs`)
        The n*r solution matrix X, as returned by ``zgbtrs``.

    Returns
    -------
    x : complex, ndarray, shape (n, `nrhs`)
        The improved solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -5)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Notes
    -----
    ``zgbrfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex band system of linear
    equations with multiple right-hand sides AX = B, A^TX = B or A^HX =
    B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``zgbrfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise` `backward` `error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise`
    `forward` `error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zgbrfs(
        trans,
        _n,
        kl,
        ku,
        _nrhs,
        _ab,
        _afb,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZgbrfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

DgtsvReturnData = _collections.namedtuple(
    'DgtsvReturnData',
    ['dl', 'd', 'du', 'b'],
)

def dgtsv(dl, d, du, b):
    u"""
    Computes the solution to a real tridiagonal system of linear
    equations.

    ``dgtsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n*n tridiagonal matrix and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07ca

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07caf.html

    Parameters
    ----------
    dl : float, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the matrix A.

    du : float, array-like, shape (n-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    dl : float, ndarray, shape (n-1)
        If no constraints are violated, `dl` is overwritten by the (n-2)
        elements of the second superdiagonal of the upper triangular
        matrix U from the LU factorization of A, in
        `dl`[0],`dl`[1],...,`dl`[n-3].

    d : float, ndarray, shape (n)
        If no constraints are violated, `d` is overwritten by the n
        diagonal elements of the upper triangular matrix U from the LU
        factorization of A.

    du : float, ndarray, shape (n-1)
        If no constraints are violated, `du` is overwritten by the (n-1)
        elements of the first superdiagonal of U.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero, and the
            solution has not been computed. The factorization has not
            been completed unless n = *<value>*.

    Notes
    -----
    ``dgtsv`` uses Gaussian elimination with partial pivoting and row
    interchanges to solve the equations AX = B.
    The matrix A is factorized as A = PLU, where P is a permutation
    matrix, L is unit lower triangular with at most one nonzero
    subdiagonal element per column, and U is an upper triangular band
    matrix, with two superdiagonals.

    Note that equations A^TX = B may be solved by interchanging the
    order of the arguments `du` and `dl`.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dgtsv'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dgtsv(
            _n,
            _nrhs,
            _dl,
            _d,
            _du,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _dl_py = _dl
    _d_py = _d
    _du_py = _du
    _b_py = _b
    _return_data = DgtsvReturnData(
        _dl_py,
        _d_py,
        _du_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DgtsvxReturnData = _collections.namedtuple(
    'DgtsvxReturnData',
    ['dlf', 'df', 'duf', 'du2', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
)

def dgtsvx(fact, trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a real tridiagonal system of linear
    equations.

    ``dgtsvx`` uses the LU factorization to compute the solution to a
    real system of linear equations

        AX = B or A^TX = B,

    where A is a tridiagonal matrix of order n and X and B are n*r
    matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07cb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07cbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `dlf`, `df`, `duf`, `du2` and `ipiv` contain the factorized
            form of the matrix A. `dlf`, `df`, `duf`, `du2` and `ipiv`
            will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `dlf`, `df` and `duf` and
            factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T' or 'C'

            A^TX = B (Transpose).

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    dl : float, array-like, shape (`n`-1)
        The (n-1) subdiagonal elements of A.

    d : float, array-like, shape (`n`)
        The n diagonal elements of A.

    du : float, array-like, shape (`n`-1)
        The (n-1) superdiagonal elements of A.

    dlf : float, array-like, shape (`n`-1)
        If `fact` = 'F', `dlf` contains the (n-1) multipliers that
        define the matrix L from the LU factorization of A.

    df : float, array-like, shape (`n`)
        If `fact` = 'F', `df` contains the n diagonal elements of the
        upper triangular matrix U from the LU factorization of A.

    duf : float, array-like, shape (`n`-1)
        If `fact` = 'F', `duf` contains the (n-1) elements of the first
        superdiagonal of U.

    du2 : float, array-like, shape (`n`-2)
        If `fact` = 'F', `du2` contains the (n-2) elements of the second
        superdiagonal of U.

    ipiv : int, array-like, shape (`n`)
        If `fact` = 'F', `ipiv` contains the pivot indices from the LU
        factorization of A.

    b : float, array-like, shape (`n`, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    dlf : float, ndarray, shape (`n`-1)
        If `fact` = 'N', `dlf` contains the (n-1) multipliers that
        define the matrix L from the LU factorization of A.

    df : float, ndarray, shape (`n`)
        If `fact` = 'N', `df` contains the n diagonal elements of the
        upper triangular matrix U from the LU factorization of A.

    duf : float, ndarray, shape (`n`-1)
        If `fact` = 'N', `duf` contains the (n-1) elements of the first
        superdiagonal of U.

    du2 : float, ndarray, shape (`n`-2)
        If `fact` = 'N', `du2` contains the (n-2) elements of the second
        superdiagonal of U.

    ipiv : int, ndarray, shape (`n`)
        If `fact` = 'N', `ipiv` contains the pivot indices from the LU
        factorization of A; row i of the matrix was interchanged with
        row `ipiv`[i-1]. `ipiv`[i-1] will always be either i or i+1;
        `ipiv`[i-1] = i indicates a row interchange was not required.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i < `n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has not been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` (i > 0) and (`n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` `n`+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``dgtsvx`` performs the following steps:

    (1) If `fact` = 'N', the LU decomposition is used to factor the
        matrix A as A = LU, where L is a product of permutation and unit
        lower bidiagonal matrices and U is upper triangular with
        nonzeros in only the main diagonal and first two superdiagonals.

    (#) If some u_{ii} = 0, so that U is exactly singular, then the
        function returns with errno = i. Otherwise, the factored form of
        A is used to estimate the condition number of the matrix A. If
        the reciprocal of the condition number is less than machine
        precision, `errno` = `n` + 1 is returned as a warning, but the
        function still goes on to solve for X and compute error bounds
        as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dgtsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _dlf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dlf'},
        dlf,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _duf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'duf'},
        duf,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.dgtsvx(
            fact,
            trans,
            _n,
            _nrhs,
            _dl,
            _d,
            _du,
            _dlf,
            _df,
            _duf,
            _du2,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _dlf_py = _dlf
    _df_py = _df
    _duf_py = _duf
    _du2_py = _du2
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DgtsvxReturnData(
        _dlf_py,
        _df_py,
        _duf_py,
        _du2_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DgttrfReturnData = _collections.namedtuple(
    'DgttrfReturnData',
    ['dl', 'd', 'du', 'du2', 'ipiv'],
)

def dgttrf(n, dl, d, du):
    u"""
    LU factorization of real tridiagonal matrix.

    ``dgttrf`` computes the LU factorization of a real n*n tridiagonal
    matrix A.

    For full information please refer to the NAG Library document for
    f07cd

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07cdf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    dl : float, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the matrix A.

    du : float, array-like, shape (`n`-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    Returns
    -------
    dl : float, ndarray, shape (`n`-1)
        Is overwritten by the (n-1) multipliers that define the matrix L
        of the LU factorization of A.

    d : float, ndarray, shape (`n`)
        Is overwritten by the n diagonal elements of the upper
        triangular matrix U from the LU factorization of A.

    du : float, ndarray, shape (`n`-1)
        Is overwritten by the (n-1) elements of the first superdiagonal
        of U.

    du2 : float, ndarray, shape (`n`-2)
        Contains the (n-2) elements of the second superdiagonal of U.

    ipiv : int, ndarray, shape (`n`)
        Contains the n pivot indices that define the permutation matrix
        P. At the ith step, row i of the matrix was interchanged with
        row `ipiv`[i-1]. `ipiv`[i-1] will always be either i or (i+1),
        `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    Notes
    -----
    ``dgttrf`` uses Gaussian elimination with partial pivoting and row
    interchanges to factorize the matrix A as

        A = PLU,

    where P is a permutation matrix, L is unit lower triangular with at
    most one nonzero subdiagonal element in each column, and U is an
    upper triangular band matrix, with two superdiagonals.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dgttrf'

    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _du2 = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py-2,),
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dgttrf(
            _n,
            _dl,
            _d,
            _du,
            _du2,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _dl_py = _dl
    _d_py = _d
    _du_py = _du
    _du2_py = _du2
    _ipiv_py = _ipiv
    _return_data = DgttrfReturnData(
        _dl_py,
        _d_py,
        _du_py,
        _du2_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgttrs(trans, dl, d, du, du2, ipiv, b):
    u"""
    Solves a real tridiagonal system of linear equations using the LU
    factorization computed by ``dgttrf``.

    ``dgttrs`` computes the solution to a real system of linear
    equations AX = B or A^TX = B, where A is an n*n tridiagonal matrix
    and X and B are n*r matrices, using the LU factorization returned by
    ``dgttrf``.

    For full information please refer to the NAG Library document for
    f07ce

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07cef.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the equations to be solved as follows:

        `trans` = 'N'

            Solve AX = B for X.

        `trans` = 'T' or 'C'

            Solve A^TX = B for X.

    dl : float, array-like, shape (n-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    du : float, array-like, shape (n-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : float, array-like, shape (n-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (n)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    b : float, array-like, shape (n, nrhs)
        The n*r matrix of right-hand sides B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dgttrs`` should be preceded by a call to ``dgttrf``, which uses
    Gaussian elimination with partial pivoting and row interchanges to
    factorize the matrix A as

        A = PLU,

    where P is a permutation matrix, L is unit lower triangular with at
    most one nonzero subdiagonal element in each column, and U is an
    upper triangular band matrix, with two superdiagonals. ``dgttrs``
    then utilizes the factorization to solve the required equations.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dgttrs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.dgttrs(
        trans,
        _n,
        _nrhs,
        _dl,
        _d,
        _du,
        _du2,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def dgtcon(norm, n, dl, d, du, du2, ipiv, anorm):
    u"""
    Estimates the reciprocal of the condition number of a real
    tridiagonal matrix using the LU factorization computed by
    ``dgttrf``.

    ``dgtcon`` estimates the reciprocal condition number of a real n*n
    tridiagonal matrix A, using the LU factorization returned by
    ``dgttrf``.

    For full information please refer to the NAG Library document for
    f07cg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07cgf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the norm to be used to estimate kappa(A).

        `norm` = '1' or 'O'

            Estimate kappa_1(A).

        `norm` = 'I'

            Estimate kappa_infinity (A).

    n : int
        n, the order of the matrix A.

    dl : float, array-like, shape (`n`-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    du : float, array-like, shape (`n`-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : float, array-like, shape (`n`-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (`n`)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.dlangt`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``dgttrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        Contains an estimate of the reciprocal condition number.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -8)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``dgtcon`` should be preceded by a call to ``dgttrf``, which uses
    Gaussian elimination with partial pivoting and row interchanges to
    factorize the matrix A as

        A = PLU,

    where P is a permutation matrix, L is unit lower triangular with at
    most one nonzero subdiagonal element in each column, and U is an
    upper triangular band matrix, with two superdiagonals. ``dgtcon``
    then utilizes the factorization to estimate either ||A^-1||_1 or
    ||A^-1||_infinity, from which the estimate of the reciprocal of the
    condition number of A, 1/kappa(A) is computed as either

        1/kappa_1(A) = 1/(||A||_1||A^-1||_1)

    or

        1/kappa_infinity (A) = 1/(||A||_infinity||A^-1||_infinity).

    1/kappa(A) is returned, rather than kappa(A), since when A is
    singular kappa(A) is infinite.

    Note that kappa_infinity (A) = kappa_1(A^T).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dgtcon'

    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.dgtcon(
        norm,
        n,
        _dl,
        _d,
        _du,
        _du2,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DgtrfsReturnData = _collections.namedtuple(
    'DgtrfsReturnData',
    ['x', 'ferr', 'berr'],
)

def dgtrfs(trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b, x):
    u"""
    Refined solution with error bounds of real tridiagonal system of
    linear equations, multiple right-hand sides.

    ``dgtrfs`` computes error bounds and refines the solution to a real
    system of linear equations AX = B or A^TX = B, where A is an n*n
    tridiagonal matrix and X and B are n*r matrices, using the LU
    factorization returned by ``dgttrf`` and an initial solution
    returned by ``dgttrs``.
    Iterative refinement is used to reduce the backward error as much as
    possible.

    For full information please refer to the NAG Library document for
    f07ch

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07chf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the equations to be solved as follows:

        `trans` = 'N'

            Solve AX = B for X.

        `trans` = 'T' or 'C'

            Solve A^TX = B for X.

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    dl : float, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the matrix A.

    du : float, array-like, shape (`n`-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    dlf : float, array-like, shape (`n`-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    df : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    duf : float, array-like, shape (`n`-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : float, array-like, shape (`n`-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (`n`)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    b : float, array-like, shape (`n`, `nrhs`)
        The n*r matrix of right-hand sides B.

    x : float, array-like, shape (`n`, `nrhs`)
        The n*r initial solution matrix X.

    Returns
    -------
    x : float, ndarray, shape (`n`, `nrhs`)
        The n*r refined solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        Estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x^_j||_infinity <=
        `ferr`[j-1], where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is almost always a
        slight overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        Estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Notes
    -----
    ``dgtrfs`` should normally be preceded by calls to ``dgttrf`` and
    ``dgttrs``. ``dgttrf`` uses Gaussian elimination with partial
    pivoting and row interchanges to factorize the matrix A as

        A = PLU,

    where P is a permutation matrix, L is unit lower triangular with at
    most one nonzero subdiagonal element in each column, and U is an
    upper triangular band matrix, with two superdiagonals. ``dgttrs``
    then utilizes the factorization to compute a solution, X^, to the
    required equations.
    Letting x^ denote a column of X^, ``dgtrfs`` computes a
    `component-wise backward error`, beta, the smallest relative
    perturbation in each element of A and b such that x^ is the exact
    solution of a perturbed system

        (A+E)x^= b+f, with |e_{ij}| <= beta|a_{ij}|, and |f_j| <=
        beta|b_j|.

    The function also estimates a bound for the `component-wise forward
    error` in the computed solution defined by max|x_i-x_i^|/max|x_i^|,
    where x is the corresponding column of the exact solution, X.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dgtrfs'

    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _dlf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dlf'},
        dlf,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _duf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'duf'},
        duf,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dgtrfs(
        trans,
        n,
        _nrhs,
        _dl,
        _d,
        _du,
        _dlf,
        _df,
        _duf,
        _du2,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DgtrfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZgtsvReturnData = _collections.namedtuple(
    'ZgtsvReturnData',
    ['dl', 'd', 'du', 'b'],
)

def zgtsv(dl, d, du, b):
    u"""
    Computes the solution to a complex tridiagonal system of linear
    equations.

    ``zgtsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n tridiagonal matrix and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07cn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07cnf.html

    Parameters
    ----------
    dl : complex, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : complex, array-like, shape (n)
        Must contain the n diagonal elements of the matrix A.

    du : complex, array-like, shape (n-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    dl : complex, ndarray, shape (n-1)
        If no constraints are violated, `dl` is overwritten by the (n-2)
        elements of the second superdiagonal of the upper triangular
        matrix U from the LU factorization of A, in
        `dl`[0],`dl`[1],...,`dl`[n-3].

    d : complex, ndarray, shape (n)
        If no constraints are violated, `d` is overwritten by the n
        diagonal elements of the upper triangular matrix U from the LU
        factorization of A.

    du : complex, ndarray, shape (n-1)
        If no constraints are violated, `du` is overwritten by the (n-1)
        elements of the first superdiagonal of U.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero, and the
            solution has not been computed. The factorization has not
            been completed unless n = *<value>*.

    Notes
    -----
    ``zgtsv`` uses Gaussian elimination with partial pivoting and row
    interchanges to solve the equations AX = B.
    The matrix A is factorized as A = PLU, where P is a permutation
    matrix, L is unit lower triangular with at most one nonzero
    subdiagonal element per column, and U is an upper triangular band
    matrix, with two superdiagonals.

    Note that the equations A^TX = B may be solved by interchanging the
    order of the arguments `du` and `dl`.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.zgtsv'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=True,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=True,
    )
    try:
        _b_lapacklin.zgtsv(
            _n,
            _nrhs,
            _dl,
            _d,
            _du,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _dl_py = _dl
    _d_py = _d
    _du_py = _du
    _b_py = _b
    _return_data = ZgtsvReturnData(
        _dl_py,
        _d_py,
        _du_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZgtsvxReturnData = _collections.namedtuple(
    'ZgtsvxReturnData',
    ['dlf', 'df', 'duf', 'du2', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
)

def zgtsvx(fact, trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a complex tridiagonal system of linear
    equations.

    ``zgtsvx`` uses the LU factorization to compute the solution to a
    complex system of linear equations

        AX = B, A^TX = B or A^HX = B,

    where A is a tridiagonal matrix of order n and X and B are n*r
    matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07cp

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07cpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `dlf`, `df`, `duf`, `du2` and `ipiv` contain the factorized
            form of the matrix A. `dlf`, `df`, `duf`, `du2` and `ipiv`
            will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `dlf`, `df` and `duf` and
            factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T'

            A^TX = B (Transpose).

        `trans` = 'C'

            A^HX = B (Conjugate transpose).

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    dl : complex, array-like, shape (`n`-1)
        The (n-1) subdiagonal elements of A.

    d : complex, array-like, shape (`n`)
        The n diagonal elements of A.

    du : complex, array-like, shape (`n`-1)
        The (n-1) superdiagonal elements of A.

    dlf : complex, array-like, shape (`n`-1)
        If `fact` = 'F', `dlf` contains the (n-1) multipliers that
        define the matrix L from the LU factorization of A.

    df : complex, array-like, shape (`n`)
        If `fact` = 'F', `df` contains the n diagonal elements of the
        upper triangular matrix U from the LU factorization of A.

    duf : complex, array-like, shape (`n`-1)
        If `fact` = 'F', `duf` contains the (n-1) elements of the first
        superdiagonal of U.

    du2 : complex, array-like, shape (`n`-2)
        If `fact` = 'F', `du2` contains the (n-2) elements of the second
        superdiagonal of U.

    ipiv : int, array-like, shape (`n`)
        If `fact` = 'F', `ipiv` contains the pivot indices from the LU
        factorization of A.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    dlf : complex, ndarray, shape (`n`-1)
        If `fact` = 'N', `dlf` contains the (n-1) multipliers that
        define the matrix L from the LU factorization of A.

    df : complex, ndarray, shape (`n`)
        If `fact` = 'N', `df` contains the n diagonal elements of the
        upper triangular matrix U from the LU factorization of A.

    duf : complex, ndarray, shape (`n`-1)
        If `fact` = 'N', `duf` contains the (n-1) elements of the first
        superdiagonal of U.

    du2 : complex, ndarray, shape (`n`-2)
        If `fact` = 'N', `du2` contains the (n-2) elements of the second
        superdiagonal of U.

    ipiv : int, ndarray, shape (`n`)
        If `fact` = 'N', `ipiv` contains the pivot indices from the LU
        factorization of A; row i of the matrix was interchanged with
        row `ipiv`[i-1]. `ipiv`[i-1] will always be either i or i+1;
        `ipiv`[i-1] = i indicates a row interchange was not required.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i < `n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has not been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` (i > 0) and (`n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` `n`+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``zgtsvx`` performs the following steps:

    (1) If `fact` = 'N', the LU decomposition is used to factor the
        matrix A as A = LU, where L is a product of permutation and unit
        lower bidiagonal matrices and U is upper triangular with
        nonzeros in only the main diagonal and first two superdiagonals.

    (#) If some u_{ii} = 0, so that U is exactly singular, then the
        function returns with errno = i. Otherwise, the factored form of
        A is used to estimate the condition number of the matrix A. If
        the reciprocal of the condition number is less than machine
        precision, `errno` = `n` + 1 is returned as a warning, but the
        function still goes on to solve for X and compute error bounds
        as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zgtsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
    )
    _dlf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dlf'},
        dlf,
        copy=True,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=True,
    )
    _duf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'duf'},
        duf,
        copy=True,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=True,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zgtsvx(
            fact,
            trans,
            _n,
            _nrhs,
            _dl,
            _d,
            _du,
            _dlf,
            _df,
            _duf,
            _du2,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _dlf_py = _dlf
    _df_py = _df
    _duf_py = _duf
    _du2_py = _du2
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZgtsvxReturnData(
        _dlf_py,
        _df_py,
        _duf_py,
        _du2_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZgttrfReturnData = _collections.namedtuple(
    'ZgttrfReturnData',
    ['dl', 'd', 'du', 'du2', 'ipiv'],
)

def zgttrf(n, dl, d, du):
    u"""
    LU factorization of complex tridiagonal matrix.

    ``zgttrf`` computes the LU factorization of a complex n*n
    tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f07cr

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07crf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    dl : complex, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : complex, array-like, shape (`n`)
        Must contain the n diagonal elements of the matrix A.

    du : complex, array-like, shape (`n`-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    Returns
    -------
    dl : complex, ndarray, shape (`n`-1)
        Is overwritten by the (n-1) multipliers that define the matrix L
        of the LU factorization of A.

    d : complex, ndarray, shape (`n`)
        Is overwritten by the n diagonal elements of the upper
        triangular matrix U from the LU factorization of A.

    du : complex, ndarray, shape (`n`-1)
        Is overwritten by the (n-1) elements of the first superdiagonal
        of U.

    du2 : complex, ndarray, shape (`n`-2)
        Contains the (n-2) elements of the second superdiagonal of U.

    ipiv : int, ndarray, shape (`n`)
        Contains the n pivot indices that define the permutation matrix
        P. At the ith step, row i of the matrix was interchanged with
        row `ipiv`[i-1]. `ipiv`[i-1] will always be either i or (i+1),
        `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    Notes
    -----
    ``zgttrf`` uses Gaussian elimination with partial pivoting and row
    interchanges to factorize the matrix A as

        A = PLU,

    where P is a permutation matrix, L is unit lower triangular with at
    most one nonzero subdiagonal element in each column, and U is an
    upper triangular band matrix, with two superdiagonals.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.zgttrf'

    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=True,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _du2 = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py-2,),
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zgttrf(
            _n,
            _dl,
            _d,
            _du,
            _du2,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _dl_py = _dl
    _d_py = _d
    _du_py = _du
    _du2_py = _du2
    _ipiv_py = _ipiv
    _return_data = ZgttrfReturnData(
        _dl_py,
        _d_py,
        _du_py,
        _du2_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgttrs(trans, dl, d, du, du2, ipiv, b):
    u"""
    Solves a complex tridiagonal system of linear equations using the LU
    factorization computed by ``dgttrf``.

    ``zgttrs`` computes the solution to a complex system of linear
    equations AX = B or A^TX = B or A^HX = B, where A is an n*n
    tridiagonal matrix and X and B are n*r matrices, using the LU
    factorization returned by ``zgttrf``.

    For full information please refer to the NAG Library document for
    f07cs

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07csf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the equations to be solved as follows:

        `trans` = 'N'

            Solve AX = B for X.

        `trans` = 'T'

            Solve A^TX = B for X.

        `trans` = 'C'

            Solve A^HX = B for X.

    dl : complex, array-like, shape (n-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    d : complex, array-like, shape (n)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    du : complex, array-like, shape (n-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : complex, array-like, shape (n-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (n)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    b : complex, array-like, shape (n, nrhs)
        The n*r matrix of right-hand sides B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zgttrs`` should be preceded by a call to ``zgttrf``, which uses
    Gaussian elimination with partial pivoting and row interchanges to
    factorize the matrix A as

        A = PLU,

    where P is a permutation matrix, L is unit lower triangular with at
    most one nonzero subdiagonal element in each column, and U is an
    upper triangular band matrix, with two superdiagonals. ``zgttrs``
    then utilizes the factorization to solve the required equations.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgttrs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zgttrs(
        trans,
        _n,
        _nrhs,
        _dl,
        _d,
        _du,
        _du2,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zgtcon(norm, dl, d, du, du2, ipiv, anorm):
    u"""
    Estimates the reciprocal of the condition number of a complex
    tridiagonal matrix using the LU factorization computed by
    ``dgttrf``.

    ``zgtcon`` estimates the reciprocal condition number of a complex
    n*n tridiagonal matrix A, using the LU factorization returned by
    ``zgttrf``.

    For full information please refer to the NAG Library document for
    f07cu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07cuf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the norm to be used to estimate kappa(A).

        `norm` = '1' or 'O'

            Estimate kappa_1(A).

        `norm` = 'I'

            Estimate kappa_infinity (A).

    dl : complex, array-like, shape (n-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    d : complex, array-like, shape (n)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    du : complex, array-like, shape (n-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : complex, array-like, shape (n-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (n)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.zlangt`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``zgttrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        Contains an estimate of the reciprocal condition number.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -8)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``zgtcon`` should be preceded by a call to ``zgttrf``, which uses
    Gaussian elimination with partial pivoting and row interchanges to
    factorize the matrix A as

        A = PLU,

    where P is a permutation matrix, L is unit lower triangular with at
    most one nonzero subdiagonal element in each column, and U is an
    upper triangular band matrix, with two superdiagonals. ``zgtcon``
    then utilizes the factorization to estimate either ||A^-1||_1 or
    ||A^-1||_infinity, from which the estimate of the reciprocal of the
    condition number of A, 1/kappa(A) is computed as either

        1/kappa_1(A) = 1/(||A||_1||A^-1||_1)

    or

        1/kappa_infinity (A) = 1/(||A||_infinity||A^-1||_infinity).

    1/kappa(A) is returned, rather than kappa(A), since when A is
    singular kappa(A) is infinite.

    Note that kappa_infinity (A) = kappa_1(A^T).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zgtcon'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zgtcon(
        norm,
        _n,
        _dl,
        _d,
        _du,
        _du2,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZgtrfsReturnData = _collections.namedtuple(
    'ZgtrfsReturnData',
    ['x', 'ferr', 'berr'],
)

def zgtrfs(trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex tridiagonal system of
    linear equations, multiple right-hand sides.

    ``zgtrfs`` computes error bounds and refines the solution to a
    complex system of linear equations AX = B or A^TX = B or A^HX = B,
    where A is an n*n tridiagonal matrix and X and B are n*r matrices,
    using the LU factorization returned by ``zgttrf`` and an initial
    solution returned by ``zgttrs``.
    Iterative refinement is used to reduce the backward error as much as
    possible.

    For full information please refer to the NAG Library document for
    f07cv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07cvf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the equations to be solved as follows:

        `trans` = 'N'

            Solve AX = B for X.

        `trans` = 'T'

            Solve A^TX = B for X.

        `trans` = 'C'

            Solve A^HX = B for X.

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    dl : complex, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : complex, array-like, shape (`n`)
        Must contain the n diagonal elements of the matrix A.

    du : complex, array-like, shape (`n`-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    dlf : complex, array-like, shape (`n`-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    df : complex, array-like, shape (`n`)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    duf : complex, array-like, shape (`n`-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : complex, array-like, shape (`n`-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (`n`)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n*r matrix of right-hand sides B.

    x : complex, array-like, shape (`n`, `nrhs`)
        The n*r initial solution matrix X.

    Returns
    -------
    x : complex, ndarray, shape (`n`, `nrhs`)
        The n*r refined solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        Estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x^_j||_infinity <=
        `ferr`[j-1], where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is almost always a
        slight overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        Estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Notes
    -----
    ``zgtrfs`` should normally be preceded by calls to ``zgttrf`` and
    ``zgttrs``. ``zgttrf`` uses Gaussian elimination with partial
    pivoting and row interchanges to factorize the matrix A as

        A = PLU,

    where P is a permutation matrix, L is unit lower triangular with at
    most one nonzero subdiagonal element in each column, and U is an
    upper triangular band matrix, with two superdiagonals. ``zgttrs``
    then utilizes the factorization to compute a solution, X^, to the
    required equations.
    Letting x^ denote a column of X^, ``zgtrfs`` computes a
    `component-wise backward error`, beta, the smallest relative
    perturbation in each element of A and b such that x^ is the exact
    solution of a perturbed system

        (A+E)x^= b+f, with |e_{ij}| <= beta|a_{ij}|, and |f_j| <=
        beta|b_j|.

    The function also estimates a bound for the `component-wise forward
    error` in the computed solution defined by max|x_i-x_i^|/max|x_i^|,
    where x is the corresponding column of the exact solution, X.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.zgtrfs'

    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
    )
    _dlf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dlf'},
        dlf,
        copy=False,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
    )
    _duf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'duf'},
        duf,
        copy=False,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zgtrfs(
        trans,
        n,
        _nrhs,
        _dl,
        _d,
        _du,
        _dlf,
        _df,
        _duf,
        _du2,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZgtrfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

DposvReturnData = _collections.namedtuple(
    'DposvReturnData',
    ['a', 'b'],
)

def dposv(uplo, a, b):
    u"""
    Computes the solution to a real symmetric positive definite system
    of linear equations.

    ``dposv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric positive definite matrix and X and B are
    n*r matrices.

    For full information please refer to the NAG Library document for
    f07fa

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07faf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : float, array-like, shape (n, n)
        The n*n symmetric matrix A.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^TU or A = LL^T.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    Notes
    -----
    ``dposv`` uses the Cholesky decomposition to factor A as A = U^TU if
    `uplo` = 'U' or A = LL^T if `uplo` = 'L', where U is an upper
    triangular matrix and L is a lower triangular matrix.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dposv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.dposv(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
    )
    _a_py = _a
    _b_py = _b
    _return_data = DposvReturnData(
        _a_py,
        _b_py,
    )
    return _return_data

DposvxReturnData = _collections.namedtuple(
    'DposvxReturnData',
    ['a', 'af', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
)

def dposvx(fact, uplo, n, nrhs, a, af, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a real symmetric positive definite system
    of linear equations.

    ``dposvx`` uses the Cholesky factorization

        A = U^TU or A = LL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric positive definite matrix and X and B are
    n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07fb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `af` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `a` and `af` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `af` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : float, array-like, shape (`n`, `n`)
        The n*n symmetric matrix A.

        If `fact` = 'F' and `equed` = 'Y', `a` must have been
        equilibrated by the scaling factor in `s` as D_S AD_S.

    af : float, array-like, shape (`n`, `n`)
        If `fact` = 'F', `af` contains the triangular factor U or L from
        the Cholesky factorization A = U^TU or A = LL^T, in the same
        storage format as `a`. If `equed` != 'N', `af` is the factorized
        form of the equilibrated matrix D_S AD_S.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : float, array-like, shape (`n`, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `a` is not modified.

        If `fact` = 'E' and `equed` = 'Y', `a` is overwritten by D_S
        AD_S.

    af : float, ndarray, shape (`n`, `n`)
        If `fact` = 'N', `af` returns the triangular factor U or L from
        the Cholesky factorization A = U^TU or A = LL^T of the original
        matrix A.

        If `fact` = 'E', `af` returns the triangular factor U or L from
        the Cholesky factorization A = U^TU or A = LL^T of the
        equilibrated matrix A (see the description of `a` for the form
        of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : float, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X to the original system of equations. Note that
        the arrays A and B are modified on exit if `equed` = 'Y', and
        the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -9)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -10)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``dposvx`` performs the following steps:

    (1) If `fact` = 'E', real diagonal scaling factors, D_S, are
        computed to equilibrate the system:

            (D_S AD_S)(D_S^-1X) = D_S B.

        Whether or not the system will be equilibrated depends on the
        scaling of the matrix A, but if equilibration is used, A is
        overwritten by D_S AD_S and B*D_S B.

    (#) If `fact` = 'N' or 'E', the Cholesky decomposition is used to
        factor the matrix A (after equilibration if `fact` = 'E') as A =
        U^TU if `uplo` = 'U' or A = LL^T if `uplo` = 'L', where U is an
        upper triangular matrix and L is a lower triangular matrix.

    (#) If the leading i*i principal minor of A is not positive
        definite, then the function returns with errno = i. Otherwise,
        the factored form of A is used to estimate the condition number
        of the matrix A. If the reciprocal of the condition number is
        less than machine precision, `errno` = `n` + 1 is returned as a
        warning, but the function still goes on to solve for X and
        compute error bounds as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    (#) If equilibration was used, the matrix X is premultiplied by D_S
        so that it solves the original system before equilibration.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dposvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dposvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _a,
            _af,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _af_py = _af
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DposvxReturnData(
        _a_py,
        _af_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DsposvReturnData = _collections.namedtuple(
    'DsposvReturnData',
    ['a', 'x', 'itera'],
)

def dsposv(uplo, n, nrhs, a, b):
    u"""
    Computes the solution to a real symmetric positive definite system
    of linear equations using mixed precision arithmetic.

    ``dsposv`` uses the Cholesky factorization

        A = U^TU or A = LL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric positive definite matrix and X and B are
    n*r matrices.

    For full information please refer to the NAG Library document for
    f07fc

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fcf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

            The upper triangular part of A is stored.

        `uplo` = 'L'

            The lower triangular part of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : float, array-like, shape (`n`, `n`)
        The n*n symmetric positive definite matrix A.

    b : float, array-like, shape (`n`, `nrhs`)
        The right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        If iterative refinement has been successfully used (no exception
        or warning is raised and `itera` >= 0, see `itera`), then `a` is
        unchanged. If full precision factorization has been used (no
        exception or warning is raised and `itera` < 0, see `itera`),
        then the array A contains the factor U or L from the Cholesky
        factorization A = U^TU or A = LL^T.

    x : float, ndarray, shape (`n`, `nrhs`)
        If no exception or warning is raised, the n*r solution matrix X.

    itera : int
        Information on the progress of the interative refinement
        process.

        `itera` < 0

            Iterative refinement has failed for one of the reasons given
            below, full precision factorization has been performed
            instead.

            [table omitted]

        `itera` > 0

            Iterative refinement has been sucessfully used. `itera`
            returns the number of iterations.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``dsposv`` first attempts to factorize the matrix in reduced
    precision and use this factorization within an iterative refinement
    procedure to produce a solution with full precision normwise
    backward error quality (see below).
    If the approach fails the method switches to a full precision
    factorization and solve.

    The iterative refinement can be more efficient than the
    corresponding direct full precision algorithm.
    Since the strategy implemented by ``dsposv`` must perform iterative
    refinement on each right-hand side, any efficiency gains will reduce
    as the number of right-hand sides increases.
    Conversely, as the matrix size increases the cost of these iterative
    refinements become less significant relative to the cost of
    factorization.
    Thus, any efficiency gains will be greatest for a very small number
    of right-hand sides and for large matrix sizes.
    The cut-off values for the number of right-hand sides and matrix
    size, for which the iterative refinement strategy performs better,
    depends on the relative performance of the reduced and full
    precision factorization and back-substitution. ``dsposv`` always
    attempts the iterative refinement strategy first; you are advised to
    compare the performance of ``dsposv`` with that of its full
    precision counterpart ``dposv`` to determine whether this strategy
    is worthwhile for your particular problem dimensions.

    The iterative refinement process is stopped if `itera` > 30 where
    `itera` is the number of iterations carried out thus far.
    The process is also stopped if for all right-hand sides we have

        ||resid|| < sqrt(`n`)||x||||A||epsilon,

    where ||resid|| is the infinity-norm of the residual, ||x|| is the
    infinity-norm of the solution, ||A|| is the infinity-norm of the
    matrix A and epsilon is the machine precision returned by
    ``machine.precision``.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dsposv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _return_scalar = _b_lapacklin.dsposv(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
        _x,
    )
    _a_py = _a
    _x_py = _x
    _itera_py = _return_scalar
    _return_data = DsposvReturnData(
        _a_py,
        _x_py,
        _itera_py,
    )
    return _return_data

def dpotrf(uplo, a):
    u"""
    Cholesky factorization of real symmetric positive definite matrix.

    ``dpotrf`` computes the Cholesky factorization of a real symmetric
    positive definite matrix.

    For full information please refer to the NAG Library document for
    f07fd

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    a : float, array-like, shape (n, n)
        The n*n symmetric positive definite matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The upper or lower triangle of A is overwritten by the Cholesky
        factor U or L as specified by `uplo`.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. To factorize a symmetric
            matrix which is not positive definite, call ``dsytrf``
            instead.

    Notes
    -----
    ``dpotrf`` forms the Cholesky factorization of a real symmetric
    positive definite matrix A either as A = U^TU if `uplo` = 'U' or A =
    LL^T if `uplo` = 'L', where U is an upper triangular matrix and L is
    lower triangular.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpotrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_lapacklin.dpotrf(
        uplo,
        _n,
        _a,
    )
    _a_py = _a
    return _a_py

def dpotrs(uplo, a, b):
    u"""
    Solution of real symmetric positive definite system of linear
    equations, multiple right-hand sides, matrix already factorized by
    ``dpotrf``.

    ``dpotrs`` solves a real symmetric positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``dpotrf``.

    For full information please refer to the NAG Library document for
    f07fe

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    a : float, array-like, shape (n, n)
        The Cholesky factor of A, as returned by ``dpotrf``.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dpotrs`` is used to solve a real symmetric positive definite
    system of linear equations AX = B, this function must be preceded by
    a call to ``dpotrf`` which computes the Cholesky factorization of A.
    The solution X is computed by forward and backward substitution.

    If `uplo` = 'U', A = U^TU, where U is upper triangular; the solution
    X is computed by solving U^TY = B and then UX = Y.

    If `uplo` = 'L', A = LL^T, where L is lower triangular; the solution
    X is computed by solving LY = B and then L^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpotrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.dpotrs(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

DpoequReturnData = _collections.namedtuple(
    'DpoequReturnData',
    ['s', 'scond', 'amax'],
)

def dpoequ(n, a):
    u"""
    Computes row and column scalings intended to equilibrate a real
    symmetric positive definite matrix and reduce its condition number.

    ``dpoequ`` computes a diagonal scaling matrix S intended to
    equilibrate a real n*n symmetric positive definite matrix A and
    reduce its condition number.

    For full information please refer to the NAG Library document for
    f07ff

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fff.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The matrix A whose scaling factors are to be computed. Only the
        diagonal elements of the array `a` are referenced.

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    Notes
    -----
    ``dpoequ`` computes a diagonal scaling matrix S chosen so that

        s_j = 1/sqrt(a_{jj}).

    This means that the matrix B given by

        B = SAS,

    has diagonal elements equal to unity.
    This in turn means that the condition number of B, kappa_2(B), is
    within a factor n of the matrix of smallest possible condition
    number over all possible choices of diagonal scalings (see Corollary
    7.6 of Higham (2002)).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dpoequ'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.dpoequ(
        _n,
        _a,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = DpoequReturnData(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def dpocon(uplo, n, a, anorm):
    u"""
    Estimate condition number of real symmetric positive definite
    matrix, matrix already factorized by ``dpotrf``.

    ``dpocon`` estimates the condition number of a real symmetric
    positive definite matrix A, where A has been factorized by
    ``dpotrf``.

    For full information please refer to the NAG Library document for
    f07fg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The Cholesky factor of A, as returned by ``dpotrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlansy`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dpotrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``dpocon`` estimates the condition number (in the 1-norm) of a real
    symmetric positive definite matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is symmetric, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.dlansy`` to
    compute ||A||_1 and a call to ``dpotrf`` to compute the Cholesky
    factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dpocon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dpocon(
        uplo,
        n,
        _a,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DporfsReturnData = _collections.namedtuple(
    'DporfsReturnData',
    ['x', 'ferr', 'berr'],
)

def dporfs(uplo, n, a, af, b, x):
    u"""
    Refined solution with error bounds of real symmetric positive
    definite system of linear equations, multiple right-hand sides.

    ``dporfs`` returns error bounds for the solution of a real symmetric
    positive definite system of linear equations with multiple
    right-hand sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07fh

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fhf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n*n original symmetric positive definite matrix A as
        supplied to ``dpotrf``.

    af : float, array-like, shape (`n`, `n`)
        The Cholesky factor of A, as returned by ``dpotrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``dpotrs``.

    Returns
    -------
    x : float, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dporfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a real symmetric positive
    definite system of linear equations with multiple right-hand sides
    AX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``dporfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dporfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dporfs(
        uplo,
        n,
        _nrhs,
        _a,
        _af,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DporfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

DpotriReturnData = _collections.namedtuple(
    'DpotriReturnData',
    ['a'],
)

def dpotri(uplo, a):
    u"""
    Inverse of real symmetric positive definite matrix, matrix already
    factorized by ``dpotrf``.

    ``dpotri`` computes the inverse of a real symmetric positive
    definite matrix A, where A has been factorized by ``dpotrf``.

    For full information please refer to the NAG Library document for
    f07fj

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    a : float, array-like, shape (n, n)
        The upper triangular matrix U if `uplo` = 'U' or the lower
        triangular matrix L if `uplo` = 'L', as returned by ``dpotrf``.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        U is overwritten by the upper triangle of A^-1 if `uplo` = 'U';
        L is overwritten by the lower triangle of A^-1 if `uplo` = 'L'.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of the Cholesky factor is zero;
            the Cholesky factor is singular and the inverse of A cannot
            be computed.

    Notes
    -----
    ``dpotri`` is used to compute the inverse of a real symmetric
    positive definite matrix A, the function must be preceded by a call
    to ``dpotrf``, which computes the Cholesky factorization of A.

    If `uplo` = 'U', A = U^TU and A^-1 is computed by first inverting U
    and then forming (U^-1)U^-T.

    If `uplo` = 'L', A = LL^T and A^-1 is computed by first inverting L
    and then forming L^-T(L^-1).

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dpotri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_lapacklin.dpotri(
            uplo,
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = DpotriReturnData(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

ZposvReturnData = _collections.namedtuple(
    'ZposvReturnData',
    ['a', 'b'],
)

def zposv(uplo, a, b):
    u"""
    Computes the solution to a complex Hermitian positive definite
    system of linear equations.

    ``zposv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n Hermitian positive definite matrix and X and B are
    n*r matrices.

    For full information please refer to the NAG Library document for
    f07fn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : complex, array-like, shape (n, n)
        The n*n Hermitian matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^HU or A = LL^H.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    Notes
    -----
    ``zposv`` uses the Cholesky decomposition to factor A as A = U^HU if
    `uplo` = 'U' or A = LL^H if `uplo` = 'L', where U is an upper
    triangular matrix and L is a lower triangular matrix.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zposv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.zposv(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
    )
    _a_py = _a
    _b_py = _b
    _return_data = ZposvReturnData(
        _a_py,
        _b_py,
    )
    return _return_data

ZposvxReturnData = _collections.namedtuple(
    'ZposvxReturnData',
    ['a', 'af', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
)

def zposvx(fact, uplo, n, nrhs, a, af, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a complex Hermitian positive definite
    system of linear equations.

    ``zposvx`` uses the Cholesky factorization

        A = U^HU or A = LL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n*n Hermitian positive definite matrix and X and B are
    n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07fp

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `af` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `a` and `af` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `af` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : complex, array-like, shape (`n`, `n`)
        The n*n Hermitian matrix A.

        If `fact` = 'F' and `equed` = 'Y', `a` must have been
        equilibrated by the scaling factor in `s` as D_S AD_S.

    af : complex, array-like, shape (`n`, `n`)
        If `fact` = 'F', `af` contains the triangular factor U or L from
        the Cholesky factorization A = U^HU or A = LL^H, in the same
        storage format as `a`. If `equed` != 'N', `af` is the factorized
        form of the equilibrated matrix D_S AD_S.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `a` is not modified.

        If `fact` = 'E' and `equed` = 'Y', `a` is overwritten by D_S
        AD_S.

    af : complex, ndarray, shape (`n`, `n`)
        If `fact` = 'N', `af` returns the triangular factor U or L from
        the Cholesky factorization A = U^HU or A = LL^H of the original
        matrix A.

        If `fact` = 'E', `af` returns the triangular factor U or L from
        the Cholesky factorization A = U^HU or A = LL^H of the
        equilibrated matrix A (see the description of `a` for the form
        of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : complex, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X to the original system of equations. Note that
        the arrays A and B are modified on exit if `equed` = 'Y', and
        the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -9)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -10)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``zposvx`` performs the following steps:

    (1) If `fact` = 'E', real diagonal scaling factors, D_S, are
        computed to equilibrate the system:

            (D_S AD_S)(D_S^-1X) = D_S B.

        Whether or not the system will be equilibrated depends on the
        scaling of the matrix A, but if equilibration is used, A is
        overwritten by D_S AD_S and B*D_S B.

    (#) If `fact` = 'N' or 'E', the Cholesky decomposition is used to
        factor the matrix A (after equilibration if `fact` = 'E') as A =
        U^HU if `uplo` = 'U' or A = LL^H if `uplo` = 'L', where U is an
        upper triangular matrix and L is a lower triangular matrix.

    (#) If the leading i*i principal minor of A is not positive
        definite, then the function returns with errno = i. Otherwise,
        the factored form of A is used to estimate the condition number
        of the matrix A. If the reciprocal of the condition number is
        less than machine precision, `errno` = `n` + 1 is returned as a
        warning, but the function still goes on to solve for X and
        compute error bounds as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    (#) If equilibration was used, the matrix X is premultiplied by D_S
        so that it solves the original system before equilibration.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zposvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zposvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _a,
            _af,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _af_py = _af
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZposvxReturnData(
        _a_py,
        _af_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZcposvReturnData = _collections.namedtuple(
    'ZcposvReturnData',
    ['a', 'x', 'itera'],
)

def zcposv(uplo, n, nrhs, a, b):
    u"""
    Computes the solution to a complex Hermitian positive definite
    system of linear equations using mixed precision arithmetic.

    ``zcposv`` uses the Cholesky factorization

        A = U^HU or A = LL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n*n Hermitian positive definite matrix and X and B are
    n*r matrices.

    For full information please refer to the NAG Library document for
    f07fq

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fqf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

            The upper triangular part of A is stored.

        `uplo` = 'L'

            The lower triangular part of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : complex, array-like, shape (`n`, `n`)
        The n*n Hermitian positive definite matrix A.

    b : complex, array-like, shape (`n`, `nrhs`)
        The right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        If iterative refinement has been successfully used (no exception
        or warning is raised and `itera` >= 0, see `itera`), then `a` is
        unchanged. If full precision factorization has been used (no
        exception or warning is raised and `itera` < 0, see `itera`),
        then the array A contains the factor U or L from the Cholesky
        factorization A = U^HU or A = LL^H.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If no exception or warning is raised, the n*r solution matrix X.

    itera : int
        Information on the progress of the interative refinement
        process.

        `itera` < 0

            Iterative refinement has failed for one of the reasons given
            below, full precision factorization has been performed
            instead.

            [table omitted]

        `itera` > 0

            Iterative refinement has been sucessfully used. `itera`
            returns the number of iterations.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``zcposv`` first attempts to factorize the matrix in reduced
    precision and use this factorization within an iterative refinement
    procedure to produce a solution with full precision normwise
    backward error quality (see below).
    If the approach fails the method switches to a full precision
    factorization and solve.

    The iterative refinement can be more efficient than the
    corresponding direct full precision algorithm.
    Since the strategy implemented by ``zcposv`` must perform iterative
    refinement on each right-hand side, any efficiency gains will reduce
    as the number of right-hand sides increases.
    Conversely, as the matrix size increases the cost of these iterative
    refinements become less significant relative to the cost of
    factorization.
    Thus, any efficiency gains will be greatest for a very small number
    of right-hand sides and for large matrix sizes.
    The cut-off values for the number of right-hand sides and matrix
    size, for which the iterative refinement strategy performs better,
    depends on the relative performance of the reduced and full
    precision factorization and back-substitution. ``zcposv`` always
    attempts the iterative refinement strategy first; you are advised to
    compare the performance of ``zcposv`` with that of its full
    precision counterpart ``zposv`` to determine whether this strategy
    is worthwhile for your particular problem dimensions.

    The iterative refinement process is stopped if `itera` > 30 where
    `itera` is the number of iterations carried out thus far.
    The process is also stopped if for all right-hand sides we have

        ||resid|| < sqrt(`n`)||x||||A||epsilon,

    where ||resid|| is the infinity-norm of the residual, ||x|| is the
    infinity-norm of the solution, ||A|| is the infinity-norm of the
    matrix A and epsilon is the machine precision returned by
    ``machine.precision``.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zcposv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _return_scalar = _b_lapacklin.zcposv(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
        _x,
    )
    _a_py = _a
    _x_py = _x
    _itera_py = _return_scalar
    _return_data = ZcposvReturnData(
        _a_py,
        _x_py,
        _itera_py,
    )
    return _return_data

def zpotrf(uplo, a):
    u"""
    Cholesky factorization of complex Hermitian positive definite
    matrix.

    ``zpotrf`` computes the Cholesky factorization of a complex
    Hermitian positive definite matrix.

    For full information please refer to the NAG Library document for
    f07fr

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07frf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        The n*n Hermitian positive definite matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The upper or lower triangle of A is overwritten by the Cholesky
        factor U or L as specified by `uplo`.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. To factorize a Hermitian
            matrix which is not positive definite, call ``zhetrf``
            instead.

    Notes
    -----
    ``zpotrf`` forms the Cholesky factorization of a complex Hermitian
    positive definite matrix A either as A = U^HU if `uplo` = 'U' or A =
    LL^H if `uplo` = 'L', where U is an upper triangular matrix and L is
    lower triangular.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpotrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_lapacklin.zpotrf(
        uplo,
        _n,
        _a,
    )
    _a_py = _a
    return _a_py

def zpotrs(uplo, a, b):
    u"""
    Solution of complex Hermitian positive definite system of linear
    equations, multiple right-hand sides, matrix already factorized by
    ``zpotrf``.

    ``zpotrs`` solves a complex Hermitian positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zpotrf``.

    For full information please refer to the NAG Library document for
    f07fs

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        The Cholesky factor of A, as returned by ``zpotrf``.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zpotrs`` is used to solve a complex Hermitian positive definite
    system of linear equations AX = B, this function must be preceded by
    a call to ``zpotrf`` which computes the Cholesky factorization of A.
    The solution X is computed by forward and backward substitution.

    If `uplo` = 'U', A = U^HU, where U is upper triangular; the solution
    X is computed by solving U^HY = B and then UX = Y.

    If `uplo` = 'L', A = LL^H, where L is lower triangular; the solution
    X is computed by solving LY = B and then L^HX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpotrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.zpotrs(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

ZpoequReturnData = _collections.namedtuple(
    'ZpoequReturnData',
    ['s', 'scond', 'amax'],
)

def zpoequ(n, a):
    u"""
    Computes row and column scalings intended to equilibrate a complex
    Hermitian positive definite matrix and reduce its condition number.

    ``zpoequ`` computes a diagonal scaling matrix S intended to
    equilibrate a complex n*n Hermitian positive definite matrix A and
    reduce its condition number.

    For full information please refer to the NAG Library document for
    f07ft

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07ftf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The matrix A whose scaling factors are to be computed. Only the
        diagonal elements of the array `a` are referenced.

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    Notes
    -----
    ``zpoequ`` computes a diagonal scaling matrix S chosen so that

        s_j = 1/sqrt(a_{jj}).

    This means that the matrix B given by

        B = SAS,

    has diagonal elements equal to unity.
    This in turn means that the condition number of B, kappa_2(B), is
    within a factor n of the matrix of smallest possible condition
    number over all possible choices of diagonal scalings (see Corollary
    7.6 of Higham (2002)).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zpoequ'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.zpoequ(
        _n,
        _a,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = ZpoequReturnData(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def zpocon(uplo, n, a, anorm):
    u"""
    Estimate condition number of complex Hermitian positive definite
    matrix, matrix already factorized by ``zpotrf``.

    ``zpocon`` estimates the condition number of a complex Hermitian
    positive definite matrix A, where A has been factorized by
    ``zpotrf``.

    For full information please refer to the NAG Library document for
    f07fu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fuf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The Cholesky factor of A, as returned by ``zpotrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanhe`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zpotrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``zpocon`` estimates the condition number (in the 1-norm) of a
    complex Hermitian positive definite matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is Hermitian, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.zlanhe`` to
    compute ||A||_1 and a call to ``zpotrf`` to compute the Cholesky
    factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zpocon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zpocon(
        uplo,
        n,
        _a,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZporfsReturnData = _collections.namedtuple(
    'ZporfsReturnData',
    ['x', 'ferr', 'berr'],
)

def zporfs(uplo, n, a, af, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian positive
    definite system of linear equations, multiple right-hand sides.

    ``zporfs`` returns error bounds for the solution of a complex
    Hermitian positive definite system of linear equations with multiple
    right-hand sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07fv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n*n original Hermitian positive definite matrix A as
        supplied to ``zpotrf``.

    af : complex, array-like, shape (`n`, `n`)
        The Cholesky factor of A, as returned by ``zpotrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``zpotrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zporfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex Hermitian positive
    definite system of linear equations with multiple right-hand sides
    AX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``zporfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zporfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zporfs(
        uplo,
        n,
        _nrhs,
        _a,
        _af,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZporfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZpotriReturnData = _collections.namedtuple(
    'ZpotriReturnData',
    ['a'],
)

def zpotri(uplo, a):
    u"""
    Inverse of complex Hermitian positive definite matrix, matrix
    already factorized by ``zpotrf``.

    ``zpotri`` computes the inverse of a complex Hermitian positive
    definite matrix A, where A has been factorized by ``zpotrf``.

    For full information please refer to the NAG Library document for
    f07fw

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07fwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        The upper triangular matrix U if `uplo` = 'U' or the lower
        triangular matrix L if `uplo` = 'L', as returned by ``zpotrf``.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        U is overwritten by the upper triangle of A^-1 if `uplo` = 'U';
        L is overwritten by the lower triangle of A^-1 if `uplo` = 'L'.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of the Cholesky factor is zero;
            the Cholesky factor is singular and the inverse of A cannot
            be computed.

    Notes
    -----
    ``zpotri`` is used to compute the inverse of a complex Hermitian
    positive definite matrix A, the function must be preceded by a call
    to ``zpotrf``, which computes the Cholesky factorization of A.

    If `uplo` = 'U', A = U^HU and A^-1 is computed by first inverting U
    and then forming (U^-1)U^-H.

    If `uplo` = 'L', A = LL^H and A^-1 is computed by first inverting L
    and then forming L^-H(L^-1).

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zpotri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_lapacklin.zpotri(
            uplo,
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = ZpotriReturnData(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

DppsvReturnData = _collections.namedtuple(
    'DppsvReturnData',
    ['ap', 'b'],
)

def dppsv(uplo, ap, b):
    u"""
    Computes the solution to a real symmetric positive definite system
    of linear equations, packed storage.

    ``dppsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric positive definite matrix stored in
    packed format and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07ga

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gaf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    ap : float, array-like, shape (n*(n+1)/2)
        The n*n symmetric matrix A, packed by columns.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    ap : float, ndarray, shape (n*(n+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^TU or A = LL^T, in the same storage
        format as A.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    Notes
    -----
    ``dppsv`` uses the Cholesky decomposition to factor A as A = U^TU if
    `uplo` = 'U' or A = LL^T if `uplo` = 'L', where U is an upper
    triangular matrix and L is a lower triangular matrix.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dppsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dppsv(
        uplo,
        _n,
        _nrhs,
        _ap,
        _b,
    )
    _ap_py = _ap
    _b_py = _b
    _return_data = DppsvReturnData(
        _ap_py,
        _b_py,
    )
    return _return_data

DppsvxReturnData = _collections.namedtuple(
    'DppsvxReturnData',
    ['ap', 'afp', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
)

def dppsvx(fact, uplo, n, nrhs, ap, afp, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a real symmetric positive definite system
    of linear equations, packed storage.

    ``dppsvx`` uses the Cholesky factorization

        A = U^TU or A = LL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric positive definite matrix stored in
    packed format and X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07gb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afp` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `ap` and `afp` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afp` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afp` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        If `fact` = 'F' and `equed` = 'Y', `ap` must contain the
        equilibrated matrix D_S AD_S; otherwise, `ap` must contain the
        n*n symmetric matrix A, packed by columns.

    afp : float, array-like, shape (`n`*(`n`+1)/2)
        If `fact` = 'F', `afp` contains the triangular factor U or L
        from the Cholesky factorization A = U^TU or A = LL^T, in the
        same storage format as `ap`. If `equed` = 'Y', `afp` is the
        factorized form of the equilibrated matrix D_S AD_S.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : float, array-like, shape (`n`, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `ap` is not modified.

        If `fact` = 'E' and `equed` = 'Y', `ap` is overwritten by D_S
        AD_S.

    afp : float, ndarray, shape (`n`*(`n`+1)/2)
        If `fact` = 'N' or if `fact` = 'E' and `equed` = 'N', `afp`
        returns the triangular factor U or L from the Cholesky
        factorization A = U^TU or A = LL^T of the original matrix A.

        If `fact` = 'E' and `equed` = 'Y', `afp` returns the triangular
        factor U or L from the Cholesky factorization A = U^TU or A =
        LL^T of the equilibrated matrix A (see the description of `ap`
        for the form of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : float, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X to the original system of equations. Note that
        the arrays A and B are modified on exit if `equed` = 'Y', and
        the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -7)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -8)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``dppsvx`` performs the following steps:

    (1) If `fact` = 'E', real diagonal scaling factors, D_S, are
        computed to equilibrate the system:

            (D_S AD_S)(D_S^-1X) = D_S B.

        Whether or not the system will be equilibrated depends on the
        scaling of the matrix A, but if equilibration is used, A is
        overwritten by D_S AD_S and B*D_S B.

    (#) If `fact` = 'N' or 'E', the Cholesky decomposition is used to
        factor the matrix A (after equilibration if `fact` = 'E') as A =
        U^TU if `uplo` = 'U' or A = LL^T if `uplo` = 'L', where U is an
        upper triangular matrix and L is a lower triangular matrix.

    (#) If the leading i*i principal minor of A is not positive
        definite, then the function returns with errno = i. Otherwise,
        the factored form of A is used to estimate the condition number
        of the matrix A. If the reciprocal of the condition number is
        less than machine precision, `errno` = `n` + 1 is returned as a
        warning, but the function still goes on to solve for X and
        compute error bounds as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    (#) If equilibration was used, the matrix X is premultiplied by D_S
        so that it solves the original system before equilibration.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dppsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dppsvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _ap,
            _afp,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ap_py = _ap
    _afp_py = _afp
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DppsvxReturnData(
        _ap_py,
        _afp_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dpptrf(uplo, n, ap):
    u"""
    Cholesky factorization of real symmetric positive definite matrix,
    packed storage.

    ``dpptrf`` computes the Cholesky factorization of a real symmetric
    positive definite matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07gd

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n*n symmetric matrix A, packed by columns.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^TU or A = LL^T, in the same storage
        format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. To factorize a symmetric
            matrix which is not positive definite, call ``dsptrf``
            instead.

    Notes
    -----
    ``dpptrf`` forms the Cholesky factorization of a real symmetric
    positive definite matrix A either as A = U^TU if `uplo` = 'U' or A =
    LL^T if `uplo` = 'L', where U is an upper triangular matrix and L is
    lower triangular, using packed storage.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpptrf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dpptrf(
        uplo,
        n,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def dpptrs(uplo, n, ap, b):
    u"""
    Solution of real symmetric positive definite system of linear
    equations, multiple right-hand sides, matrix already factorized by
    ``dpptrf``, packed storage.

    ``dpptrs`` solves a real symmetric positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``dpptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07ge

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``dpptrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (`n`, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dpptrs`` is used to solve a real symmetric positive definite
    system of linear equations AX = B, the function must be preceded by
    a call to ``dpptrf`` which computes the Cholesky factorization of A,
    using packed storage.
    The solution X is computed by forward and backward substitution.

    If `uplo` = 'U', A = U^TU, where U is upper triangular; the solution
    X is computed by solving U^TY = B and then UX = Y.

    If `uplo` = 'L', A = LL^T, where L is lower triangular; the solution
    X is computed by solving LY = B and then L^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dpptrs(
        uplo,
        n,
        _nrhs,
        _ap,
        _b,
    )
    _b_py = _b
    return _b_py

DppequReturnData = _collections.namedtuple(
    'DppequReturnData',
    ['s', 'scond', 'amax'],
)

def dppequ(uplo, n, ap):
    u"""
    Computes row and column scalings intended to equilibrate a real
    symmetric positive definite matrix and reduce its condition number,
    packed storage.

    ``dppequ`` computes a diagonal scaling matrix S intended to
    equilibrate a real n*n symmetric positive definite matrix A, stored
    in packed format, and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07gf

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gff.html

    Parameters
    ----------
    uplo : str, length 1
        Indicates whether the upper or lower triangular part of A is
        stored in the array `ap`, as follows:

        `uplo` = 'U'

            The upper triangle of A is stored.

        `uplo` = 'L'

            The lower triangle of A is stored.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n*n symmetric matrix A, packed by columns.

        Only the elements of `ap` corresponding to the diagonal elements
        A are referenced.

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    Notes
    -----
    ``dppequ`` computes a diagonal scaling matrix S chosen so that

        s_j = 1/sqrt(a_{jj}).

    This means that the matrix B given by

        B = SAS,

    has diagonal elements equal to unity.
    This in turn means that the condition number of B, kappa_2(B), is
    within a factor n of the matrix of smallest possible condition
    number over all possible choices of diagonal scalings (see Corollary
    7.6 of Higham (2002)).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dppequ'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.dppequ(
        uplo,
        _n,
        _ap,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = DppequReturnData(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def dppcon(uplo, n, ap, anorm):
    u"""
    Estimate condition number of real symmetric positive definite
    matrix, matrix already factorized by ``dpptrf``, packed storage.

    ``dppcon`` estimates the condition number of a real symmetric
    positive definite matrix A, where A has been factorized by
    ``dpptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07gg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07ggf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``dpptrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlansp`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dpptrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``dppcon`` estimates the condition number (in the 1-norm) of a real
    symmetric positive definite matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is symmetric, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.dlansp`` to
    compute ||A||_1 and a call to ``dpptrf`` to compute the Cholesky
    factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dppcon'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dppcon(
        uplo,
        n,
        _ap,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DpprfsReturnData = _collections.namedtuple(
    'DpprfsReturnData',
    ['x', 'ferr', 'berr'],
)

def dpprfs(uplo, n, ap, afp, b, x):
    u"""
    Refined solution with error bounds of real symmetric positive
    definite system of linear equations, multiple right-hand sides,
    packed storage.

    ``dpprfs`` returns error bounds for the solution of a real symmetric
    positive definite system of linear equations with multiple
    right-hand sides, AX = B, using packed storage.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07gh

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07ghf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n*n original symmetric positive definite matrix A as
        supplied to ``dpptrf``.

    afp : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``dpptrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``dpptrs``.

    Returns
    -------
    x : float, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dpprfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a real symmetric positive
    definite system of linear equations with multiple right-hand sides
    AX = B, using packed storage.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``dpprfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dpprfs(
        uplo,
        n,
        _nrhs,
        _ap,
        _afp,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DpprfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

DpptriReturnData = _collections.namedtuple(
    'DpptriReturnData',
    ['ap'],
)

def dpptri(uplo, n, ap):
    u"""
    Inverse of real symmetric positive definite matrix, matrix already
    factorized by ``dpptrf``, packed storage.

    ``dpptri`` computes the inverse of a real symmetric positive
    definite matrix A, where A has been factorized by ``dpptrf``, using
    packed storage.

    For full information please refer to the NAG Library document for
    f07gj

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``dpptrf``.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n*n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of the Cholesky factor is zero;
            the Cholesky factor is singular and the inverse of A cannot
            be computed.

    Notes
    -----
    ``dpptri`` is used to compute the inverse of a real symmetric
    positive definite matrix A, the function must be preceded by a call
    to ``dpptrf``, which computes the Cholesky factorization of A, using
    packed storage.

    If `uplo` = 'U', A = U^TU and A^-1 is computed by first inverting U
    and then forming (U^-1)U^-T.

    If `uplo` = 'L', A = LL^T and A^-1 is computed by first inverting L
    and then forming L^-T(L^-1).

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dpptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dpptri(
            uplo,
            n,
            _ap,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = DpptriReturnData(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

ZppsvReturnData = _collections.namedtuple(
    'ZppsvReturnData',
    ['ap', 'b'],
)

def zppsv(uplo, ap, b):
    u"""
    Computes the solution to a complex Hermitian positive definite
    system of linear equations, packed storage.

    ``zppsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n Hermitian positive definite matrix stored in
    packed format and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07gn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n*n Hermitian matrix A, packed by columns.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    ap : complex, ndarray, shape (n*(n+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^HU or A = LL^H, in the same storage
        format as A.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    Notes
    -----
    ``zppsv`` uses the Cholesky decomposition to factor A as A = U^HU if
    `uplo` = 'U' or A = LL^H if `uplo` = 'L', where U is an upper
    triangular matrix and L is a lower triangular matrix.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zppsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _b_lapacklin.zppsv(
        uplo,
        _n,
        _nrhs,
        _ap,
        _b,
    )
    _ap_py = _ap
    _b_py = _b
    _return_data = ZppsvReturnData(
        _ap_py,
        _b_py,
    )
    return _return_data

ZppsvxReturnData = _collections.namedtuple(
    'ZppsvxReturnData',
    ['ap', 'afp', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
)

def zppsvx(fact, uplo, n, nrhs, ap, afp, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a complex Hermitian positive definite
    system of linear equations, packed storage.

    ``zppsvx`` uses the Cholesky factorization

        A = U^HU or A = LL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n*n Hermitian positive definite matrix stored in
    packed format and X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07gp

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afp` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `ap` and `afp` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afp` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afp` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        If `fact` = 'F' and `equed` = 'Y', `ap` must contain the
        equilibrated matrix D_S AD_S; otherwise, `ap` must contain the
        n*n Hermitian matrix A, packed by columns.

    afp : complex, array-like, shape (`n`*(`n`+1)/2)
        If `fact` = 'F', `afp` contains the triangular factor U or L
        from the Cholesky factorization A = U^HU or A = LL^H, in the
        same storage format as `ap`. If `equed` = 'Y', `afp` is the
        factorized form of the equilibrated matrix D_S AD_S.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `ap` is not modified.

        If `fact` = 'E' and `equed` = 'Y', `ap` is overwritten by D_S
        AD_S.

    afp : complex, ndarray, shape (`n`*(`n`+1)/2)
        If `fact` = 'N' or if `fact` = 'E' and `equed` = 'N', `afp`
        returns the triangular factor U or L from the Cholesky
        factorization A = U^HU or A = LL^H of the original matrix A.

        If `fact` = 'E' and `equed` = 'Y', `afp` returns the triangular
        factor U or L from the Cholesky factorization A = U^HU or A =
        LL^H of the equilibrated matrix A (see the description of `ap`
        for the form of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : complex, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X to the original system of equations. Note that
        the arrays A and B are modified on exit if `equed` = 'Y', and
        the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -7)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -8)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``zppsvx`` performs the following steps:

    (1) If `fact` = 'E', real diagonal scaling factors, D_S, are
        computed to equilibrate the system:

            (D_S AD_S)(D_S^-1X) = D_S B.

        Whether or not the system will be equilibrated depends on the
        scaling of the matrix A, but if equilibration is used, A is
        overwritten by D_S AD_S and B*D_S B.

    (#) If `fact` = 'N' or 'E', the Cholesky decomposition is used to
        factor the matrix A (after equilibration if `fact` = 'E') as A =
        U^HU if `uplo` = 'U' or A = LL^H if `uplo` = 'L', where U is an
        upper triangular matrix and L is a lower triangular matrix.

    (#) If the leading i*i principal minor of A is not positive
        definite, then the function returns with errno = i. Otherwise,
        the factored form of A is used to estimate the condition number
        of the matrix A. If the reciprocal of the condition number is
        less than machine precision, `errno` = `n` + 1 is returned as a
        warning, but the function still goes on to solve for X and
        compute error bounds as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    (#) If equilibration was used, the matrix X is premultiplied by D_S
        so that it solves the original system before equilibration.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zppsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=True,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zppsvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _ap,
            _afp,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ap_py = _ap
    _afp_py = _afp
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZppsvxReturnData(
        _ap_py,
        _afp_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zpptrf(uplo, n, ap):
    u"""
    Cholesky factorization of complex Hermitian positive definite
    matrix, packed storage.

    ``zpptrf`` computes the Cholesky factorization of a complex
    Hermitian positive definite matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07gr

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07grf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n*n Hermitian matrix A, packed by columns.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^HU or A = LL^H, in the same storage
        format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. To factorize a Hermitian
            matrix which is not positive definite, call ``zhptrf``
            instead.

    Notes
    -----
    ``zpptrf`` forms the Cholesky factorization of a complex Hermitian
    positive definite matrix A either as A = U^HU if `uplo` = 'U' or A =
    LL^H if `uplo` = 'L', where U is an upper triangular matrix and L is
    lower triangular, using packed storage.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpptrf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _b_lapacklin.zpptrf(
        uplo,
        n,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def zpptrs(uplo, n, ap, b):
    u"""
    Solution of complex Hermitian positive definite system of linear
    equations, multiple right-hand sides, matrix already factorized by
    ``zpptrf``, packed storage.

    ``zpptrs`` solves a complex Hermitian positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zpptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07gs

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``zpptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (`n`, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zpptrs`` is used to solve a complex Hermitian positive definite
    system of linear equations AX = B, the function must be preceded by
    a call to ``zpptrf`` which computes the Cholesky factorization of A,
    using packed storage.
    The solution X is computed by forward and backward substitution.

    If `uplo` = 'U', A = U^HU, where U is upper triangular; the solution
    X is computed by solving U^HY = B and then UX = Y.

    If `uplo` = 'L', A = LL^H, where L is lower triangular; the solution
    X is computed by solving LY = B and then L^HX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _b_lapacklin.zpptrs(
        uplo,
        n,
        _nrhs,
        _ap,
        _b,
    )
    _b_py = _b
    return _b_py

ZppequReturnData = _collections.namedtuple(
    'ZppequReturnData',
    ['s', 'scond', 'amax'],
)

def zppequ(uplo, n, ap):
    u"""
    Computes row and column scalings intended to equilibrate a complex
    Hermitian positive definite matrix and reduce its condition number,
    packed storage.

    ``zppequ`` computes a diagonal scaling matrix S intended to
    equilibrate a complex n*n Hermitian positive definite matrix A,
    stored in packed format, and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07gt

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gtf.html

    Parameters
    ----------
    uplo : str, length 1
        Indicates whether the upper or lower triangular part of A is
        stored in the array `ap`, as follows:

        `uplo` = 'U'

            The upper triangle of A is stored.

        `uplo` = 'L'

            The lower triangle of A is stored.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n*n Hermitian matrix A, packed by columns.

        Only the elements of `ap` corresponding to the diagonal elements
        A are referenced.

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    Notes
    -----
    ``zppequ`` computes a diagonal scaling matrix S chosen so that

        s_j = 1/sqrt(a_{jj}).

    This means that the matrix B given by

        B = SAS,

    has diagonal elements equal to unity.
    This in turn means that the condition number of B, kappa_2(B), is
    within a factor n of the matrix of smallest possible condition
    number over all possible choices of diagonal scalings (see Corollary
    7.6 of Higham (2002)).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zppequ'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.zppequ(
        uplo,
        _n,
        _ap,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = ZppequReturnData(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def zppcon(uplo, n, ap, anorm):
    u"""
    Estimate condition number of complex Hermitian positive definite
    matrix, matrix already factorized by ``zpptrf``, packed storage.

    ``zppcon`` estimates the condition number of a complex Hermitian
    positive definite matrix A, where A has been factorized by
    ``zpptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07gu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07guf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``zpptrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanhp`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zpptrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``zppcon`` estimates the condition number (in the 1-norm) of a
    complex Hermitian positive definite matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is Hermitian, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.zlanhp`` to
    compute ||A||_1 and a call to ``zpptrf`` to compute the Cholesky
    factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zppcon'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zppcon(
        uplo,
        n,
        _ap,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZpprfsReturnData = _collections.namedtuple(
    'ZpprfsReturnData',
    ['x', 'ferr', 'berr'],
)

def zpprfs(uplo, n, ap, afp, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian positive
    definite system of linear equations, multiple right-hand sides,
    packed storage.

    ``zpprfs`` returns error bounds for the solution of a complex
    Hermitian positive definite system of linear equations with multiple
    right-hand sides, AX = B, using packed storage.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07gv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n*n original Hermitian positive definite matrix A as
        supplied to ``zpptrf``.

    afp : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``zpptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``zpptrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zpprfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex Hermitian positive
    definite system of linear equations with multiple right-hand sides
    AX = B, using packed storage.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``zpprfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zpprfs(
        uplo,
        n,
        _nrhs,
        _ap,
        _afp,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZpprfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZpptriReturnData = _collections.namedtuple(
    'ZpptriReturnData',
    ['ap'],
)

def zpptri(uplo, n, ap):
    u"""
    Inverse of complex Hermitian positive definite matrix, matrix
    already factorized by ``zpptrf``, packed storage.

    ``zpptri`` computes the inverse of a complex Hermitian positive
    definite matrix A, where A has been factorized by ``zpptrf``, using
    packed storage.

    For full information please refer to the NAG Library document for
    f07gw

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07gwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``zpptrf``.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n*n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of the Cholesky factor is zero;
            the Cholesky factor is singular and the inverse of A cannot
            be computed.

    Notes
    -----
    ``zpptri`` is used to compute the inverse of a complex Hermitian
    positive definite matrix A, the function must be preceded by a call
    to ``zpptrf``, which computes the Cholesky factorization of A, using
    packed storage.

    If `uplo` = 'U', A = U^HU and A^-1 is computed by first inverting U
    and then forming (U^-1)U^-H.

    If `uplo` = 'L', A = LL^H and A^-1 is computed by first inverting L
    and then forming L^-H(L^-1).

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zpptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    try:
        _b_lapacklin.zpptri(
            uplo,
            n,
            _ap,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = ZpptriReturnData(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

DpbsvReturnData = _collections.namedtuple(
    'DpbsvReturnData',
    ['ab', 'b'],
)

def dpbsv(uplo, kd, ab, b):
    u"""
    Computes the solution to a real symmetric positive definite banded
    system of linear equations.

    ``dpbsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric positive definite band matrix of
    bandwidth (2k_d+1) and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07ha

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07haf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : float, array-like, shape (`kd`+1, n)
        The upper or lower triangle of the symmetric band matrix A.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    ab : float, ndarray, shape (`kd`+1, n)
        If no exception or warning is raised, the triangular factor U or
        L from the Cholesky factorization A = U^TU or A = LL^T of the
        band matrix A, in the same storage format as A.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    Notes
    -----
    ``dpbsv`` uses the Cholesky decomposition to factor A as A = U^TU if
    `uplo` = 'U' or A = LL^T if `uplo` = 'L', where U is an upper
    triangular band matrix, and L is a lower triangular band matrix,
    with the same number of superdiagonals or subdiagonals as A.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbsv'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.dpbsv(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
    )
    _ab_py = _ab
    _b_py = _b
    _return_data = DpbsvReturnData(
        _ab_py,
        _b_py,
    )
    return _return_data

DpbsvxReturnData = _collections.namedtuple(
    'DpbsvxReturnData',
    ['ab', 'afb', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
)

def dpbsvx(fact, uplo, n, kd, nrhs, ab, afb, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a real symmetric positive definite banded
    system of linear equations.

    ``dpbsvx`` uses the Cholesky factorization

        A = U^TU or A = LL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric positive definite band matrix of
    bandwidth (2k_d+1) and X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07hb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07hbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afb` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `ab` and `afb` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afb` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afb` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ab : float, array-like, shape (`kd`+1, `n`)
        The upper or lower triangle of the symmetric band matrix A,
        except if `fact` = 'F' and `equed` = 'Y', in which case `ab`
        must contain the equilibrated matrix D_S AD_S.

    afb : float, array-like, shape (`kd`+1, `n`)
        If `fact` = 'F', `afb` contains the triangular factor U or L
        from the Cholesky factorization A = U^TU or A = LL^T of the band
        matrix A, in the same storage format as A. If `equed` = 'Y',
        `afb` is the factorized form of the equilibrated matrix A.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : float, array-like, shape (`n`, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    ab : float, ndarray, shape (`kd`+1, `n`)
        If `fact` = 'E' and `equed` = 'Y', `ab` is overwritten by D_S
        AD_S.

    afb : float, ndarray, shape (`kd`+1, `n`)
        If `fact` = 'N', `afb` returns the triangular factor U or L from
        the Cholesky factorization A = U^TU or A = LL^T.

        If `fact` = 'E', `afb` returns the triangular factor U or L from
        the Cholesky factorization A = U^TU or A = LL^T of the
        equilibrated matrix A (see the description of `ab` for the form
        of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : float, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X to the original system of equations. Note that
        the arrays A and B are modified on exit if `equed` = 'Y', and
        the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -5)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -10)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -11)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``dpbsvx`` performs the following steps:

    (1) If `fact` = 'E', real diagonal scaling factors, D_S, are
        computed to equilibrate the system:

            (D_S AD_S)(D_S^-1X) = D_S B.

        Whether or not the system will be equilibrated depends on the
        scaling of the matrix A, but if equilibration is used, A is
        overwritten by D_S AD_S and B*D_S B.

    (#) If `fact` = 'N' or 'E', the Cholesky decomposition is used to
        factor the matrix A (after equilibration if `fact` = 'E') as A =
        U^TU if `uplo` = 'U' or A = LL^T if `uplo` = 'L', where U is an
        upper triangular matrix and L is a lower triangular matrix.

    (#) If the leading i*i principal minor of A is not positive
        definite, then the function returns with errno = i. Otherwise,
        the factored form of A is used to estimate the condition number
        of the matrix A. If the reciprocal of the condition number is
        less than machine precision, `errno` = `n` + 1 is returned as a
        warning, but the function still goes on to solve for X and
        compute error bounds as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    (#) If equilibration was used, the matrix X is premultiplied by D_S
        so that it solves the original system before equilibration.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbsvx'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_ab, 'ab'), (_afb, 'afb'), (_b, 'b')),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dpbsvx(
            fact,
            uplo,
            _n,
            kd,
            _nrhs,
            _ab,
            _afb,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ab_py = _ab
    _afb_py = _afb
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DpbsvxReturnData(
        _ab_py,
        _afb_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dpbtrf(uplo, kd, ab):
    u"""
    Cholesky factorization of real symmetric positive definite band
    matrix.

    ``dpbtrf`` computes the Cholesky factorization of a real symmetric
    positive definite band matrix.

    For full information please refer to the NAG Library document for
    f07hd

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07hdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : float, array-like, shape (`kd`+1, n)
        The n*n symmetric positive definite band matrix A.

    Returns
    -------
    ab : float, ndarray, shape (`kd`+1, n)
        The upper or lower triangle of A is overwritten by the Cholesky
        factor U or L as specified by `uplo`, using the same storage
        format as described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to factorize a symmetric band matrix
            which is not positive definite; the matrix must be treated
            either as a nonsymmetric band matrix, by calling ``dgbtrf``
            or as a full symmetric matrix, by calling ``dsytrf``.

    Notes
    -----
    ``dpbtrf`` forms the Cholesky factorization of a real symmetric
    positive definite band matrix A either as A = U^TU if `uplo` = 'U'
    or A = LL^T if `uplo` = 'L', where U (or L) is an upper (or lower)
    triangular band matrix with the same number of superdiagonals (or
    subdiagonals) as A.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbtrf'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_lapacklin.dpbtrf(
        uplo,
        _n,
        kd,
        _ab,
    )
    _ab_py = _ab
    return _ab_py

def dpbtrs(uplo, kd, ab, b):
    u"""
    Solution of real symmetric positive definite band system of linear
    equations, multiple right-hand sides, matrix already factorized by
    ``dpbtrf``.

    ``dpbtrs`` solves a real symmetric positive definite band system of
    linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``dpbtrf``.

    For full information please refer to the NAG Library document for
    f07he

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07hef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : float, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``dpbtrf``.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dpbtrs`` is used to solve a real symmetric positive definite band
    system of linear equations AX = B, the function must be preceded by
    a call to ``dpbtrf`` which computes the Cholesky factorization of A.
    The solution X is computed by forward and backward substitution.

    If `uplo` = 'U', A = U^TU, where U is upper triangular; the solution
    X is computed by solving U^TY = B and then UX = Y.

    If `uplo` = 'L', A = LL^T, where L is lower triangular; the solution
    X is computed by solving LY = B and then L^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.dpbtrs(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
    )
    _b_py = _b
    return _b_py

DpbequReturnData = _collections.namedtuple(
    'DpbequReturnData',
    ['s', 'scond', 'amax'],
)

def dpbequ(uplo, n, kd, ab):
    u"""
    Computes row and column scalings intended to equilibrate a real
    symmetric positive definite banded matrix and reduce its condition
    number.

    ``dpbequ`` computes a diagonal scaling matrix S intended to
    equilibrate a real n*n symmetric positive definite band matrix A,
    with bandwidth (2k_d+1), and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07hf

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07hff.html

    Parameters
    ----------
    uplo : str, length 1
        Indicates whether the upper or lower triangular part of A is
        stored in the array `ab`, as follows:

        `uplo` = 'U'

            The upper triangle of A is stored.

        `uplo` = 'L'

            The lower triangle of A is stored.

    n : int
        n, the order of the matrix A.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : float, array-like, shape (`kd`+1, `n`)
        The upper or lower triangle of the symmetric positive definite
        band matrix A whose scaling factors are to be computed.

        Only the elements of the array `ab` corresponding to the
        diagonal elements of A are referenced. (Row (k_d+1) of `ab` when
        `uplo` = 'U', row 1 of `ab` when `uplo` = 'L'.)

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    Notes
    -----
    ``dpbequ`` computes a diagonal scaling matrix S chosen so that

        s_j = 1/sqrt(a_{jj}).

    This means that the matrix B given by

        B = SAS,

    has diagonal elements equal to unity.
    This in turn means that the condition number of B, kappa_2(B), is
    within a factor n of the matrix of smallest possible condition
    number over all possible choices of diagonal scalings (see Corollary
    7.6 of Higham (2002)).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbequ'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.dpbequ(
        uplo,
        _n,
        kd,
        _ab,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = DpbequReturnData(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def dpbcon(uplo, kd, ab, anorm):
    u"""
    Estimate condition number of real symmetric positive definite band
    matrix, matrix already factorized by ``dpbtrf``.

    ``dpbcon`` estimates the condition number of a real symmetric
    positive definite band matrix A, where A has been factorized by
    ``dpbtrf``.

    For full information please refer to the NAG Library document for
    f07hg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07hgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : float, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``dpbtrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlansb`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dpbtrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``dpbcon`` estimates the condition number (in the 1-norm) of a real
    symmetric positive definite band matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is symmetric, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.dlansb`` to
    compute ||A||_1 and a call to ``dpbtrf`` to compute the Cholesky
    factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_lapacklin.dpbcon(
        uplo,
        _n,
        kd,
        _ab,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DpbrfsReturnData = _collections.namedtuple(
    'DpbrfsReturnData',
    ['x', 'ferr', 'berr'],
)

def dpbrfs(uplo, kd, nrhs, ab, afb, b, x):
    u"""
    Refined solution with error bounds of real symmetric positive
    definite band system of linear equations, multiple right-hand sides.

    ``dpbrfs`` returns error bounds for the solution of a real symmetric
    positive definite band system of linear equations with multiple
    right-hand sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07hh

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07hhf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides.

    ab : float, array-like, shape (`kd`+1, n)
        The n*n original symmetric positive definite band matrix A as
        supplied to ``dpbtrf``.

    afb : float, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``dpbtrf``.

    b : float, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    x : float, array-like, shape (n, `nrhs`)
        The n*r solution matrix X, as returned by ``dpbtrs``.

    Returns
    -------
    x : float, ndarray, shape (n, `nrhs`)
        The improved solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Notes
    -----
    ``dpbrfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a real symmetric positive
    definite band system of linear equations with multiple right-hand
    sides AX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``dpbrfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dpbrfs(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _afb,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DpbrfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZpbsvReturnData = _collections.namedtuple(
    'ZpbsvReturnData',
    ['ab', 'b'],
)

def zpbsv(uplo, kd, ab, b):
    u"""
    Computes the solution to a complex Hermitian positive definite
    banded system of linear equations.

    ``zpbsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n Hermitian positive definite band matrix of
    bandwidth (2k_d+1) and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07hn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07hnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : complex, array-like, shape (`kd`+1, n)
        The upper or lower triangle of the Hermitian band matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    ab : complex, ndarray, shape (`kd`+1, n)
        If no exception or warning is raised, the triangular factor U or
        L from the Cholesky factorization A = U^HU or A = LL^H of the
        band matrix A, in the same storage format as A.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    Notes
    -----
    ``zpbsv`` uses the Cholesky decomposition to factor A as A = U^HU if
    `uplo` = 'U' or A = LL^H if `uplo` = 'L', where U is an upper
    triangular band matrix, and L is a lower triangular band matrix,
    with the same number of superdiagonals or subdiagonals as A.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbsv'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.zpbsv(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
    )
    _ab_py = _ab
    _b_py = _b
    _return_data = ZpbsvReturnData(
        _ab_py,
        _b_py,
    )
    return _return_data

ZpbsvxReturnData = _collections.namedtuple(
    'ZpbsvxReturnData',
    ['ab', 'afb', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
)

def zpbsvx(fact, uplo, n, kd, nrhs, ab, afb, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a complex Hermitian positive definite
    banded system of linear equations.

    ``zpbsvx`` uses the Cholesky factorization

        A = U^HU or A = LL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n*n Hermitian positive definite band matrix of
    bandwidth (2k_d+1) and X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07hp

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07hpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afb` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `ab` and `afb` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afb` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afb` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ab : complex, array-like, shape (`kd`+1, `n`)
        The upper or lower triangle of the Hermitian band matrix A,
        except if `fact` = 'F' and `equed` = 'Y', in which case `ab`
        must contain the equilibrated matrix D_S AD_S.

    afb : complex, array-like, shape (`kd`+1, `n`)
        If `fact` = 'F', `afb` contains the triangular factor U or L
        from the Cholesky factorization A = U^HU or A = LL^H of the band
        matrix A, in the same storage format as A. If `equed` = 'Y',
        `afb` is the factorized form of the equilibrated matrix A.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    ab : complex, ndarray, shape (`kd`+1, `n`)
        If `fact` = 'E' and `equed` = 'Y', `ab` is overwritten by D_S
        AD_S.

    afb : complex, ndarray, shape (`kd`+1, `n`)
        If `fact` = 'N', `afb` returns the triangular factor U or L from
        the Cholesky factorization A = U^HU or A = LL^H.

        If `fact` = 'E', `afb` returns the triangular factor U or L from
        the Cholesky factorization A = U^HU or A = LL^H of the
        equilibrated matrix A (see the description of `ab` for the form
        of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : complex, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X to the original system of equations. Note that
        the arrays A and B are modified on exit if `equed` = 'Y', and
        the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -5)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -10)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -11)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``zpbsvx`` performs the following steps:

    (1) If `fact` = 'E', real diagonal scaling factors, D_S, are
        computed to equilibrate the system:

            (D_S AD_S)(D_S^-1X) = D_S B.

        Whether or not the system will be equilibrated depends on the
        scaling of the matrix A, but if equilibration is used, A is
        overwritten by D_S AD_S and B*D_S B.

    (#) If `fact` = 'N' or 'E', the Cholesky decomposition is used to
        factor the matrix A (after equilibration if `fact` = 'E') as A =
        U^HU if `uplo` = 'U' or A = LL^H if `uplo` = 'L', where U is an
        upper triangular matrix and L is a lower triangular matrix.

    (#) If the leading i*i principal minor of A is not positive
        definite, then the function returns with errno = i. Otherwise,
        the factored form of A is used to estimate the condition number
        of the matrix A. If the reciprocal of the condition number is
        less than machine precision, `errno` = `n` + 1 is returned as a
        warning, but the function still goes on to solve for X and
        compute error bounds as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    (#) If equilibration was used, the matrix X is premultiplied by D_S
        so that it solves the original system before equilibration.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbsvx'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_ab, 'ab'), (_afb, 'afb'), (_b, 'b')),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zpbsvx(
            fact,
            uplo,
            _n,
            kd,
            _nrhs,
            _ab,
            _afb,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ab_py = _ab
    _afb_py = _afb
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZpbsvxReturnData(
        _ab_py,
        _afb_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zpbtrf(uplo, kd, ab):
    u"""
    Cholesky factorization of complex Hermitian positive definite band
    matrix.

    ``zpbtrf`` computes the Cholesky factorization of a complex
    Hermitian positive definite band matrix.

    For full information please refer to the NAG Library document for
    f07hr

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07hrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : complex, array-like, shape (`kd`+1, n)
        The n*n Hermitian positive definite band matrix A.

    Returns
    -------
    ab : complex, ndarray, shape (`kd`+1, n)
        The upper or lower triangle of A is overwritten by the Cholesky
        factor U or L as specified by `uplo`, using the same storage
        format as described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to factorize a Hermitian band matrix
            which is not positive definite; the matrix must be treated
            either as a nonsymmetric band matrix, by calling ``zgbtrf``
            or as a full Hermitian matrix, by calling ``zhetrf``.

    Notes
    -----
    ``zpbtrf`` forms the Cholesky factorization of a complex Hermitian
    positive definite band matrix A either as A = U^HU if `uplo` = 'U'
    or A = LL^H if `uplo` = 'L', where U (or L) is an upper (or lower)
    triangular band matrix with the same number of superdiagonals (or
    subdiagonals) as A.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbtrf'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_lapacklin.zpbtrf(
        uplo,
        _n,
        kd,
        _ab,
    )
    _ab_py = _ab
    return _ab_py

def zpbtrs(uplo, kd, ab, b):
    u"""
    Solution of complex Hermitian positive definite band system of
    linear equations, multiple right-hand sides, matrix already
    factorized by ``zpbtrf``.

    ``zpbtrs`` solves a complex Hermitian positive definite band system
    of linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zpbtrf``.

    For full information please refer to the NAG Library document for
    f07hs

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07hsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : complex, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``zpbtrf``.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zpbtrs`` is used to solve a complex Hermitian positive definite
    band system of linear equations AX = B, the function must be
    preceded by a call to ``zpbtrf`` which computes the Cholesky
    factorization of A.
    The solution X is computed by forward and backward substitution.

    If `uplo` = 'U', A = U^HU, where U is upper triangular; the solution
    X is computed by solving U^HY = B and then UX = Y.

    If `uplo` = 'L', A = LL^H, where L is lower triangular; the solution
    X is computed by solving LY = B and then L^HX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.zpbtrs(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
    )
    _b_py = _b
    return _b_py

ZpbequReturnData = _collections.namedtuple(
    'ZpbequReturnData',
    ['s', 'scond', 'amax'],
)

def zpbequ(uplo, n, kd, ab):
    u"""
    Computes row and column scalings intended to equilibrate a complex
    Hermitian positive definite banded matrix and reduce its condition
    number.

    ``zpbequ`` computes a diagonal scaling matrix S intended to
    equilibrate a complex n*n Hermitian positive definite band matrix A,
    with bandwidth (2k_d+1), and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07ht

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07htf.html

    Parameters
    ----------
    uplo : str, length 1
        Indicates whether the upper or lower triangular part of A is
        stored in the array `ab`, as follows:

        `uplo` = 'U'

            The upper triangle of A is stored.

        `uplo` = 'L'

            The lower triangle of A is stored.

    n : int
        n, the order of the matrix A.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : complex, array-like, shape (`kd`+1, `n`)
        The upper or lower triangle of the Hermitian positive definite
        band matrix A whose scaling factors are to be computed.

        Only the elements of the array `ab` corresponding to the
        diagonal elements of A are referenced. (Row (k_d+1) of `ab` when
        `uplo` = 'U', row 1 of `ab` when `uplo` = 'L'.)

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    Notes
    -----
    ``zpbequ`` computes a diagonal scaling matrix S chosen so that

        s_j = 1/sqrt(a_{jj}).

    This means that the matrix B given by

        B = SAS,

    has diagonal elements equal to unity.
    This in turn means that the condition number of B, kappa_2(B), is
    within a factor n of the matrix of smallest possible condition
    number over all possible choices of diagonal scalings (see Corollary
    7.6 of Higham (2002)).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbequ'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.zpbequ(
        uplo,
        _n,
        kd,
        _ab,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = ZpbequReturnData(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def zpbcon(uplo, kd, ab, anorm):
    u"""
    Estimate condition number of complex Hermitian positive definite
    band matrix, matrix already factorized by ``zpbtrf``.

    ``zpbcon`` estimates the condition number of a complex Hermitian
    positive definite band matrix A, where A has been factorized by
    ``zpbtrf``.

    For full information please refer to the NAG Library document for
    f07hu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07huf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : complex, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``zpbtrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanhb`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zpbtrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``zpbcon`` estimates the condition number (in the 1-norm) of a
    complex Hermitian positive definite band matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is Hermitian, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.zlanhb`` to
    compute ||A||_1 and a call to ``zpbtrf`` to compute the Cholesky
    factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_lapacklin.zpbcon(
        uplo,
        _n,
        kd,
        _ab,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZpbrfsReturnData = _collections.namedtuple(
    'ZpbrfsReturnData',
    ['x', 'ferr', 'berr'],
)

def zpbrfs(uplo, kd, nrhs, ab, afb, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian positive
    definite band system of linear equations, multiple right-hand sides.

    ``zpbrfs`` returns error bounds for the solution of a complex
    Hermitian positive definite band system of linear equations with
    multiple right-hand sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07hv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07hvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides.

    ab : complex, array-like, shape (`kd`+1, n)
        The n*n original Hermitian positive definite band matrix A as
        supplied to ``zpbtrf``.

    afb : complex, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``zpbtrf``.

    b : complex, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (n, `nrhs`)
        The n*r solution matrix X, as returned by ``zpbtrs``.

    Returns
    -------
    x : complex, ndarray, shape (n, `nrhs`)
        The improved solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Notes
    -----
    ``zpbrfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex Hermitian positive
    definite band system of linear equations with multiple right-hand
    sides AX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``zpbrfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zpbrfs(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _afb,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZpbrfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

DptsvReturnData = _collections.namedtuple(
    'DptsvReturnData',
    ['d', 'e', 'b'],
)

def dptsv(d, e, b):
    u"""
    Computes the solution to a real symmetric positive definite
    tridiagonal system of linear equations.

    ``dptsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric positive definite tridiagonal matrix,
    and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07ja

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07jaf.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        The n diagonal elements of the tridiagonal matrix A.

    e : float, array-like, shape (n-1)
        The (n-1) subdiagonal elements of the tridiagonal matrix A.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    d : float, ndarray, shape (n)
        The n diagonal elements of the diagonal matrix D from the
        factorization A = LDL^T.

    e : float, ndarray, shape (n-1)
        The (n-1) subdiagonal elements of the unit bidiagonal factor L
        from the LDL^T factorization of A. (`e` can also be regarded as
        the superdiagonal of the unit bidiagonal factor U from the U^TDU
        factorization of A.)

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite, and the solution has not been computed. The
            factorization has not been completed unless n = *<value>*.

    Notes
    -----
    ``dptsv`` factors A as A = LDL^T.
    The factored form of A is then used to solve the system of
    equations.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dptsv'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dptsv(
        _n,
        _nrhs,
        _d,
        _e,
        _b,
    )
    _d_py = _d
    _e_py = _e
    _b_py = _b
    _return_data = DptsvReturnData(
        _d_py,
        _e_py,
        _b_py,
    )
    return _return_data

DptsvxReturnData = _collections.namedtuple(
    'DptsvxReturnData',
    ['df', 'ef', 'x', 'rcond', 'ferr', 'berr'],
)

def dptsvx(fact, nrhs, d, e, df, ef, b):
    u"""
    Uses the LDL^T factorization to compute the solution, error-bound
    and condition estimate for a real symmetric positive definite
    tridiagonal system of linear equations.

    ``dptsvx`` uses the factorization

        A = LDL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric positive definite tridiagonal matrix and
    X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07jb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07jbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `df` and `ef` contain the factorized form of the matrix A.
            `df` and `ef` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `df` and `ef` and factorized.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    d : float, array-like, shape (n)
        The n diagonal elements of the tridiagonal matrix A.

    e : float, array-like, shape (n-1)
        The (n-1) subdiagonal elements of the tridiagonal matrix A.

    df : float, array-like, shape (n)
        If `fact` = 'F', `df` must contain the n diagonal elements of
        the diagonal matrix D from the LDL^T factorization of A.

    ef : float, array-like, shape (n-1)
        If `fact` = 'F', `ef` must contain the (n-1) subdiagonal
        elements of the unit bidiagonal factor L from the LDL^T
        factorization of A.

    b : float, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    df : float, ndarray, shape (n)
        If `fact` = 'N', `df` contains the n diagonal elements of the
        diagonal matrix D from the LDL^T factorization of A.

    ef : float, ndarray, shape (n-1)
        If `fact` = 'N', `ef` contains the (n-1) subdiagonal elements of
        the unit bidiagonal factor L from the LDL^T factorization of A.

    x : float, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X.

    rcond : float
        The reciprocal condition number of the matrix A. If `rcond` is
        less than the machine precision (in particular, if `rcond` =
        0.0), the matrix is singular to working precision. This
        condition is indicated by a return code of `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        The forward error bound for each solution vector x^_j (the jth
        column of the solution matrix X). If x_j is the true solution
        corresponding to x^_j, `ferr`[j-1] is an estimated upper bound
        for the magnitude of the largest element in (x^_j-x_j) divided
        by the magnitude of the largest element in x^_j.

    berr : float, ndarray, shape (`nrhs`)
        The component-wise relative backward error of each solution
        vector x^_j (i.e., the smallest relative change in any element
        of A or B that makes x^_j an exact solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` (i > 0) and (i<=n))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``dptsvx`` performs the following steps:

    (1) If `fact` = 'N', the matrix A is factorized as A = LDL^T, where
        L is a unit lower bidiagonal matrix and D is diagonal. The
        factorization can also be regarded as having the form A = U^TDU.

    (#) If the leading i*i principal minor is not positive definite,
        then the function returns with errno = i. Otherwise, the
        factored form of A is used to estimate the condition number of
        the matrix A. If the reciprocal of the condition number is less
        than machine precision, `errno` = `n` + 1 is returned as a
        warning, but the function still goes on to solve for X and
        compute error bounds as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dptsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ef = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ef'},
        ef,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.dptsvx(
            fact,
            _n,
            _nrhs,
            _d,
            _e,
            _df,
            _ef,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _df_py = _df
    _ef_py = _ef
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DptsvxReturnData(
        _df_py,
        _ef_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DpttrfReturnData = _collections.namedtuple(
    'DpttrfReturnData',
    ['d', 'e'],
)

def dpttrf(d, e):
    u"""
    Computes the LDL^T factorization of a real symmetric positive
    definite tridiagonal matrix.

    ``dpttrf`` computes the modified Cholesky factorization of a real
    n*n symmetric positive definite tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f07jd

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07jdf.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the matrix A.

    e : float, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    Returns
    -------
    d : float, ndarray, shape (n)
        Is overwritten by the n diagonal elements of the diagonal matrix
        D from the LDL^T factorization of A.

    e : float, ndarray, shape (n-1)
        Is overwritten by the (n-1) subdiagonal elements of the lower
        bidiagonal matrix L. (`e` can also be regarded as containing the
        (n-1) superdiagonal elements of the upper bidiagonal matrix U.)

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` (i > 0) and (i < n))
            The leading minor of order *<value>* is not positive
            definite, the factorization could not be completed.

        (`errno` (i > 0) and (n))
            The leading minor of order n is not positive definite, the
            factorization was completed, but `d`[n-1] <= 0.

    Notes
    -----
    ``dpttrf`` factorizes the matrix A as

        A = LDL^T,

    where L is a unit lower bidiagonal matrix and D is a diagonal matrix
    with positive diagonal elements.
    The factorization may also be regarded as having the form U^TDU,
    where U is a unit upper bidiagonal matrix.
    """

    fun_name = 'naginterfaces.library.lapacklin.dpttrf'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dpttrf(
        _n,
        _d,
        _e,
    )
    _d_py = _d
    _e_py = _e
    _return_data = DpttrfReturnData(
        _d_py,
        _e_py,
    )
    return _return_data

def dpttrs(d, e, b):
    u"""
    Solves a real symmetric positive definite tridiagonal system using
    the LDL^T factorization computed by ``dpttrf``.

    ``dpttrs`` computes the solution to a real system of linear
    equations AX = B, where A is an n*n symmetric positive definite
    tridiagonal matrix and X and B are n*r matrices, using the LDL^T
    factorization returned by ``dpttrf``.

    For full information please refer to the NAG Library document for
    f07je

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07jef.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^T factorization of A.

    e : float, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the unit lower
        bidiagonal matrix L. (`e` can also be regarded as the
        superdiagonal of the unit upper bidiagonal matrix U from the
        U^TDU factorization of A.)

    b : float, array-like, shape (n, nrhs)
        The n*r matrix of right-hand sides B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dpttrs`` should be preceded by a call to ``dpttrf``, which
    computes a modified Cholesky factorization of the matrix A as

        A = LDL^T,

    where L is a unit lower bidiagonal matrix and D is a diagonal
    matrix, with positive diagonal elements. ``dpttrs`` then utilizes
    the factorization to solve the required equations.
    Note that the factorization may also be regarded as having the form
    U^TDU, where U is a unit upper bidiagonal matrix.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dpttrs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dpttrs(
        _n,
        _nrhs,
        _d,
        _e,
        _b,
    )
    _b_py = _b
    return _b_py

def dptcon(n, d, e, anorm):
    u"""
    Computes the reciprocal of the condition number of a real symmetric
    positive definite tridiagonal system using the LDL^T factorization
    computed by ``dpttrf``.

    ``dptcon`` computes the reciprocal condition number of a real n*n
    symmetric positive definite tridiagonal matrix A, using the LDL^T
    factorization returned by ``dpttrf``.

    For full information please refer to the NAG Library document for
    f07jg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07jgf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^T factorization of A.

    e : float, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the unit lower
        bidiagonal matrix L. (`e` can also be regarded as the
        superdiagonal of the unit upper bidiagonal matrix U from the
        U^TDU factorization of A.)

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlanst`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dpttrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        The reciprocal condition number, 1/kappa_1(A) =
        1/(||A||_1||A^-1||_1).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``dptcon`` should be preceded by a call to ``dpttrf``, which
    computes a modified Cholesky factorization of the matrix A as

        A = LDL^T,

    where L is a unit lower bidiagonal matrix and D is a diagonal
    matrix, with positive diagonal elements. ``dptcon`` then utilizes
    the factorization to compute ||A^-1||_1 by a direct method, from
    which the reciprocal of the condition number of A, 1/kappa(A) is
    computed as

        1/kappa_1(A) = 1/(||A||_1||A^-1||_1).

    1/kappa(A) is returned, rather than kappa(A), since when A is
    singular kappa(A) is infinite.

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dptcon'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dptcon(
        n,
        _d,
        _e,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DptrfsReturnData = _collections.namedtuple(
    'DptrfsReturnData',
    ['x', 'ferr', 'berr'],
)

def dptrfs(nrhs, d, e, df, ef, b, x):
    u"""
    Refined solution with error bounds of real symmetric positive
    definite tridiagonal system of linear equations, multiple right-hand
    sides.

    ``dptrfs`` computes error bounds and refines the solution to a real
    system of linear equations AX = B, where A is an n*n symmetric
    positive definite tridiagonal matrix and X and B are n*r matrices,
    using the modified Cholesky factorization returned by ``dpttrf`` and
    an initial solution returned by ``dpttrs``.
    Iterative refinement is used to reduce the backward error as much as
    possible.

    For full information please refer to the NAG Library document for
    f07jh

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07jhf.html

    Parameters
    ----------
    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the matrix of A.

    e : float, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    df : float, array-like, shape (n)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^T factorization of A.

    ef : float, array-like, shape (n)
        Must contain the (n-1) subdiagonal elements of the unit
        bidiagonal matrix L from the LDL^T factorization of A.

    b : float, array-like, shape (n, `nrhs`)
        The n*r matrix of right-hand sides B.

    x : float, array-like, shape (n, `nrhs`)
        The n*r initial solution matrix X.

    Returns
    -------
    x : float, ndarray, shape (n, `nrhs`)
        The n*r refined solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        Estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x^_j||_infinity <=
        `ferr`[j-1], where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is almost always a
        slight overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        Estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Notes
    -----
    ``dptrfs`` should normally be preceded by calls to ``dpttrf`` and
    ``dpttrs``. ``dpttrf`` computes a modified Cholesky factorization of
    the matrix A as

        A = LDL^T,

    where L is a unit lower bidiagonal matrix and D is a diagonal
    matrix, with positive diagonal elements. ``dpttrs`` then utilizes
    the factorization to compute a solution, X^, to the required
    equations.
    Letting x^ denote a column of X^, ``dptrfs`` computes a
    `component-wise backward error`, beta, the smallest relative
    perturbation in each element of A and b such that x^ is the exact
    solution of a perturbed system

        (A+E)x^= b+f, with |e_{ij}| <= beta|a_{ij}|, and |f_j| <=
        beta|b_j|.

    The function also estimates a bound for the `component-wise forward
    error` in the computed solution defined by max|x_i-x_i^|/max|x_i^|,
    where x is the corresponding column of the exact solution, X.

    Note that the modified Cholesky factorization of A can also be
    expressed as

        A = U^TDU,

    where U is unit upper bidiagonal.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dptrfs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ef = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ef'},
        ef,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dptrfs(
        _n,
        _nrhs,
        _d,
        _e,
        _df,
        _ef,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DptrfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZptsvReturnData = _collections.namedtuple(
    'ZptsvReturnData',
    ['d', 'e', 'b'],
)

def zptsv(d, e, b):
    u"""
    Computes the solution to a complex Hermitian positive definite
    tridiagonal system of linear equations.

    ``zptsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n Hermitian positive definite tridiagonal matrix,
    and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07jn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07jnf.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        The n diagonal elements of the tridiagonal matrix A.

    e : complex, array-like, shape (n-1)
        The (n-1) subdiagonal elements of the tridiagonal matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    d : float, ndarray, shape (n)
        The n diagonal elements of the diagonal matrix D from the
        factorization A = LDL^H.

    e : complex, ndarray, shape (n-1)
        The (n-1) subdiagonal elements of the unit bidiagonal factor L
        from the LDL^H factorization of A. (`e` can also be regarded as
        the superdiagonal of the unit bidiagonal factor U from the U^HDU
        factorization of A.)

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite, and the solution has not been computed. The
            factorization has not been completed unless n = *<value>*.

    Notes
    -----
    ``zptsv`` factors A as A = LDL^H.
    The factored form of A is then used to solve the system of
    equations.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zptsv'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
    )
    _b_lapacklin.zptsv(
        _n,
        _nrhs,
        _d,
        _e,
        _b,
    )
    _d_py = _d
    _e_py = _e
    _b_py = _b
    _return_data = ZptsvReturnData(
        _d_py,
        _e_py,
        _b_py,
    )
    return _return_data

ZptsvxReturnData = _collections.namedtuple(
    'ZptsvxReturnData',
    ['df', 'ef', 'x', 'rcond', 'ferr', 'berr'],
)

def zptsvx(fact, n, nrhs, d, e, df, ef, b):
    u"""
    Uses the LDL^T factorization to compute the solution, error-bound
    and condition estimate for a complex Hermitian positive definite
    tridiagonal system of linear equations.

    ``zptsvx`` uses the factorization

        A = LDL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n*n Hermitian positive definite tridiagonal matrix and
    X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07jp

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07jpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `df` and `ef` contain the factorized form of the matrix A.
            `df` and `ef` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `df` and `ef` and factorized.

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    d : float, array-like, shape (`n`)
        The n diagonal elements of the tridiagonal matrix A.

    e : complex, array-like, shape (`n`-1)
        The (n-1) subdiagonal elements of the tridiagonal matrix A.

    df : float, array-like, shape (`n`)
        If `fact` = 'F', `df` must contain the n diagonal elements of
        the diagonal matrix D from the LDL^H factorization of A.

    ef : complex, array-like, shape (`n`-1)
        If `fact` = 'F', `ef` must contain the (n-1) subdiagonal
        elements of the unit bidiagonal factor L from the LDL^H
        factorization of A.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    df : float, ndarray, shape (`n`)
        If `fact` = 'N', `df` contains the n diagonal elements of the
        diagonal matrix D from the LDL^H factorization of A.

    ef : complex, ndarray, shape (`n`-1)
        If `fact` = 'N', `ef` contains the (n-1) subdiagonal elements of
        the unit bidiagonal factor L from the LDL^H factorization of A.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X.

    rcond : float
        The reciprocal condition number of the matrix A. If `rcond` is
        less than the machine precision (in particular, if `rcond` =
        0.0), the matrix is singular to working precision. This
        condition is indicated by a return code of `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        The forward error bound for each solution vector x^_j (the jth
        column of the solution matrix X). If x_j is the true solution
        corresponding to x^_j, `ferr`[j-1] is an estimated upper bound
        for the magnitude of the largest element in (x^_j-x_j) divided
        by the magnitude of the largest element in x^_j.

    berr : float, ndarray, shape (`nrhs`)
        The component-wise relative backward error of each solution
        vector x^_j (i.e., the smallest relative change in any element
        of A or B that makes x^_j an exact solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``zptsvx`` performs the following steps:

    (1) If `fact` = 'N', the matrix A is factorized as A = LDL^H, where
        L is a unit lower bidiagonal matrix and D is diagonal. The
        factorization can also be regarded as having the form A = U^HDU.

    (#) If the leading i*i principal minor is not positive definite,
        then the function returns with errno = i. Otherwise, the
        factored form of A is used to estimate the condition number of
        the matrix A. If the reciprocal of the condition number is less
        than machine precision, `errno` = `n` + 1 is returned as a
        warning, but the function still goes on to solve for X and
        compute error bounds as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zptsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ef = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ef'},
        ef,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zptsvx(
            fact,
            _n,
            _nrhs,
            _d,
            _e,
            _df,
            _ef,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _df_py = _df
    _ef_py = _ef
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZptsvxReturnData(
        _df_py,
        _ef_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZpttrfReturnData = _collections.namedtuple(
    'ZpttrfReturnData',
    ['d', 'e'],
)

def zpttrf(d, e):
    u"""
    Computes the LDL^H factorization of a complex Hermitian positive
    definite tridiagonal matrix.

    ``zpttrf`` computes the modified Cholesky factorization of a complex
    n*n Hermitian positive definite tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f07jr

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07jrf.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the matrix A.

    e : complex, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    Returns
    -------
    d : float, ndarray, shape (n)
        Is overwritten by the n diagonal elements of the diagonal matrix
        D from the LDL^H factorization of A.

    e : complex, ndarray, shape (n-1)
        Is overwritten by the (n-1) subdiagonal elements of the lower
        bidiagonal matrix L. (`e` can also be regarded as containing the
        (n-1) superdiagonal elements of the upper bidiagonal matrix U.)

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` (i > 0) and (i < n))
            The leading minor of order *<value>* is not positive
            definite, the factorization could not be completed.

        (`errno` (i > 0) and (n))
            The leading minor of order n is not positive definite, the
            factorization was completed, but `d`[n-1] <= 0.

    Notes
    -----
    ``zpttrf`` factorizes the matrix A as

        A = LDL^H,

    where L is a unit lower bidiagonal matrix and D is a diagonal matrix
    with positive diagonal elements.
    The factorization may also be regarded as having the form U^HDU,
    where U is a unit upper bidiagonal matrix.
    """

    fun_name = 'naginterfaces.library.lapacklin.zpttrf'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
    )
    _b_lapacklin.zpttrf(
        _n,
        _d,
        _e,
    )
    _d_py = _d
    _e_py = _e
    _return_data = ZpttrfReturnData(
        _d_py,
        _e_py,
    )
    return _return_data

def zpttrs(uplo, d, e, b):
    u"""
    Solves a complex Hermitian positive definite tridiagonal system
    using the LDL^H factorization computed by ``zpttrf``.

    ``zpttrs`` computes the solution to a complex system of linear
    equations AX = B, where A is an n*n Hermitian positive definite
    tridiagonal matrix and X and B are n*r matrices, using the LDL^H
    factorization returned by ``zpttrf``.

    For full information please refer to the NAG Library document for
    f07js

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07jsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies the form of the factorization as follows:

        `uplo` = 'U'

            A = U^HDU.

        `uplo` = 'L'

            A = LDL^H.

    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^H or U^HDU factorization of A.

    e : complex, array-like, shape (n-1)
        If `uplo` = 'U', `e` must contain the (n-1) superdiagonal
        elements of the unit upper bidiagonal matrix U from the U^HDU
        factorization of A.

        If `uplo` = 'L', `e` must contain the (n-1) subdiagonal elements
        of the unit lower bidiagonal matrix L from the LDL^H
        factorization of A.

    b : complex, array-like, shape (n, nrhs)
        The n*r matrix of right-hand sides B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zpttrs`` should be preceded by a call to ``zpttrf``, which
    computes a modified Cholesky factorization of the matrix A as

        A = LDL^H,

    where L is a unit lower bidiagonal matrix and D is a diagonal
    matrix, with positive diagonal elements. ``zpttrs`` then utilizes
    the factorization to solve the required equations.
    Note that the factorization may also be regarded as having the form
    U^HDU, where U is a unit upper bidiagonal matrix.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.zpttrs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
    )
    _b_lapacklin.zpttrs(
        uplo,
        _n,
        _nrhs,
        _d,
        _e,
        _b,
    )
    _b_py = _b
    return _b_py

def zptcon(n, d, e, anorm):
    u"""
    Computes the reciprocal of the condition number of a complex
    Hermitian positive definite tridiagonal system using the LDL^H
    factorization computed by ``zpttrf``.

    ``zptcon`` computes the reciprocal condition number of a complex n*n
    Hermitian positive definite tridiagonal matrix A, using the LDL^H
    factorization returned by ``zpttrf``.

    For full information please refer to the NAG Library document for
    f07ju

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07juf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^H factorization of A.

    e : complex, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the unit lower
        bidiagonal matrix L. (`e` can also be regarded as the
        superdiagonal of the unit upper bidiagonal matrix U from the
        U^HDU factorization of A.)

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanht`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zpttrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        The reciprocal condition number, 1/kappa_1(A) =
        1/(||A||_1||A^-1||_1).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``zptcon`` should be preceded by a call to ``zpttrf``, which
    computes a modified Cholesky factorization of the matrix A as

        A = LDL^H,

    where L is a unit lower bidiagonal matrix and D is a diagonal
    matrix, with positive diagonal elements. ``zptcon`` then utilizes
    the factorization to compute ||A^-1||_1 by a direct method, from
    which the reciprocal of the condition number of A, 1/kappa(A) is
    computed as

        1/kappa_1(A) = 1/(||A||_1||A^-1||_1).

    1/kappa(A) is returned, rather than kappa(A), since when A is
    singular kappa(A) is infinite.

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zptcon'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zptcon(
        n,
        _d,
        _e,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZptrfsReturnData = _collections.namedtuple(
    'ZptrfsReturnData',
    ['x', 'ferr', 'berr'],
)

def zptrfs(uplo, n, nrhs, d, e, df, ef, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian positive
    definite tridiagonal system of linear equations, multiple right-hand
    sides.

    ``zptrfs`` computes error bounds and refines the solution to a
    complex system of linear equations AX = B, where A is an n*n
    Hermitian positive definite tridiagonal matrix and X and B are n*r
    matrices, using the modified Cholesky factorization returned by
    ``zpttrf`` and an initial solution returned by ``zpttrs``.
    Iterative refinement is used to reduce the backward error as much as
    possible.

    For full information please refer to the NAG Library document for
    f07jv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07jvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies the form of the factorization as follows:

        `uplo` = 'U'

            A = U^HDU.

        `uplo` = 'L'

            A = LDL^H.

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the matrix of A.

    e : complex, array-like, shape (`n`-1)
        If `uplo` = 'U', `e` must contain the (n-1) superdiagonal
        elements of the matrix A.

        If `uplo` = 'L', `e` must contain the (n-1) subdiagonal elements
        of the matrix A.

    df : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^T factorization of A.

    ef : complex, array-like, shape (`n`-1)
        If `uplo` = 'U', `ef` must contain the (n-1) superdiagonal
        elements of the unit upper bidiagonal matrix U from the U^HDU
        factorization of A.

        If `uplo` = 'L', `ef` must contain the (n-1) subdiagonal
        elements of the unit lower bidiagonal matrix L from the LDL^H
        factorization of A.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n*r matrix of right-hand sides B.

    x : complex, array-like, shape (`n`, `nrhs`)
        The n*r initial solution matrix X.

    Returns
    -------
    x : complex, ndarray, shape (`n`, `nrhs`)
        The n*r refined solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        Estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x^_j||_infinity <=
        `ferr`[j-1], where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is almost always a
        slight overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        Estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Notes
    -----
    ``zptrfs`` should normally be preceded by calls to ``zpttrf`` and
    ``zpttrs``. ``zpttrf`` computes a modified Cholesky factorization of
    the matrix A as

        A = LDL^H,

    where L is a unit lower bidiagonal matrix and D is a diagonal
    matrix, with positive diagonal elements. ``zpttrs`` then utilizes
    the factorization to compute a solution, X^, to the required
    equations.
    Letting x^ denote a column of X^, ``zptrfs`` computes a
    `component-wise backward error`, beta, the smallest relative
    perturbation in each element of A and b such that x^ is the exact
    solution of a perturbed system

        (A+E)x^= b+f, with |e_{ij}| <= beta|a_{ij}|, and |f_j| <=
        beta|b_j|.

    The function also estimates a bound for the `component-wise forward
    error` in the computed solution defined by max|x_i-x_i^|/max|x_i^|,
    where x is the corresponding column of the exact solution, X.

    Note that the modified Cholesky factorization of A can also be
    expressed as

        A = U^HDU,

    where U is unit upper bidiagonal.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.zptrfs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ef = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ef'},
        ef,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zptrfs(
        uplo,
        n,
        _nrhs,
        _d,
        _e,
        _df,
        _ef,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZptrfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

DpstrfReturnData = _collections.namedtuple(
    'DpstrfReturnData',
    ['a', 'piv', 'rank'],
)

def dpstrf(uplo, n, a, tol=-1):
    u"""
    Cholesky factorization, with complete pivoting, of a real,
    symmetric, positive semidefinite matrix.

    ``dpstrf`` computes the Cholesky factorization with complete
    pivoting of a real symmetric positive semidefinite matrix.

    For full information please refer to the NAG Library document for
    f07kd

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07kdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n*n symmetric positive semidefinite matrix A.

    tol : float, optional
        User defined tolerance. If `tol` < 0, n*max_{k =
        1,n}|A_{kk}|*machine precision will be used. The algorithm
        terminates at the rth step if the (r+1)th step pivot < `tol`.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        If `uplo` = 'U', the first `rank` rows of the upper triangle of
        A are overwritten with the nonzero elements of the Cholesky
        factor U, and the remaining rows of the triangle are destroyed.

        If `uplo` = 'L', the first `rank` columns of the lower triangle
        of A are overwritten with the nonzero elements of the Cholesky
        factor L, and the remaining columns of the triangle are
        destroyed.

    piv : int, ndarray, shape (`n`)
        `piv` is such that the nonzero entries of P are P(`piv`[k-1],k)
        = 1, for k = 1,2,...,n.

    rank : int
        The computed rank of A given by the number of steps the
        algorithm completed.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            The matrix A is not positive definite. It is either positive
            semidefinite with computed rank as returned in `rank` and
            less than n, or it may be indefinite, see [Further
            Comments].

    Notes
    -----
    ``dpstrf`` forms the Cholesky factorization of a real symmetric
    positive semidefinite matrix A either as P^TAP = U^TU if `uplo` =
    'U' or P^TAP = LL^T if `uplo` = 'L', where P is a permutation
    matrix, U is an upper triangular matrix and L is lower triangular.

    This algorithm does not attempt to check that A is positive
    semidefinite.

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia

    Lucas, C, 2004, `LAPACK-style codes for Level 2 and 3 pivoted
    Cholesky factorizations`, LAPACK Working Note No. 161. Technical
    Report CS-04-522, Department of Computer Science, University of
    Tennessee, 107 Ayres Hall, Knoxville, TN 37996-1301, USA,
    https://www.netlib.org/lapack/lawnspdf/lawn161.pdf
    """

    fun_name = 'naginterfaces.library.lapacklin.dpstrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _piv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_lapacklin.dpstrf(
            uplo,
            _n,
            _a,
            _piv,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _piv_py = _piv
    _rank_py = _return_scalar
    _return_data = DpstrfReturnData(
        _a_py,
        _piv_py,
        _rank_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZpstrfReturnData = _collections.namedtuple(
    'ZpstrfReturnData',
    ['a', 'piv', 'rank'],
)

def zpstrf(uplo, n, a, tol=-1):
    u"""
    Cholesky factorization of complex Hermitian positive semidefinite
    matrix.

    ``zpstrf`` computes the Cholesky factorization with complete
    pivoting of a complex Hermitian positive semidefinite matrix.

    For full information please refer to the NAG Library document for
    f07kr

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07krf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n*n Hermitian positive semidefinite matrix A.

    tol : float, optional
        User defined tolerance. If `tol` < 0, n*max_{k =
        1,n}|A_{kk}|*machine precision will be used. The algorithm
        terminates at the rth step if the (r+1)th step pivot < `tol`.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        If `uplo` = 'U', the first `rank` rows of the upper triangle of
        A are overwritten with the nonzero elements of the Cholesky
        factor U, and the remaining rows of the triangle are destroyed.

        If `uplo` = 'L', the first `rank` columns of the lower triangle
        of A are overwritten with the nonzero elements of the Cholesky
        factor L, and the remaining columns of the triangle are
        destroyed.

    piv : int, ndarray, shape (`n`)
        `piv` is such that the nonzero entries of P are P(`piv`[k-1],k)
        = 1, for k = 1,2,...,n.

    rank : int
        The computed rank of A given by the number of steps the
        algorithm completed.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            The matrix A is not positive definite. It is either positive
            semidefinite with computed rank as returned in `rank` and
            less than n, or it may be indefinite, see [Further
            Comments].

    Notes
    -----
    ``zpstrf`` forms the Cholesky factorization of a complex Hermitian
    positive semidefinite matrix A either as P^TAP = U^HU if `uplo` =
    'U' or P^TAP = LL^H if `uplo` = 'L', where P is a permutation
    matrix, U is an upper triangular matrix and L is lower triangular.

    This algorithm does not attempt to check that A is positive
    semidefinite.

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia

    Lucas, C, 2004, `LAPACK-style codes for Level 2 and 3 pivoted
    Cholesky factorizations`, LAPACK Working Note No. 161. Technical
    Report CS-04-522, Department of Computer Science, University of
    Tennessee, 107 Ayres Hall, Knoxville, TN 37996-1301, USA,
    https://www.netlib.org/lapack/lawnspdf/lawn161.pdf
    """

    fun_name = 'naginterfaces.library.lapacklin.zpstrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _piv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zpstrf(
            uplo,
            _n,
            _a,
            _piv,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _piv_py = _piv
    _rank_py = _return_scalar
    _return_data = ZpstrfReturnData(
        _a_py,
        _piv_py,
        _rank_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DsysvReturnData = _collections.namedtuple(
    'DsysvReturnData',
    ['a', 'ipiv', 'b'],
)

def dsysv(uplo, a, b):
    u"""
    Computes the solution to a real symmetric system of linear
    equations.

    ``dsysv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric matrix and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07ma

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07maf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : float, array-like, shape (n, n)
        The n*n symmetric matrix A.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        If no exception or warning is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^T or A = LDL^T as computed by ``dsytrf``.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    Notes
    -----
    ``dsysv`` uses the diagonal pivoting method to factor A as A = UDU^T
    if `uplo` = 'U' or A = LDL^T if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices,
    and D is symmetric and block diagonal with 1*1 and 2*2 diagonal
    blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    Note that, in general, different permutations (pivot sequences) and
    diagonal block structures are obtained for `uplo` = 'U' or 'L'

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsysv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dsysv(
            uplo,
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = DsysvReturnData(
        _a_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DsysvxReturnData = _collections.namedtuple(
    'DsysvxReturnData',
    ['af', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
)

def dsysvx(fact, uplo, a, af, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a real symmetric system of linear equations.

    ``dsysvx`` uses the diagonal pivoting factorization to compute the
    solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric matrix and X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07mb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07mbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `af` and `ipiv` contain the factorized form of the matrix A.
            `af` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : float, array-like, shape (n, n)
        The n*n symmetric matrix A.

    af : float, array-like, shape (n, n)
        If `fact` = 'F', `af` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^T or `a` = LDL^T as computed by
        ``dsytrf``.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``dsytrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    af : float, ndarray, shape (n, n)
        If `fact` = 'N', `af` returns the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^T or `a` = LDL^T.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``dsytrf``, as
        described above.

    x : float, ndarray, shape (n, nrhs)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``dsysvx`` performs the following steps:

    (1) If `fact` = 'N', the diagonal pivoting method is used to factor
        A. The form of the factorization is A = UDU^T if `uplo` = 'U' or
        A = LDL^T if `uplo` = 'L', where U (or L) is a product of
        permutation and unit upper (lower) triangular matrices, and D is
        symmetric and block diagonal with 1*1 and 2*2 diagonal blocks.

    (#) If some d_{ii} = 0, so that D is exactly singular, then the
        function returns with errno = i. Otherwise, the factored form of
        A is used to estimate the condition number of the matrix A. If
        the reciprocal of the condition number is less than machine
        precision, `errno` = `n` + 1 is returned as a warning, but the
        function still goes on to solve for X and compute error bounds
        as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dsysvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.dsysvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _a,
            _af,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _af_py = _af
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DsysvxReturnData(
        _af_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DsytrfReturnData = _collections.namedtuple(
    'DsytrfReturnData',
    ['a', 'ipiv'],
)

def dsytrf(uplo, a):
    u"""
    Bunch--Kaufman factorization of real symmetric indefinite matrix.

    ``dsytrf`` computes the Bunch--Kaufman factorization of a real
    symmetric indefinite matrix.

    For full information please refer to the NAG Library document for
    f07md

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07mdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^TP^T, where L is lower triangular.

    a : float, array-like, shape (n, n)
        The n*n symmetric indefinite matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The upper or lower triangle of A is overwritten by details of
        the block diagonal matrix D and the multipliers used to obtain
        the factor U or L as specified by `uplo`.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    Notes
    -----
    ``dsytrf`` factorizes a real symmetric matrix A, using the
    Bunch--Kaufman diagonal pivoting method. A is factorized as either A
    = PUDU^TP^T if `uplo` = 'U' or A = PLDL^TP^T if `uplo` = 'L', where
    P is a permutation matrix, U (or L) is a unit upper (or lower)
    triangular matrix and D is a symmetric block diagonal matrix with
    1*1 and 2*2 diagonal blocks; U (or L) has 2*2 unit diagonal blocks
    corresponding to the 2*2 blocks of D.
    Row and column interchanges are performed to ensure numerical
    stability while preserving symmetry.

    This method is suitable for symmetric matrices which are not known
    to be positive definite.
    If A is in fact positive definite, no interchanges are performed and
    no 2*2 blocks occur in D.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsytrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dsytrf(
            uplo,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _ipiv_py = _ipiv
    _return_data = DsytrfReturnData(
        _a_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dsytrs(uplo, a, ipiv, b):
    u"""
    Solution of real symmetric indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by ``dsytrf``.

    ``dsytrs`` solves a real symmetric indefinite system of linear
    equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``dsytrf``.

    For full information please refer to the NAG Library document for
    f07me

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07mef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    a : float, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``dsytrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``dsytrf``.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dsytrs`` is used to solve a real symmetric indefinite system of
    linear equations AX = B, this function must be preceded by a call to
    ``dsytrf`` which computes the Bunch--Kaufman factorization of A.

    If `uplo` = 'U', A = PUDU^TP^T, where P is a permutation matrix, U
    is an upper triangular matrix and D is a symmetric block diagonal
    matrix with 1*1 and 2*2 blocks; the solution X is computed by
    solving PUDY = B and then U^TP^TX = Y.

    If `uplo` = 'L', A = PLDL^TP^T, where L is a lower triangular
    matrix; the solution X is computed by solving PLDY = B and then
    L^TP^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsytrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.dsytrs(
        uplo,
        _n,
        _nrhs,
        _a,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def dsycon(uplo, n, a, ipiv, anorm):
    u"""
    Estimate condition number of real symmetric indefinite matrix,
    matrix already factorized by ``dsytrf``.

    ``dsycon`` estimates the condition number of a real symmetric
    indefinite matrix A, where A has been factorized by ``dsytrf``.

    For full information please refer to the NAG Library document for
    f07mg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07mgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``dsytrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsytrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlansy`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dsytrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -6)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``dsycon`` estimates the condition number (in the 1-norm) of a real
    symmetric indefinite matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is symmetric, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.dlansy`` to
    compute ||A||_1 and a call to ``dsytrf`` to compute the
    Bunch--Kaufman factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dsycon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.dsycon(
        uplo,
        n,
        _a,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DsyrfsReturnData = _collections.namedtuple(
    'DsyrfsReturnData',
    ['x', 'ferr', 'berr'],
)

def dsyrfs(uplo, n, a, af, ipiv, b, x):
    u"""
    Refined solution with error bounds of real symmetric indefinite
    system of linear equations, multiple right-hand sides.

    ``dsyrfs`` returns error bounds for the solution of a real symmetric
    indefinite system of linear equations with multiple right-hand
    sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07mh

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07mhf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n*n original symmetric matrix A as supplied to ``dsytrf``.

    af : float, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``dsytrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsytrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``dsytrs``.

    Returns
    -------
    x : float, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dsyrfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a real symmetric indefinite
    system of linear equations with multiple right-hand sides AX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``dsyrfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsyrfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dsyrfs(
        uplo,
        n,
        _nrhs,
        _a,
        _af,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DsyrfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

DsytriReturnData = _collections.namedtuple(
    'DsytriReturnData',
    ['a'],
)

def dsytri(uplo, n, a, ipiv):
    u"""
    Inverse of real symmetric indefinite matrix, matrix already
    factorized by ``dsytrf``.

    ``dsytri`` computes the inverse of a real symmetric indefinite
    matrix A, where A has been factorized by ``dsytrf``.

    For full information please refer to the NAG Library document for
    f07mj

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07mjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``dsytrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsytrf``.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        The factorization is overwritten by the n*n symmetric matrix
        A^-1.

        If `uplo` = 'U', the upper triangle of A^-1 is stored in the
        upper triangular part of the array.

        If `uplo` = 'L', the lower triangle of A^-1 is stored in the
        lower triangular part of the array.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    Notes
    -----
    ``dsytri`` is used to compute the inverse of a real symmetric
    indefinite matrix A, the function must be preceded by a call to
    ``dsytrf``, which computes the Bunch--Kaufman factorization of A.

    If `uplo` = 'U', A = PUDU^TP^T and A^-1 is computed by solving
    U^TP^TXPU = D^-1 for X.

    If `uplo` = 'L', A = PLDL^TP^T and A^-1 is computed by solving
    L^TP^TXPL = D^-1 for X.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dsytri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.dsytri(
            uplo,
            n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = DsytriReturnData(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

ZhesvReturnData = _collections.namedtuple(
    'ZhesvReturnData',
    ['a', 'ipiv', 'b'],
)

def zhesv(uplo, a, b):
    u"""
    Computes the solution to a complex Hermitian system of linear
    equations.

    ``zhesv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n Hermitian matrix and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07mn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07mnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : complex, array-like, shape (n, n)
        The n*n Hermitian matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If no exception or warning is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zhetrf``.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    Notes
    -----
    ``zhesv`` uses the diagonal pivoting method to factor A as A = UDU^H
    if `uplo` = 'U' or A = LDL^H if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices,
    and D is Hermitian and block diagonal with 1*1 and 2*2 diagonal
    blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhesv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zhesv(
            uplo,
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = ZhesvReturnData(
        _a_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZhesvxReturnData = _collections.namedtuple(
    'ZhesvxReturnData',
    ['af', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
)

def zhesvx(fact, uplo, a, af, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a complex Hermitian system of linear equations.

    ``zhesvx`` uses the diagonal pivoting factorization to compute the
    solution to a complex system of linear equations

        AX = B,

    where A is an n*n Hermitian matrix and X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07mp

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07mpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `af` and `ipiv` contain the factorized form of the matrix A.
            `af` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : complex, array-like, shape (n, n)
        The n*n Hermitian matrix A.

    af : complex, array-like, shape (n, n)
        If `fact` = 'F', `af` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^H or `a` = LDL^H as computed by
        ``zhetrf``.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zhetrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    af : complex, ndarray, shape (n, n)
        If `fact` = 'N', `af` returns the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^H or `a` = LDL^H.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zhetrf``, as
        described above.

    x : complex, ndarray, shape (n, nrhs)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``zhesvx`` performs the following steps:

    (1) If `fact` = 'N', the diagonal pivoting method is used to factor
        A. The form of the factorization is A = UDU^H if `uplo` = 'U' or
        A = LDL^H if `uplo` = 'L', where U (or L) is a product of
        permutation and unit upper (lower) triangular matrices, and D is
        Hermitian and block diagonal with 1*1 and 2*2 diagonal blocks.

    (#) If some d_{ii} = 0, so that D is exactly singular, then the
        function returns with errno = i. Otherwise, the factored form of
        A is used to estimate the condition number of the matrix A. If
        the reciprocal of the condition number is less than machine
        precision, `errno` = `n` + 1 is returned as a warning, but the
        function still goes on to solve for X and compute error bounds
        as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zhesvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zhesvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _a,
            _af,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _af_py = _af
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZhesvxReturnData(
        _af_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZhetrfReturnData = _collections.namedtuple(
    'ZhetrfReturnData',
    ['a', 'ipiv'],
)

def zhetrf(uplo, n, a):
    u"""
    Bunch--Kaufman factorization of complex Hermitian indefinite matrix.

    ``zhetrf`` computes the Bunch--Kaufman factorization of a complex
    Hermitian indefinite matrix.

    For full information please refer to the NAG Library document for
    f07mr

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07mrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n*n Hermitian indefinite matrix A.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        The upper or lower triangle of A is overwritten by details of
        the block diagonal matrix D and the multipliers used to obtain
        the factor U or L as specified by `uplo`.

    ipiv : int, ndarray, shape (`n`)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    Notes
    -----
    ``zhetrf`` factorizes a complex Hermitian matrix A, using the
    Bunch--Kaufman diagonal pivoting method. A is factorized either as A
    = PUDU^HP^T if `uplo` = 'U' or A = PLDL^HP^T if `uplo` = 'L', where
    P is a permutation matrix, U (or L) is a unit upper (or lower)
    triangular matrix and D is an Hermitian block diagonal matrix with
    1*1 and 2*2 diagonal blocks; U (or L) has 2*2 unit diagonal blocks
    corresponding to the 2*2 blocks of D.
    Row and column interchanges are performed to ensure numerical
    stability while keeping the matrix Hermitian.

    This method is suitable for Hermitian matrices which are not known
    to be positive definite.
    If A is in fact positive definite, no interchanges are performed and
    no 2*2 blocks occur in D.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhetrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zhetrf(
            uplo,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _ipiv_py = _ipiv
    _return_data = ZhetrfReturnData(
        _a_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zhetrs(uplo, a, ipiv, b):
    u"""
    Solution of complex Hermitian indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by ``zhetrf``.

    ``zhetrs`` solves a complex Hermitian indefinite system of linear
    equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zhetrf``.

    For full information please refer to the NAG Library document for
    f07ms

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07msf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``zhetrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zhetrf``.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zhetrs`` is used to solve a complex Hermitian indefinite system of
    linear equations AX = B, this function must be preceded by a call to
    ``zhetrf`` which computes the Bunch--Kaufman factorization of A.

    If `uplo` = 'U', A = PUDU^HP^T, where P is a permutation matrix, U
    is an upper triangular matrix and D is an Hermitian block diagonal
    matrix with 1*1 and 2*2 blocks; the solution X is computed by
    solving PUDY = B and then U^HP^TX = Y.

    If `uplo` = 'L', A = PLDL^HP^T, where L is a lower triangular
    matrix; the solution X is computed by solving PLDY = B and then
    L^HP^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhetrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zhetrs(
        uplo,
        _n,
        _nrhs,
        _a,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zhecon(uplo, a, ipiv, anorm):
    u"""
    Estimate condition number of complex Hermitian indefinite matrix,
    matrix already factorized by ``zhetrf``.

    ``zhecon`` estimates the condition number of a complex Hermitian
    indefinite matrix A, where A has been factorized by ``zhetrf``.

    For full information please refer to the NAG Library document for
    f07mu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07muf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``zhetrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zhetrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanhe`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zhetrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -6)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``zhecon`` estimates the condition number (in the 1-norm) of a
    complex Hermitian indefinite matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is Hermitian, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.zlanhe`` to
    compute ||A||_1 and a call to ``zhetrf`` to compute the
    Bunch--Kaufman factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zhecon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zhecon(
        uplo,
        _n,
        _a,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZherfsReturnData = _collections.namedtuple(
    'ZherfsReturnData',
    ['x', 'ferr', 'berr'],
)

def zherfs(uplo, n, a, af, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian indefinite
    system of linear equations, multiple right-hand sides.

    ``zherfs`` returns error bounds for the solution of a complex
    Hermitian indefinite system of linear equations with multiple
    right-hand sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07mv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07mvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n*n original Hermitian matrix A as supplied to ``zhetrf``.

    af : complex, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``zhetrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zhetrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``zhetrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zherfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex Hermitian indefinite
    system of linear equations with multiple right-hand sides AX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``zherfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zherfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zherfs(
        uplo,
        n,
        _nrhs,
        _a,
        _af,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZherfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZhetriReturnData = _collections.namedtuple(
    'ZhetriReturnData',
    ['a'],
)

def zhetri(uplo, n, a, ipiv):
    u"""
    Inverse of complex Hermitian indefinite matrix, matrix already
    factorized by ``zhetrf``.

    ``zhetri`` computes the inverse of a complex Hermitian indefinite
    matrix A, where A has been factorized by ``zhetrf``.

    For full information please refer to the NAG Library document for
    f07mw

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07mwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``zhetrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zhetrf``.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        The factorization is overwritten by the n*n Hermitian matrix
        A^-1.

        If `uplo` = 'U', the upper triangle of A^-1 is stored in the
        upper triangular part of the array.

        If `uplo` = 'L', the lower triangle of A^-1 is stored in the
        lower triangular part of the array.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    Notes
    -----
    ``zhetri`` is used to compute the inverse of a complex Hermitian
    indefinite matrix A, the function must be preceded by a call to
    ``zhetrf``, which computes the Bunch--Kaufman factorization of A.

    If `uplo` = 'U', A = PUDU^HP^T and A^-1 is computed by solving
    U^HP^TXPU = D^-1 for X.

    If `uplo` = 'L', A = PLDL^HP^T and A^-1 is computed by solving
    L^HP^TXPL = D^-1 for X.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zhetri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.zhetri(
            uplo,
            n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = ZhetriReturnData(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

ZsysvReturnData = _collections.namedtuple(
    'ZsysvReturnData',
    ['a', 'ipiv', 'b'],
)

def zsysv(uplo, a, b):
    u"""
    Computes the solution to a complex symmetric system of linear
    equations.

    ``zsysv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n symmetric matrix and X and B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07nn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07nnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : complex, array-like, shape (n, n)
        The n*n symmetric matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If no exception or warning is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zsytrf``.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    Notes
    -----
    ``zsysv`` uses the diagonal pivoting method to factor A as A = UDU^T
    if `uplo` = 'U' or A = LDL^T if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices,
    and D is symmetric and block diagonal with 1*1 and 2*2 diagonal
    blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zsysv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zsysv(
            uplo,
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = ZsysvReturnData(
        _a_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZsysvxReturnData = _collections.namedtuple(
    'ZsysvxReturnData',
    ['af', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
)

def zsysvx(fact, uplo, a, af, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a complex symmetric system of linear equations.

    ``zsysvx`` uses the diagonal pivoting factorization to compute the
    solution to a complex system of linear equations

        AX = B,

    where A is an n*n symmetric matrix and X and B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07np

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07npf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `af` and `ipiv` contain the factorized form of the matrix A.
            `af` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : complex, array-like, shape (n, n)
        The n*n symmetric matrix A.

    af : complex, array-like, shape (n, n)
        If `fact` = 'F', `af` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^H or `a` = LDL^H as computed by
        ``zsytrf``.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zsytrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    af : complex, ndarray, shape (n, n)
        If `fact` = 'N', `af` returns the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^H or `a` = LDL^H.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zsytrf``, as
        described above.

    x : complex, ndarray, shape (n, nrhs)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``zsysvx`` performs the following steps:

    (1) If `fact` = 'N', the diagonal pivoting method is used to factor
        A. The form of the factorization is A = UDU^T if `uplo` = 'U' or
        A = LDL^T if `uplo` = 'L', where U (or L) is a product of
        permutation and unit upper (lower) triangular matrices, and D is
        symmetric and block diagonal with 1*1 and 2*2 diagonal blocks.

    (#) If some d_{ii} = 0, so that D is exactly singular, then the
        function returns with errno = i. Otherwise, the factored form of
        A is used to estimate the condition number of the matrix A. If
        the reciprocal of the condition number is less than machine
        precision, `errno` = `n` + 1 is returned as a warning, but the
        function still goes on to solve for X and compute error bounds
        as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zsysvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zsysvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _a,
            _af,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _af_py = _af
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZsysvxReturnData(
        _af_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZsytrfReturnData = _collections.namedtuple(
    'ZsytrfReturnData',
    ['a', 'ipiv'],
)

def zsytrf(uplo, a):
    u"""
    Bunch--Kaufman factorization of complex symmetric matrix.

    ``zsytrf`` computes the Bunch--Kaufman factorization of a complex
    symmetric matrix.

    For full information please refer to the NAG Library document for
    f07nr

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07nrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        The n*n symmetric indefinite matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The upper or lower triangle of A is overwritten by details of
        the block diagonal matrix D and the multipliers used to obtain
        the factor U or L as specified by `uplo`.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    Notes
    -----
    ``zsytrf`` factorizes a complex symmetric matrix A, using the
    Bunch--Kaufman diagonal pivoting method. A is factorized as either A
    = PUDU^TP^T if `uplo` = 'U' or A = PLDL^TP^T if `uplo` = 'L', where
    P is a permutation matrix, U (or L) is a unit upper (or lower)
    triangular matrix and D is a symmetric block diagonal matrix with
    1*1 and 2*2 diagonal blocks; U (or L) has 2*2 unit diagonal blocks
    corresponding to the 2*2 blocks of D.
    Row and column interchanges are performed to ensure numerical
    stability while preserving symmetry.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsytrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zsytrf(
            uplo,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _ipiv_py = _ipiv
    _return_data = ZsytrfReturnData(
        _a_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zsytrs(uplo, a, ipiv, b):
    u"""
    Solution of complex symmetric system of linear equations, multiple
    right-hand sides, matrix already factorized by ``zsytrf``.

    ``zsytrs`` solves a complex symmetric system of linear equations
    with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zsytrf``.

    For full information please refer to the NAG Library document for
    f07ns

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07nsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``zsytrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zsytrf``.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zsytrs`` is used to solve a complex symmetric system of linear
    equations AX = B, this function must be preceded by a call to
    ``zsytrf`` which computes the Bunch--Kaufman factorization of A.

    If `uplo` = 'U', A = PUDU^TP^T, where P is a permutation matrix, U
    is an upper triangular matrix and D is a symmetric block diagonal
    matrix with 1*1 and 2*2 blocks; the solution X is computed by
    solving PUDY = B and then U^TP^TX = Y.

    If `uplo` = 'L', A = PLDL^TP^T, where L is a lower triangular
    matrix; the solution X is computed by solving PLDY = B and then
    L^TP^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsytrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zsytrs(
        uplo,
        _n,
        _nrhs,
        _a,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zsycon(uplo, a, ipiv, anorm):
    u"""
    Estimate condition number of complex symmetric matrix, matrix
    already factorized by ``zsytrf``.

    ``zsycon`` estimates the condition number of a complex symmetric
    matrix A, where A has been factorized by ``zsytrf``.

    For full information please refer to the NAG Library document for
    f07nu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07nuf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``zsytrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zsytrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlansy`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zsytrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -6)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``zsycon`` estimates the condition number (in the 1-norm) of a
    complex symmetric matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is symmetric, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.zlansy`` to
    compute ||A||_1 and a call to ``zsytrf`` to compute the
    Bunch--Kaufman factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zsycon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zsycon(
        uplo,
        _n,
        _a,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZsyrfsReturnData = _collections.namedtuple(
    'ZsyrfsReturnData',
    ['x', 'ferr', 'berr'],
)

def zsyrfs(uplo, n, a, af, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex symmetric system of
    linear equations, multiple right-hand sides.

    ``zsyrfs`` returns error bounds for the solution of a complex
    symmetric system of linear equations with multiple right-hand sides,
    AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07nv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07nvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n*n original symmetric matrix A as supplied to ``zsytrf``.

    af : complex, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``zsytrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zsytrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``zsytrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zsyrfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex symmetric system of
    linear equations with multiple right-hand sides AX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``zsyrfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsyrfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zsyrfs(
        uplo,
        n,
        _nrhs,
        _a,
        _af,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZsyrfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZsytriReturnData = _collections.namedtuple(
    'ZsytriReturnData',
    ['a'],
)

def zsytri(uplo, a, ipiv):
    u"""
    Inverse of complex symmetric matrix, matrix already factorized by
    ``zsytrf``.

    ``zsytri`` computes the inverse of a complex symmetric matrix A,
    where A has been factorized by ``zsytrf``.

    For full information please refer to the NAG Library document for
    f07nw

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07nwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``zsytrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zsytrf``.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The factorization is overwritten by the n*n symmetric matrix
        A^-1.

        If `uplo` = 'U', the upper triangle of A^-1 is stored in the
        upper triangular part of the array.

        If `uplo` = 'L', the lower triangle of A^-1 is stored in the
        lower triangular part of the array.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    Notes
    -----
    ``zsytri`` is used to compute the inverse of a complex symmetric
    matrix A, the function must be preceded by a call to ``zsytrf``,
    which computes the Bunch--Kaufman factorization of A.

    If `uplo` = 'U', A = PUDU^TP^T and A^-1 is computed by solving
    U^TP^TXPU = D^-1 for X.

    If `uplo` = 'L', A = PLDL^TP^T and A^-1 is computed by solving
    L^TP^TXPL = D^-1 for X.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zsytri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.zsytri(
            uplo,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = ZsytriReturnData(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

DspsvReturnData = _collections.namedtuple(
    'DspsvReturnData',
    ['ap', 'ipiv', 'b'],
)

def dspsv(uplo, ap, b):
    u"""
    Computes the solution to a real symmetric system of linear
    equations, packed storage.

    ``dspsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric matrix stored in packed format and X and
    B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07pa

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07paf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    ap : float, array-like, shape (n*(n+1)/2)
        The n*n symmetric matrix A, packed by columns.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    ap : float, ndarray, shape (n*(n+1)/2)
        The block diagonal matrix D and the multipliers used to obtain
        the factor U or L from the factorization A = UDU^T or A = LDL^T
        as computed by ``dsptrf``, stored as a packed triangular matrix
        in the same storage format as A.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    Notes
    -----
    ``dspsv`` uses the diagonal pivoting method to factor A as A = UDU^T
    if `uplo` = 'U' or A = LDL^T if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices, D
    is symmetric and block diagonal with 1*1 and 2*2 diagonal blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dspsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dspsv(
            uplo,
            _n,
            _nrhs,
            _ap,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = DspsvReturnData(
        _ap_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DspsvxReturnData = _collections.namedtuple(
    'DspsvxReturnData',
    ['afp', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
)

def dspsvx(fact, uplo, nrhs, ap, afp, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a real symmetric system of linear equations, packed storage. Error
    bounds and a condition estimate are also computed.

    ``dspsvx`` uses the diagonal pivoting factorization

        A = UDU^T or A = LDL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n*n symmetric matrix stored in packed format and X and
    B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07pb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07pbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `afp` and `ipiv` contain the factorized form of the matrix
            A. `afp` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afp` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ap : float, array-like, shape (n*(n+1)/2)
        The n*n symmetric matrix A, packed by columns.

    afp : float, array-like, shape (n*(n+1)/2)
        If `fact` = 'F', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^T or A = LDL^T as computed by ``dsptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``dsptrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : float, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    afp : float, ndarray, shape (n*(n+1)/2)
        If `fact` = 'N', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^T or A = LDL^T as computed by ``dsptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``dsptrf``, as
        described above.

    x : float, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``dspsvx`` performs the following steps:

    (1) If `fact` = 'N', the diagonal pivoting method is used to factor
        A as A = UDU^T if `uplo` = 'U' or A = LDL^T if `uplo` = 'L',
        where U (or L) is a product of permutation and unit upper
        (lower) triangular matrices and D is symmetric and block
        diagonal with 1*1 and 2*2 diagonal blocks.

    (#) If some d_{ii} = 0, so that D is exactly singular, then the
        function returns with errno = i. Otherwise, the factored form of
        A is used to estimate the condition number of the matrix A. If
        the reciprocal of the condition number is less than machine
        precision, `errno` = `n` + 1 is returned as a warning, but the
        function still goes on to solve for X and compute error bounds
        as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dspsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_ipiv, {'fun_name': fun_name, 'entity': 'ipiv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.dspsvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _ap,
            _afp,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _afp_py = _afp
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DspsvxReturnData(
        _afp_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DsptrfReturnData = _collections.namedtuple(
    'DsptrfReturnData',
    ['ap', 'ipiv'],
)

def dsptrf(uplo, n, ap):
    u"""
    Bunch--Kaufman factorization of real symmetric indefinite matrix,
    packed storage.

    ``dsptrf`` computes the Bunch--Kaufman factorization of a real
    symmetric indefinite matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07pd

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07pdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n*n symmetric matrix A, packed by columns.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by details of the block diagonal matrix D and
        the multipliers used to obtain the factor U or L as specified by
        `uplo`.

    ipiv : int, ndarray, shape (`n`)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    Notes
    -----
    ``dsptrf`` factorizes a real symmetric matrix A, using the
    Bunch--Kaufman diagonal pivoting method and packed storage. A is
    factorized as either A = PUDU^TP^T if `uplo` = 'U' or A = PLDL^TP^T
    if `uplo` = 'L', where P is a permutation matrix, U (or L) is a unit
    upper (or lower) triangular matrix and D is a symmetric block
    diagonal matrix with 1*1 and 2*2 diagonal blocks; U (or L) has 2*2
    unit diagonal blocks corresponding to the 2*2 blocks of D.
    Row and column interchanges are performed to ensure numerical
    stability while preserving symmetry.

    This method is suitable for symmetric matrices which are not known
    to be positive definite.
    If A is in fact positive definite, no interchanges are performed and
    no 2*2 blocks occur in D.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsptrf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dsptrf(
            uplo,
            _n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _ipiv_py = _ipiv
    _return_data = DsptrfReturnData(
        _ap_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dsptrs(uplo, n, ap, ipiv, b):
    u"""
    Solution of real symmetric indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by ``dsptrf``,
    packed storage.

    ``dsptrs`` solves a real symmetric indefinite system of linear
    equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``dsptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07pe

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07pef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``dsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsptrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (`n`, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dsptrs`` is used to solve a real symmetric indefinite system of
    linear equations AX = B, the function must be preceded by a call to
    ``dsptrf`` which computes the Bunch--Kaufman factorization of A,
    using packed storage.

    If `uplo` = 'U', A = PUDU^TP^T, where P is a permutation matrix, U
    is an upper triangular matrix and D is a symmetric block diagonal
    matrix with 1*1 and 2*2 blocks; the solution X is computed by
    solving PUDY = B and then U^TP^TX = Y.

    If `uplo` = 'L', A = PLDL^TP^T, where L is a lower triangular
    matrix; the solution X is computed by solving PLDY = B and then
    L^TP^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.dsptrs(
        uplo,
        n,
        _nrhs,
        _ap,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def dspcon(uplo, n, ap, ipiv, anorm):
    u"""
    Estimate condition number of real symmetric indefinite matrix,
    matrix already factorized by ``dsptrf``, packed storage.

    ``dspcon`` estimates the condition number of a real symmetric
    indefinite matrix A, where A has been factorized by ``dsptrf``,
    using packed storage.

    For full information please refer to the NAG Library document for
    f07pg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07pgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``dsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsptrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlansp`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dsptrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``dspcon`` estimates the condition number (in the 1-norm) of a real
    symmetric indefinite matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is symmetric, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.dlansp`` to
    compute ||A||_1 and a call to ``dsptrf`` to compute the
    Bunch--Kaufman factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dspcon'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.dspcon(
        uplo,
        n,
        _ap,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DsprfsReturnData = _collections.namedtuple(
    'DsprfsReturnData',
    ['x', 'ferr', 'berr'],
)

def dsprfs(uplo, n, ap, afp, ipiv, b, x):
    u"""
    Refined solution with error bounds of real symmetric indefinite
    system of linear equations, multiple right-hand sides, packed
    storage.

    ``dsprfs`` returns error bounds for the solution of a real symmetric
    indefinite system of linear equations with multiple right-hand
    sides, AX = B, using packed storage.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07ph

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07phf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n*n original symmetric matrix A as supplied to ``dsptrf``.

    afp : float, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``dsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsptrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``dsptrs``.

    Returns
    -------
    x : float, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dsprfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a real symmetric indefinite
    system of linear equations with multiple right-hand sides AX = B,
    using packed storage.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``dsprfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dsprfs(
        uplo,
        n,
        _nrhs,
        _ap,
        _afp,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DsprfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

DsptriReturnData = _collections.namedtuple(
    'DsptriReturnData',
    ['ap'],
)

def dsptri(uplo, n, ap, ipiv):
    u"""
    Inverse of real symmetric indefinite matrix, matrix already
    factorized by ``dsptrf``, packed storage.

    ``dsptri`` computes the inverse of a real symmetric indefinite
    matrix A, where A has been factorized by ``dsptrf``, using packed
    storage.

    For full information please refer to the NAG Library document for
    f07pj

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07pjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``dsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsptrf``.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n*n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    Notes
    -----
    ``dsptri`` is used to compute the inverse of a real symmetric
    indefinite matrix A, the function must be preceded by a call to
    ``dsptrf``, which computes the Bunch--Kaufman factorization of A,
    using packed storage.

    If `uplo` = 'U', A = PUDU^TP^T and A^-1 is computed by solving
    U^TP^TXPU = D^-1.

    If `uplo` = 'L', A = PLDL^TP^T and A^-1 is computed by solving
    L^TP^TXPL = D^-1.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dsptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.dsptri(
            uplo,
            n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = DsptriReturnData(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

ZhpsvReturnData = _collections.namedtuple(
    'ZhpsvReturnData',
    ['ap', 'ipiv', 'b'],
)

def zhpsv(uplo, ap, b):
    u"""
    Computes the solution to a complex Hermitian system of linear
    equations, packed storage.

    ``zhpsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n Hermitian matrix stored in packed format and X and
    B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07pn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07pnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n*n Hermitian matrix A, packed by columns.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    ap : complex, ndarray, shape (n*(n+1)/2)
        The block diagonal matrix D and the multipliers used to obtain
        the factor U or L from the factorization A = UDU^H or A = LDL^H
        as computed by ``zhptrf``, stored as a packed triangular matrix
        in the same storage format as A.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    Notes
    -----
    ``zhpsv`` uses the diagonal pivoting method to factor A as A = UDU^H
    if `uplo` = 'U' or A = LDL^H if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices, D
    is Hermitian and block diagonal with 1*1 and 2*2 diagonal blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zhpsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zhpsv(
            uplo,
            _n,
            _nrhs,
            _ap,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = ZhpsvReturnData(
        _ap_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZhpsvxReturnData = _collections.namedtuple(
    'ZhpsvxReturnData',
    ['afp', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
)

def zhpsvx(fact, uplo, nrhs, ap, afp, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a complex, Hermitian, system of linear equations, error bounds and
    condition estimates. Packed storage.

    ``zhpsvx`` uses the diagonal pivoting factorization

        A = UDU^H or A = LDL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n*n Hermitian matrix stored in packed format and X and
    B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07pp

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07ppf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `afp` and `ipiv` contain the factorized form of the matrix
            A. `afp` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afp` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n*n Hermitian matrix A, packed by columns.

    afp : complex, array-like, shape (n*(n+1)/2)
        If `fact` = 'F', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zhptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zhptrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : complex, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    afp : complex, ndarray, shape (n*(n+1)/2)
        If `fact` = 'N', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zhptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zhptrf``, as
        described above.

    x : complex, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``zhpsvx`` performs the following steps:

    (1) If `fact` = 'N', the diagonal pivoting method is used to factor
        A as A = UDU^H if `uplo` = 'U' or A = LDL^H if `uplo` = 'L',
        where U (or L) is a product of permutation and unit upper
        (lower) triangular matrices and D is Hermitian and block
        diagonal with 1*1 and 2*2 diagonal blocks.

    (#) If some d_{ii} = 0, so that D is exactly singular, then the
        function returns with errno = i. Otherwise, the factored form of
        A is used to estimate the condition number of the matrix A. If
        the reciprocal of the condition number is less than machine
        precision, `errno` = `n` + 1 is returned as a warning, but the
        function still goes on to solve for X and compute error bounds
        as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zhpsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_ipiv, {'fun_name': fun_name, 'entity': 'ipiv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zhpsvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _ap,
            _afp,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _afp_py = _afp
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZhpsvxReturnData(
        _afp_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZhptrfReturnData = _collections.namedtuple(
    'ZhptrfReturnData',
    ['ap', 'ipiv'],
)

def zhptrf(uplo, n, ap):
    u"""
    Bunch--Kaufman factorization of complex Hermitian indefinite matrix,
    packed storage.

    ``zhptrf`` computes the Bunch--Kaufman factorization of a complex
    Hermitian indefinite matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07pr

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07prf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n*n Hermitian matrix A, packed by columns.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by details of the block diagonal matrix D and
        the multipliers used to obtain the factor U or L as specified by
        `uplo`.

    ipiv : int, ndarray, shape (`n`)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    Notes
    -----
    ``zhptrf`` factorizes a complex Hermitian matrix A, using the
    Bunch--Kaufman diagonal pivoting method and packed storage. A is
    factorized as either A = PUDU^HP^T if `uplo` = 'U' or A = PLDL^HP^T
    if `uplo` = 'L', where P is a permutation matrix, U (or L) is a unit
    upper (or lower) triangular matrix and D is an Hermitian block
    diagonal matrix with 1*1 and 2*2 diagonal blocks; U (or L) has 2*2
    unit diagonal blocks corresponding to the 2*2 blocks of D.
    Row and column interchanges are performed to ensure numerical
    stability while keeping the matrix Hermitian.

    This method is suitable for Hermitian matrices which are not known
    to be positive definite.
    If A is in fact positive definite, no interchanges are performed and
    no 2*2 blocks occur in D.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhptrf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zhptrf(
            uplo,
            _n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _ipiv_py = _ipiv
    _return_data = ZhptrfReturnData(
        _ap_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zhptrs(uplo, n, ap, ipiv, b):
    u"""
    Solution of complex Hermitian indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by ``zhptrf``,
    packed storage.

    ``zhptrs`` solves a complex Hermitian indefinite system of linear
    equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zhptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07ps

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07psf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zhptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zhptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (`n`, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zhptrs`` is used to solve a complex Hermitian indefinite system of
    linear equations AX = B, the function must be preceded by a call to
    ``zhptrf`` which computes the Bunch--Kaufman factorization of A,
    using packed storage.

    If `uplo` = 'U', A = PUDU^HP^T, where P is a permutation matrix, U
    is an upper triangular matrix and D is an Hermitian block diagonal
    matrix with 1*1 and 2*2 blocks; the solution X is computed by
    solving PUDY = B and then U^HP^TX = Y.

    If `uplo` = 'L', A = PLDL^HP^T, where L is a lower triangular
    matrix; the solution X is computed by solving PLDY = B and then
    L^HP^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zhptrs(
        uplo,
        n,
        _nrhs,
        _ap,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zhpcon(uplo, ap, ipiv, anorm):
    u"""
    Estimate condition number of complex Hermitian indefinite matrix,
    matrix already factorized by ``zhptrf``, packed storage.

    ``zhpcon`` estimates the condition number of a complex Hermitian
    indefinite matrix A, where A has been factorized by ``zhptrf``,
    using packed storage.

    For full information please refer to the NAG Library document for
    f07pu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07puf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    ap : complex, array-like, shape (n*(n+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zhptrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zhptrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanhp`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zhptrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``zhpcon`` estimates the condition number (in the 1-norm) of a
    complex Hermitian indefinite matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is Hermitian, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.zlanhp`` to
    compute ||A||_1 and a call to ``zhptrf`` to compute the
    Bunch--Kaufman factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zhpcon'

    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_ipiv, {'fun_name': fun_name, 'entity': 'ipiv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zhpcon(
        uplo,
        _n,
        _ap,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZhprfsReturnData = _collections.namedtuple(
    'ZhprfsReturnData',
    ['x', 'ferr', 'berr'],
)

def zhprfs(uplo, n, ap, afp, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian indefinite
    system of linear equations, multiple right-hand sides, packed
    storage.

    ``zhprfs`` returns error bounds for the solution of a complex
    Hermitian indefinite system of linear equations with multiple
    right-hand sides, AX = B, using packed storage.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07pv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07pvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n*n original Hermitian matrix A as supplied to ``zhptrf``.

    afp : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zhptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zhptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``zhptrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zhprfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex Hermitian indefinite
    system of linear equations with multiple right-hand sides AX = B,
    using packed storage.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``zhprfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zhprfs(
        uplo,
        n,
        _nrhs,
        _ap,
        _afp,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZhprfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZhptriReturnData = _collections.namedtuple(
    'ZhptriReturnData',
    ['ap'],
)

def zhptri(uplo, n, ap, ipiv):
    u"""
    Inverse of complex Hermitian indefinite matrix, matrix already
    factorized by ``zhptrf``, packed storage.

    ``zhptri`` computes the inverse of a complex Hermitian indefinite
    matrix A, where A has been factorized by ``zhptrf``, using packed
    storage.

    For full information please refer to the NAG Library document for
    f07pw

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07pwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zhptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zhptrf``.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n*n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    Notes
    -----
    ``zhptri`` is used to compute the inverse of a complex Hermitian
    indefinite matrix A, the function must be preceded by a call to
    ``zhptrf``, which computes the Bunch--Kaufman factorization of A,
    using packed storage.

    If `uplo` = 'U', A = PUDU^HP^T and A^-1 is computed by solving
    U^HP^TXPU = D^-1 for X.

    If `uplo` = 'L', A = PLDL^HP^T and A^-1 is computed by solving
    L^HP^TXPL = D^-1 for X.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zhptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.zhptri(
            uplo,
            n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = ZhptriReturnData(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

ZspsvReturnData = _collections.namedtuple(
    'ZspsvReturnData',
    ['ap', 'ipiv', 'b'],
)

def zspsv(uplo, ap, b):
    u"""
    Computes the solution to a complex symmetric system of linear
    equations, packed storage.

    ``zspsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n*n symmetric matrix stored in packed format and X and
    B are n*r matrices.

    For full information please refer to the NAG Library document for
    f07qn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07qnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n*n symmetric matrix A, packed by columns.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    ap : complex, ndarray, shape (n*(n+1)/2)
        The block diagonal matrix D and the multipliers used to obtain
        the factor U or L from the factorization A = UDU^H or A = LDL^H
        as computed by ``zsptrf``, stored as a packed triangular matrix
        in the same storage format as A.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    Notes
    -----
    ``zspsv`` uses the diagonal pivoting method to factor A as A = UDU^T
    if `uplo` = 'U' or A = LDL^T if `uplo` = 'L', where U (or L) is a
    product of permutation and unit upper (lower) triangular matrices, D
    is symmetric and block diagonal with 1*1 and 2*2 diagonal blocks.
    The factored form of A is then used to solve the system of equations
    AX = B.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zspsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zspsv(
            uplo,
            _n,
            _nrhs,
            _ap,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = ZspsvReturnData(
        _ap_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZspsvxReturnData = _collections.namedtuple(
    'ZspsvxReturnData',
    ['afp', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
)

def zspsvx(fact, uplo, nrhs, ap, afp, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a complex, symmetric, system of linear equations, error bounds and
    condition estimates. Packed storage.

    ``zspsvx`` uses the diagonal pivoting factorization

        A = UDU^T or A = LDL^T

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n*n symmetric matrix stored in packed format and X and
    B are n*r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07qp

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07qpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `afp` and `ipiv` contain the factorized form of the matrix
            A. `afp` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afp` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n*n symmetric matrix A, packed by columns.

    afp : complex, array-like, shape (n*(n+1)/2)
        If `fact` = 'F', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zsptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zsptrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    b : complex, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    Returns
    -------
    afp : complex, ndarray, shape (n*(n+1)/2)
        If `fact` = 'N', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zsptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zsptrf``, as
        described above.

    x : complex, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n*r
        solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    Notes
    -----
    ``zspsvx`` performs the following steps:

    (1) If `fact` = 'N', the diagonal pivoting method is used to factor
        A as A = UDU^T if `uplo` = 'U' or A = LDL^T if `uplo` = 'L',
        where U (or L) is a product of permutation and unit upper
        (lower) triangular matrices and D is symmetric and block
        diagonal with 1*1 and 2*2 diagonal blocks.

    (#) If some d_{ii} = 0, so that D is exactly singular, then the
        function returns with errno = i. Otherwise, the factored form of
        A is used to estimate the condition number of the matrix A. If
        the reciprocal of the condition number is less than machine
        precision, `errno` = `n` + 1 is returned as a warning, but the
        function still goes on to solve for X and compute error bounds
        as described below.

    (#) The system of equations is solved for X using the factored form
        of A.

    (#) Iterative refinement is applied to improve the computed solution
        matrix and to calculate error bounds and backward error
        estimates for it.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zspsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_ipiv, {'fun_name': fun_name, 'entity': 'ipiv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zspsvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _ap,
            _afp,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _afp_py = _afp
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZspsvxReturnData(
        _afp_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ZsptrfReturnData = _collections.namedtuple(
    'ZsptrfReturnData',
    ['ap', 'ipiv'],
)

def zsptrf(uplo, n, ap):
    u"""
    Bunch--Kaufman factorization of complex symmetric matrix, packed
    storage.

    ``zsptrf`` computes the Bunch--Kaufman factorization of a complex
    symmetric matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07qr

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07qrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n*n symmetric matrix A, packed by columns.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by details of the block diagonal matrix D and
        the multipliers used to obtain the factor U or L as specified by
        `uplo`.

    ipiv : int, ndarray, shape (`n`)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1*1 pivot block and the
            ith row and column of A were interchanged with the kth row
            and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2*2 pivot block and the (i-1)th row and
            column of A were interchanged with the lth row and column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2*2 pivot block and the (i+1)th row and column
            of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    Notes
    -----
    ``zsptrf`` factorizes a complex symmetric matrix A, using the
    Bunch--Kaufman diagonal pivoting method and packed storage. A is
    factorized as either A = PUDU^TP^T if `uplo` = 'U' or A = PLDL^TP^T
    if `uplo` = 'L', where P is a permutation matrix, U (or L) is a unit
    upper (or lower) triangular matrix and D is a symmetric block
    diagonal matrix with 1*1 and 2*2 diagonal blocks; U (or L) has 2*2
    unit diagonal blocks corresponding to the 2*2 blocks of D.
    Row and column interchanges are performed to ensure numerical
    stability while preserving symmetry.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsptrf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zsptrf(
            uplo,
            _n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _ipiv_py = _ipiv
    _return_data = ZsptrfReturnData(
        _ap_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zsptrs(uplo, n, ap, ipiv, b):
    u"""
    Solution of complex symmetric system of linear equations, multiple
    right-hand sides, matrix already factorized by ``zsptrf``, packed
    storage.

    ``zsptrs`` solves a complex symmetric system of linear equations
    with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zsptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07qs

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07qsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zsptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (`n`, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zsptrs`` is used to solve a complex symmetric system of linear
    equations AX = B, the function must be preceded by a call to
    ``zsptrf`` which computes the Bunch--Kaufman factorization of A,
    using packed storage.

    If `uplo` = 'U', A = PUDU^TP^T, where P is a permutation matrix, U
    is an upper triangular matrix and D is a symmetric block diagonal
    matrix with 1*1 and 2*2 blocks; the solution X is computed by
    solving PUDY = B and then U^TP^TX = Y.

    If `uplo` = 'L', A = PLDL^TP^T, where L is a lower triangular
    matrix; the solution X is computed by solving PLDY = B and then
    L^TP^TX = Y.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zsptrs(
        uplo,
        n,
        _nrhs,
        _ap,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zspcon(uplo, ap, ipiv, anorm):
    u"""
    Estimate condition number of complex symmetric matrix, matrix
    already factorized by ``zsptrf``, packed storage.

    ``zspcon`` estimates the condition number of a complex symmetric
    matrix A, where A has been factorized by ``zsptrf``, using packed
    storage.

    For full information please refer to the NAG Library document for
    f07qu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07quf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    ap : complex, array-like, shape (n*(n+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zsptrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zsptrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlansp`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zsptrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    Notes
    -----
    ``zspcon`` estimates the condition number (in the 1-norm) of a
    complex symmetric matrix A:

        kappa_1(A) = ||A||_1||A^-1||_1.

    Since A is symmetric, kappa_1(A) = kappa_infinity (A) =
    ||A||_infinity||A^-1||_infinity.

    Because kappa_1(A) is infinite if A is singular, the function
    actually returns an estimate of the **reciprocal** of kappa_1(A).

    The function should be preceded by a call to ``blas.zlansp`` to
    compute ||A||_1 and a call to ``zsptrf`` to compute the
    Bunch--Kaufman factorization of A.
    The function then uses Higham's implementation of Hager's method
    (see Higham (1988)) to estimate ||A^-1||_1.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zspcon'

    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_ipiv, {'fun_name': fun_name, 'entity': 'ipiv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zspcon(
        uplo,
        _n,
        _ap,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZsprfsReturnData = _collections.namedtuple(
    'ZsprfsReturnData',
    ['x', 'ferr', 'berr'],
)

def zsprfs(uplo, n, ap, afp, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex symmetric system of
    linear equations, multiple right-hand sides, packed storage.

    ``zsprfs`` returns error bounds for the solution of a complex
    symmetric system of linear equations with multiple right-hand sides,
    AX = B, using packed storage.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07qv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07qvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n*n original symmetric matrix A as supplied to ``zsptrf``.

    afp : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zsptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``zsptrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zsprfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex symmetric system of
    linear equations with multiple right-hand sides AX = B, using packed
    storage.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``zsprfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zsprfs(
        uplo,
        n,
        _nrhs,
        _ap,
        _afp,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZsprfsReturnData(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZsptriReturnData = _collections.namedtuple(
    'ZsptriReturnData',
    ['ap'],
)

def zsptri(uplo, n, ap, ipiv):
    u"""
    Inverse of complex symmetric matrix, matrix already factorized by
    ``zsptrf``, packed storage.

    ``zsptri`` computes the inverse of a complex symmetric matrix A,
    where A has been factorized by ``zsptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07qw

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07qwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zsptrf``.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n*n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    Notes
    -----
    ``zsptri`` is used to compute the inverse of a complex symmetric
    matrix A, the function must be preceded by a call to ``zsptrf``,
    which computes the Bunch--Kaufman factorization of A, using packed
    storage.

    If `uplo` = 'U', A = PUDU^TP^T and A^-1 is computed by solving
    U^TP^TXPU = D^-1.

    If `uplo` = 'L', A = PLDL^TP^T and A^-1 is computed by solving
    L^TP^TXPL = D^-1.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zsptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.zsptri(
            uplo,
            n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = ZsptriReturnData(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

DtrtrsReturnData = _collections.namedtuple(
    'DtrtrsReturnData',
    ['b'],
)

def dtrtrs(uplo, trans, diag, a, b):
    u"""
    Solution of real triangular system of linear equations, multiple
    right-hand sides.

    ``dtrtrs`` solves a real triangular system of linear equations with
    multiple right-hand sides, AX = B or A^TX = B.

    For full information please refer to the NAG Library document for
    f07te

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07tef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    a : float, array-like, shape (n, n)
        The n*n triangular matrix A.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    Notes
    -----
    ``dtrtrs`` solves a real triangular system of linear equations AX =
    B or A^TX = B.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.dtrtrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    try:
        _b_lapacklin.dtrtrs(
            uplo,
            trans,
            diag,
            _n,
            _nrhs,
            _a,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = DtrtrsReturnData(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def dtrcon(norm, uplo, diag, n, a):
    u"""
    Estimate condition number of real triangular matrix.

    ``dtrcon`` estimates the condition number of a real triangular
    matrix.

    For full information please refer to the NAG Library document for
    f07tg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07tgf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n*n triangular matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or if
        the estimate underflows. If `rcond` is less than machine
        precision, A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    ``dtrcon`` estimates the condition number of a real triangular
    matrix A, in either the 1-norm or the infinity-norm:

        kappa_1(A) = ||A||_1||A^-1||_1 or kappa_infinity (A) =
        ||A||_infinity||A^-1||_infinity.

    Note that kappa_infinity (A) = kappa_1(A^T).

    Because the condition number is infinite if A is singular, the
    function actually returns an estimate of the **reciprocal** of the
    condition number.

    The function computes ||A||_1 or ||A||_infinity exactly, and uses
    Higham's implementation of Hager's method (see Higham (1988)) to
    estimate ||A^-1||_1 or ||A^-1||_infinity.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dtrcon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dtrcon(
        norm,
        uplo,
        diag,
        n,
        _a,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DtrrfsReturnData = _collections.namedtuple(
    'DtrrfsReturnData',
    ['ferr', 'berr'],
)

def dtrrfs(uplo, trans, diag, n, a, b, x):
    u"""
    Error bounds for solution of real triangular system of linear
    equations, multiple right-hand sides.

    ``dtrrfs`` returns error bounds for the solution of a real
    triangular system of linear equations with multiple right-hand
    sides, AX = B or A^TX = B.

    For full information please refer to the NAG Library document for
    f07th

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07thf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n*n triangular matrix A.

    b : float, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``dtrtrs``.

    Returns
    -------
    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dtrrfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a real triangular system of
    linear equations with multiple right-hand sides AX = B or A^TX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``dtrrfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dtrrfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dtrrfs(
        uplo,
        trans,
        diag,
        n,
        _nrhs,
        _a,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DtrrfsReturnData(
        _ferr_py,
        _berr_py,
    )
    return _return_data

DtrtriReturnData = _collections.namedtuple(
    'DtrtriReturnData',
    ['a'],
)

def dtrtri(uplo, diag, a):
    u"""
    Inverse of real triangular matrix.

    ``dtrtri`` computes the inverse of a real triangular matrix.

    For full information please refer to the NAG Library document for
    f07tj

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07tjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    a : float, array-like, shape (n, n)
        The n*n triangular matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        A is overwritten by A^-1, using the same storage format as
        described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular its inverse cannot be computed.

    Notes
    -----
    ``dtrtri`` forms the inverse of a real triangular matrix A.
    Note that the inverse of an upper (lower) triangular matrix is also
    upper (lower) triangular.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dtrtri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_lapacklin.dtrtri(
            uplo,
            diag,
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = DtrtriReturnData(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

ZtrtrsReturnData = _collections.namedtuple(
    'ZtrtrsReturnData',
    ['b'],
)

def ztrtrs(uplo, trans, diag, a, b):
    u"""
    Solution of complex triangular system of linear equations, multiple
    right-hand sides.

    ``ztrtrs`` solves a complex triangular system of linear equations
    with multiple right-hand sides, AX = B, A^TX = B or A^HX = B.

    For full information please refer to the NAG Library document for
    f07ts

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07tsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    a : complex, array-like, shape (n, n)
        The n*n triangular matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    Notes
    -----
    ``ztrtrs`` solves a complex triangular system of linear equations AX
    = B, A^TX = B or A^HX = B.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.ztrtrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    try:
        _b_lapacklin.ztrtrs(
            uplo,
            trans,
            diag,
            _n,
            _nrhs,
            _a,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = ZtrtrsReturnData(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def ztrcon(norm, uplo, diag, n, a):
    u"""
    Estimate condition number of complex triangular matrix.

    ``ztrcon`` estimates the condition number of a complex triangular
    matrix.

    For full information please refer to the NAG Library document for
    f07tu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07tuf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n*n triangular matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    ``ztrcon`` estimates the condition number of a complex triangular
    matrix A, in either the 1-norm or the infinity-norm:

        kappa_1(A) = ||A||_1||A^-1||_1 or kappa_infinity (A) =
        ||A||_infinity||A^-1||_infinity.

    Note that kappa_infinity (A) = kappa_1(A^T).

    Because the condition number is infinite if A is singular, the
    function actually returns an estimate of the **reciprocal** of the
    condition number.

    The function computes ||A||_1 or ||A||_infinity exactly, and uses
    Higham's implementation of Hager's method (see Higham (1988)) to
    estimate ||A^-1||_1 or ||A^-1||_infinity.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.ztrcon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _return_scalar = _b_lapacklin.ztrcon(
        norm,
        uplo,
        diag,
        n,
        _a,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZtrrfsReturnData = _collections.namedtuple(
    'ZtrrfsReturnData',
    ['ferr', 'berr'],
)

def ztrrfs(uplo, trans, diag, n, a, b, x):
    u"""
    Error bounds for solution of complex triangular system of linear
    equations, multiple right-hand sides.

    ``ztrrfs`` returns error bounds for the solution of a complex
    triangular system of linear equations with multiple right-hand
    sides, AX = B, A^TX = B or A^HX = B.

    For full information please refer to the NAG Library document for
    f07tv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07tvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n*n triangular matrix A.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``ztrtrs``.

    Returns
    -------
    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``ztrrfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex triangular system of
    linear equations with multiple right-hand sides AX = B, A^TX = B or
    A^HX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``ztrrfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.ztrrfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.ztrrfs(
        uplo,
        trans,
        diag,
        n,
        _nrhs,
        _a,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZtrrfsReturnData(
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZtrtriReturnData = _collections.namedtuple(
    'ZtrtriReturnData',
    ['a'],
)

def ztrtri(uplo, diag, a):
    u"""
    Inverse of complex triangular matrix.

    ``ztrtri`` computes the inverse of a complex triangular matrix.

    For full information please refer to the NAG Library document for
    f07tw

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07twf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    a : complex, array-like, shape (n, n)
        The n*n triangular matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        A is overwritten by A^-1, using the same storage format as
        described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular its inverse cannot be computed.

    Notes
    -----
    ``ztrtri`` forms the inverse of a complex triangular matrix A.
    Note that the inverse of an upper (lower) triangular matrix is also
    upper (lower) triangular.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.ztrtri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_lapacklin.ztrtri(
            uplo,
            diag,
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = ZtrtriReturnData(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

DtptrsReturnData = _collections.namedtuple(
    'DtptrsReturnData',
    ['b'],
)

def dtptrs(uplo, trans, diag, n, ap, b):
    u"""
    Solution of real triangular system of linear equations, multiple
    right-hand sides, packed storage.

    ``dtptrs`` solves a real triangular system of linear equations with
    multiple right-hand sides, AX = B or A^TX = B, using packed storage.

    For full information please refer to the NAG Library document for
    f07ue

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07uef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n*n triangular matrix A, packed by columns.

    b : float, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (`n`, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    Notes
    -----
    ``dtptrs`` solves a real triangular system of linear equations AX =
    B or A^TX = B, using packed storage.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.dtptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dtptrs(
            uplo,
            trans,
            diag,
            n,
            _nrhs,
            _ap,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = DtptrsReturnData(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def dtpcon(norm, uplo, diag, n, ap):
    u"""
    Estimate condition number of real triangular matrix, packed storage.

    ``dtpcon`` estimates the condition number of a real triangular
    matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07ug

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07ugf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n*n triangular matrix A, packed by columns.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    ``dtpcon`` estimates the condition number of a real triangular
    matrix A, in either the 1-norm or the infinity-norm, using packed
    storage:

        kappa_1(A) = ||A||_1||A^-1||_1 or kappa_infinity (A) =
        ||A||_infinity||A^-1||_infinity.

    Note that kappa_infinity (A) = kappa_1(A^T).

    Because the condition number is infinite if A is singular, the
    function actually returns an estimate of the **reciprocal** of the
    condition number.

    The function computes ||A||_1 or ||A||_infinity exactly, and uses
    Higham's implementation of Hager's method (see Higham (1988)) to
    estimate ||A^-1||_1 or ||A^-1||_infinity.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dtpcon'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dtpcon(
        norm,
        uplo,
        diag,
        n,
        _ap,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DtprfsReturnData = _collections.namedtuple(
    'DtprfsReturnData',
    ['ferr', 'berr'],
)

def dtprfs(uplo, trans, diag, n, ap, b, x):
    u"""
    Error bounds for solution of real triangular system of linear
    equations, multiple right-hand sides, packed storage.

    ``dtprfs`` returns error bounds for the solution of a real
    triangular system of linear equations with multiple right-hand
    sides, AX = B or A^TX = B, using packed storage.

    For full information please refer to the NAG Library document for
    f07uh

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07uhf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n*n triangular matrix A, packed by columns.

    b : float, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``dtptrs``.

    Returns
    -------
    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dtprfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a real triangular system of
    linear equations with multiple right-hand sides AX = B or A^TX = B,
    using packed storage.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``dtprfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dtprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dtprfs(
        uplo,
        trans,
        diag,
        n,
        _nrhs,
        _ap,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DtprfsReturnData(
        _ferr_py,
        _berr_py,
    )
    return _return_data

DtptriReturnData = _collections.namedtuple(
    'DtptriReturnData',
    ['ap'],
)

def dtptri(uplo, diag, n, ap):
    u"""
    Inverse of real triangular matrix, packed storage.

    ``dtptri`` computes the inverse of a real triangular matrix, using
    packed storage.

    For full information please refer to the NAG Library document for
    f07uj

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07ujf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n*n triangular matrix A, packed by columns.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by A^-1, using the same storage format as
        described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular its inverse cannot be computed.

    Notes
    -----
    ``dtptri`` forms the inverse of a real triangular matrix A, using
    packed storage.
    Note that the inverse of an upper (lower) triangular matrix is also
    upper (lower) triangular.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dtptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dtptri(
            uplo,
            diag,
            n,
            _ap,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = DtptriReturnData(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

ZtptrsReturnData = _collections.namedtuple(
    'ZtptrsReturnData',
    ['b'],
)

def ztptrs(uplo, trans, diag, n, ap, b):
    u"""
    Solution of complex triangular system of linear equations, multiple
    right-hand sides, packed storage.

    ``ztptrs`` solves a complex triangular system of linear equations
    with multiple right-hand sides, AX = B, A^TX = B or A^HX = B, using
    packed storage.

    For full information please refer to the NAG Library document for
    f07us

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07usf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n*n triangular matrix A, packed by columns.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (`n`, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    Notes
    -----
    ``ztptrs`` solves a complex triangular system of linear equations AX
    = B, A^TX = B or A^HX = B, using packed storage.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.ztptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    try:
        _b_lapacklin.ztptrs(
            uplo,
            trans,
            diag,
            n,
            _nrhs,
            _ap,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = ZtptrsReturnData(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def ztpcon(norm, uplo, diag, n, ap):
    u"""
    Estimate condition number of complex triangular matrix, packed
    storage.

    ``ztpcon`` estimates the condition number of a complex triangular
    matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07uu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07uuf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n*n triangular matrix A, packed by columns.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Notes
    -----
    ``ztpcon`` estimates the condition number of a complex triangular
    matrix A, in either the 1-norm or the infinity-norm, using packed
    storage:

        kappa_1(A) = ||A||_1||A^-1||_1 or kappa_infinity (A) =
        ||A||_infinity||A^-1||_infinity.

    Note that kappa_infinity (A) = kappa_1(A^T).

    Because the condition number is infinite if A is singular, the
    function actually returns an estimate of the **reciprocal** of the
    condition number.

    The function computes ||A||_1 or ||A||_infinity exactly, and uses
    Higham's implementation of Hager's method (see Higham (1988)) to
    estimate ||A^-1||_1 or ||A^-1||_infinity.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.ztpcon'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _return_scalar = _b_lapacklin.ztpcon(
        norm,
        uplo,
        diag,
        n,
        _ap,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZtprfsReturnData = _collections.namedtuple(
    'ZtprfsReturnData',
    ['ferr', 'berr'],
)

def ztprfs(uplo, trans, diag, n, ap, b, x):
    u"""
    Error bounds for solution of complex triangular system of linear
    equations, multiple right-hand sides, packed storage.

    ``ztprfs`` returns error bounds for the solution of a complex
    triangular system of linear equations with multiple right-hand
    sides, AX = B, A^TX = B or A^HX = B, using packed storage.

    For full information please refer to the NAG Library document for
    f07uv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07uvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n*n triangular matrix A, packed by columns.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n*r solution matrix X, as returned by ``ztptrs``.

    Returns
    -------
    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``ztprfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex triangular system of
    linear equations with multiple right-hand sides AX = B, A^TX = B or
    A^HX = B, using packed storage.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``ztprfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise backward error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise forward
    error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.ztprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.ztprfs(
        uplo,
        trans,
        diag,
        n,
        _nrhs,
        _ap,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZtprfsReturnData(
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZtptriReturnData = _collections.namedtuple(
    'ZtptriReturnData',
    ['ap'],
)

def ztptri(uplo, diag, n, ap):
    u"""
    Inverse of complex triangular matrix, packed storage.

    ``ztptri`` computes the inverse of a complex triangular matrix,
    using packed storage.

    For full information please refer to the NAG Library document for
    f07uw

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07uwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n*n triangular matrix A, packed by columns.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by A^-1, using the same storage format as
        described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular its inverse cannot be computed.

    Notes
    -----
    ``ztptri`` forms the inverse of a complex triangular matrix A, using
    packed storage.
    Note that the inverse of an upper (lower) triangular matrix is also
    upper (lower) triangular.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.ztptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    try:
        _b_lapacklin.ztptri(
            uplo,
            diag,
            n,
            _ap,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = ZtptriReturnData(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

DtbtrsReturnData = _collections.namedtuple(
    'DtbtrsReturnData',
    ['b'],
)

def dtbtrs(uplo, trans, diag, kd, ab, b):
    u"""
    Solution of real band triangular system of linear equations,
    multiple right-hand sides.

    ``dtbtrs`` solves a real triangular band system of linear equations
    with multiple right-hand sides, AX = B or A^TX = B.

    For full information please refer to the NAG Library document for
    f07ve

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07vef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : float, array-like, shape (`kd`+1, n)
        The n*n triangular band matrix A.

    b : float, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    Notes
    -----
    ``dtbtrs`` solves a real triangular band system of linear equations
    AX = B or A^TX = B.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.dtbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    try:
        _b_lapacklin.dtbtrs(
            uplo,
            trans,
            diag,
            _n,
            kd,
            _nrhs,
            _ab,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = DtbtrsReturnData(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def dtbcon(norm, uplo, diag, kd, ab):
    u"""
    Estimate condition number of real band triangular matrix.

    ``dtbcon`` estimates the condition number of a real triangular band
    matrix.

    For full information please refer to the NAG Library document for
    f07vg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07vgf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : float, array-like, shape (`kd`+1, n)
        The n*n triangular band matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

    Notes
    -----
    ``dtbcon`` estimates the condition number of a real triangular band
    matrix A, in either the 1-norm or the infinity-norm:

        kappa_1(A) = ||A||_1||A^-1||_1 or kappa_infinity (A) =
        ||A||_infinity||A^-1||_infinity.

    Note that kappa_infinity (A) = kappa_1(A^T).

    Because the condition number is infinite if A is singular, the
    function actually returns an estimate of the **reciprocal** of the
    condition number.

    The function computes ||A||_1 or ||A||_infinity exactly, and uses
    Higham's implementation of Hager's method (see Higham (1988)) to
    estimate ||A^-1||_1 or ||A^-1||_infinity.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dtbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_lapacklin.dtbcon(
        norm,
        uplo,
        diag,
        _n,
        kd,
        _ab,
    )
    _rcond_py = _return_scalar
    return _rcond_py

DtbrfsReturnData = _collections.namedtuple(
    'DtbrfsReturnData',
    ['ferr', 'berr'],
)

def dtbrfs(uplo, trans, diag, kd, nrhs, ab, b, x):
    u"""
    Error bounds for solution of real band triangular system of linear
    equations, multiple right-hand sides.

    ``dtbrfs`` returns error bounds for the solution of a real
    triangular band system of linear equations with multiple right-hand
    sides, AX = B or A^TX = B.

    For full information please refer to the NAG Library document for
    f07vh

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07vhf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    nrhs : int
        r, the number of right-hand sides.

    ab : float, array-like, shape (`kd`+1, n)
        The n*n triangular band matrix A.

    b : float, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    x : float, array-like, shape (n, `nrhs`)
        The n*r solution matrix X, as returned by ``dtbtrs``.

    Returns
    -------
    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Notes
    -----
    ``dtbrfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a real triangular band system of
    linear equations with multiple right-hand sides AX = B or A^TX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``dtbrfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise` `backward` `error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise`
    `forward` `error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dtbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dtbrfs(
        uplo,
        trans,
        diag,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = DtbrfsReturnData(
        _ferr_py,
        _berr_py,
    )
    return _return_data

ZtbtrsReturnData = _collections.namedtuple(
    'ZtbtrsReturnData',
    ['b'],
)

def ztbtrs(uplo, trans, diag, kd, ab, b):
    u"""
    Solution of complex band triangular system of linear equations,
    multiple right-hand sides.

    ``ztbtrs`` solves a complex triangular band system of linear
    equations with multiple right-hand sides, AX = B, A^TX = B or A^HX =
    B.

    For full information please refer to the NAG Library document for
    f07vs

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07vsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : complex, array-like, shape (`kd`+1, n)
        The n*n triangular band matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    Notes
    -----
    ``ztbtrs`` solves a complex triangular band system of linear
    equations AX = B, A^TX = B or A^HX = B.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.ztbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    try:
        _b_lapacklin.ztbtrs(
            uplo,
            trans,
            diag,
            _n,
            kd,
            _nrhs,
            _ab,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = ZtbtrsReturnData(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def ztbcon(norm, uplo, diag, kd, ab):
    u"""
    Estimate condition number of complex band triangular matrix.

    ``ztbcon`` estimates the condition number of a complex triangular
    band matrix.

    For full information please refer to the NAG Library document for
    f07vu

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07vuf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : complex, array-like, shape (`kd`+1, n)
        The n*n triangular band matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

    Notes
    -----
    ``ztbcon`` estimates the condition number of a complex triangular
    band matrix A, in either the 1-norm or the infinity-norm:

        kappa_1(A) = ||A||_1||A^-1||_1 or kappa_infinity (A) =
        ||A||_infinity||A^-1||_infinity.

    Note that kappa_infinity (A) = kappa_1(A^T).

    Because the condition number is infinite if A is singular, the
    function actually returns an estimate of the **reciprocal** of the
    condition number.

    The function computes ||A||_1 or ||A||_infinity exactly, and uses
    Higham's implementation of Hager's method (see Higham (1988)) to
    estimate ||A^-1||_1 or ||A^-1||_infinity.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.ztbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_lapacklin.ztbcon(
        norm,
        uplo,
        diag,
        _n,
        kd,
        _ab,
    )
    _rcond_py = _return_scalar
    return _rcond_py

ZtbrfsReturnData = _collections.namedtuple(
    'ZtbrfsReturnData',
    ['ferr', 'berr'],
)

def ztbrfs(uplo, trans, diag, kd, nrhs, ab, b, x):
    u"""
    Error bounds for solution of complex band triangular system of
    linear equations, multiple right-hand sides.

    ``ztbrfs`` returns error bounds for the solution of a complex
    triangular band system of linear equations with multiple right-hand
    sides, AX = B, A^TX = B or A^HX = B.

    For full information please refer to the NAG Library document for
    f07vv

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07vvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    nrhs : int
        r, the number of right-hand sides.

    ab : complex, array-like, shape (`kd`+1, n)
        The n*n triangular band matrix A.

    b : complex, array-like, shape (n, `nrhs`)
        The n*r right-hand side matrix B.

    x : complex, array-like, shape (n, `nrhs`)
        The n*r solution matrix X, as returned by ``ztbtrs``.

    Returns
    -------
    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Notes
    -----
    ``ztbrfs`` returns the backward errors and estimated bounds on the
    forward errors for the solution of a complex triangular band system
    of linear equations with multiple right-hand sides AX = B, A^TX = B
    or A^HX = B.
    The function handles each right-hand side vector (stored as a column
    of the matrix B) independently, so we describe the function of
    ``ztbrfs`` in terms of a single right-hand side b and solution x.

    Given a computed solution x, the function computes the
    `component-wise` `backward` `error` beta.
    This is the size of the smallest relative perturbation in each
    element of A and b such that x is the exact solution of a perturbed
    system

        (A+delta A)x = b+delta b;|delta a_{ij}| <= beta|a_{ij}| and
        |delta b_i| <= beta|b_i| .

    Then the function estimates a bound for the `component-wise`
    `forward` `error` in the computed solution, defined by:

        max_i|x_i-x^_i|/max_i|x_i|

    where x^ is the true solution.

    For details of the method, see [the F07 Introduction].

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.ztbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.ztbrfs(
        uplo,
        trans,
        diag,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = ZtbrfsReturnData(
        _ferr_py,
        _berr_py,
    )
    return _return_data

DpftrfReturnData = _collections.namedtuple(
    'DpftrfReturnData',
    ['ar'],
)

def dpftrf(transr, uplo, n, ar):
    u"""
    Cholesky factorization of real symmetric positive definite matrix,
    Rectangular Full Packed format.

    ``dpftrf`` computes the Cholesky factorization of a real symmetric
    positive definite matrix stored in Rectangular Full Packed (RFP)
    format.

    For full information please refer to the NAG Library document for
    f07wd

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07wdf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'T'

            The matrix A is stored in transposed RFP format.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

            The upper triangular part of A is stored, and A is
            factorized as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored, and A is
            factorized as LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : float, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n*n symmetric matrix A, in either normal or transposed RFP
        format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    ar : float, ndarray, shape (`n`*(`n`+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^TU or A = LL^T, in the same storage
        format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to factorize a symmetric matrix stored
            in RFP format which is not positive definite; the matrix
            must be treated as a full symmetric matrix, by calling
            ``dsytrf``.

    Notes
    -----
    ``dpftrf`` forms the Cholesky factorization of a real symmetric
    positive definite matrix A either as A = U^TU if `uplo` = 'U' or A =
    LL^T if `uplo` = 'L', where U is an upper triangular matrix and L is
    a lower triangular, stored in RFP format. The RFP storage format is
    described in [the F07 Introduction].

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.dpftrf'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dpftrf(
            transr,
            uplo,
            n,
            _ar,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ar_py = _ar
    _return_data = DpftrfReturnData(
        _ar_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ar_py

def dpftrs(transr, uplo, n, ar, b):
    u"""
    Solution of real symmetric positive definite system of linear
    equations, multiple right-hand sides, coefficient matrix already
    factorized by ``dpftrf``, Rectangular Full Packed format.

    ``dpftrs`` solves a real symmetric positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    using the Cholesky factorization computed by ``dpftrf`` stored in
    Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07we

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07wef.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'T'

            The matrix A is stored in transposed RFP format.

    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factorization of A stored in RFP format, as
        returned by ``dpftrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (`n`, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``dpftrs`` is used to solve a real symmetric positive definite
    system of linear equations AX = B, the function must be preceded by
    a call to ``dpftrf`` which computes the Cholesky factorization of A,
    stored in RFP format. The RFP storage format is described in [the
    F07 Introduction]. The solution X is computed by forward and
    backward substitution.

    If `uplo` = 'U', A = U^TU, where U is upper triangular; the solution
    X is computed by solving U^TY = B and then UX = Y.

    If `uplo` = 'L', A = LL^T, where L is lower triangular; the solution
    X is computed by solving LY = B and then L^TX = Y.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.dpftrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dpftrs(
        transr,
        uplo,
        n,
        _nrhs,
        _ar,
        _b,
    )
    _b_py = _b
    return _b_py

DpftriReturnData = _collections.namedtuple(
    'DpftriReturnData',
    ['ar'],
)

def dpftri(transr, uplo, n, ar):
    u"""
    Inverse of real symmetric positive definite matrix, matrix already
    factorized by ``dpftrf``, Rectangular Full Packed format.

    ``dpftri`` computes the inverse of a real symmetric positive
    definite matrix using the Cholesky factorization computed by
    ``dpftrf`` stored in Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07wj

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07wjf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'T'

            The matrix A is stored in transposed RFP format.

    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factorization of A stored in RFP format, as
        returned by ``dpftrf``.

    Returns
    -------
    ar : float, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n*n matrix A^-1 stored
        in RFP format.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to invert a symmetric matrix stored in
            RFP format which is not positive definite; the matrix must
            be treated as a full symmetric matrix, by calling
            ``dsytri``.

    Notes
    -----
    ``dpftri`` is used to compute the inverse of a real symmetric
    positive definite matrix A, stored in RFP format. The RFP storage
    format is described in [the F07 Introduction]. The function must be
    preceded by a call to ``dpftrf``, which computes the Cholesky
    factorization of A.

    If `uplo` = 'U', A = U^TU and A^-1 is computed by first inverting U
    and then forming (U^-1)U^-T.

    If `uplo` = 'L', A = LL^T and A^-1 is computed by first inverting L
    and then forming L^-T(L^-1).

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.dpftri'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dpftri(
            transr,
            uplo,
            n,
            _ar,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ar_py = _ar
    _return_data = DpftriReturnData(
        _ar_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ar_py

DtftriReturnData = _collections.namedtuple(
    'DtftriReturnData',
    ['ar'],
)

def dtftri(transr, uplo, diag, n, ar):
    u"""
    Inverse of real triangular matrix, Rectangular Full Packed format.

    ``dtftri`` computes the inverse of a real triangular matrix stored
    in Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07wk

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07wkf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'T'

            The matrix A is stored in transposed RFP format.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ar : float, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n*n symmetric matrix A, in either normal or transposed RFP
        format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    ar : float, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by A^-1, in the same storage format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of A is exactly zero. A is
            singular its inverse cannot be computed.

    Notes
    -----
    ``dtftri`` forms the inverse of a real triangular matrix A, stored
    using RFP format. The RFP storage format is described in [the F07
    Introduction]. Note that the inverse of an upper (lower) triangular
    matrix is also upper (lower) triangular.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.dtftri'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dtftri(
            transr,
            uplo,
            diag,
            n,
            _ar,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ar_py = _ar
    _return_data = DtftriReturnData(
        _ar_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ar_py

def zpftrf(transr, uplo, n, ar):
    u"""
    Cholesky factorization of complex Hermitian positive definite
    matrix, Rectangular Full Packed format.

    ``zpftrf`` computes the Cholesky factorization of a complex
    Hermitian positive definite matrix stored in Rectangular Full Packed
    (RFP) format.

    For full information please refer to the NAG Library document for
    f07wr

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07wrf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

            The upper triangular part of A is stored, and A is
            factorized as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored, and A is
            factorized as LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : complex, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n*n Hermitian matrix A, in either normal or transposed RFP
        format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    ar : complex, ndarray, shape (`n`*(`n`+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^HU or A = LL^H, in the same storage
        format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to factorize a Hermitian matrix stored
            in RFP format which is not positive definite; the matrix
            must be treated as a full Hermitian matrix, by calling
            ``zhetrf``.

    Notes
    -----
    ``zpftrf`` forms the Cholesky factorization of a complex Hermitian
    positive definite matrix A either as A = U^HU if `uplo` = 'U' or A =
    LL^H if `uplo` = 'L', where U is an upper triangular matrix and L is
    a lower triangular, stored in RFP format. The RFP storage format is
    described in [the F07 Introduction].

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.zpftrf'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
    )
    _b_lapacklin.zpftrf(
        transr,
        uplo,
        n,
        _ar,
    )
    _ar_py = _ar
    return _ar_py

def zpftrs(transr, uplo, n, ar, b):
    u"""
    Solution of complex Hermitian positive definite system of linear
    equations, multiple right-hand sides, coefficient matrix already
    factorized by ``zpftrf``, Rectangular Full Packed format.

    ``zpftrs`` solves a complex Hermitian positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    using the Cholesky factorization computed by ``zpftrf`` stored in
    Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07ws

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07wsf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factorization of A stored in RFP format, as
        returned by ``zpftrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n*r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (`n`, nrhs)
        The n*r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Notes
    -----
    ``zpftrs`` is used to solve a complex Hermitian positive definite
    system of linear equations AX = B, the function must be preceded by
    a call to ``zpftrf`` which computes the Cholesky factorization of A,
    stored in RFP format. The RFP storage format is described in [the
    F07 Introduction]. The solution X is computed by forward and
    backward substitution.

    If `uplo` = 'U', A = U^HU, where U is upper triangular; the solution
    X is computed by solving U^HY = B and then UX = Y.

    If `uplo` = 'L', A = LL^H, where L is lower triangular; the solution
    X is computed by solving LY = B and then L^HX = Y.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.zpftrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=False,
    )
    _b_lapacklin.zpftrs(
        transr,
        uplo,
        n,
        _nrhs,
        _ar,
        _b,
    )
    _b_py = _b
    return _b_py

def zpftri(transr, uplo, n, ar):
    u"""
    Inverse of complex Hermitian positive definite matrix, matrix
    already factorized by ``zpftrf``, Rectangular Full Packed format.

    ``zpftri`` computes the inverse of a complex Hermitian positive
    definite matrix using the Cholesky factorization computed by
    ``zpftrf`` stored in Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07ww

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07wwf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factorization of A stored in RFP format, as
        returned by ``zpftrf``.

    Returns
    -------
    ar : complex, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n*n matrix A^-1 stored
        in RFP format.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to invert a Hermitian matrix stored in
            RFP format which is not positive definite; the matrix must
            be treated as a full Hermitian matrix, by calling
            ``zhetri``.

    Notes
    -----
    ``zpftri`` is used to compute the inverse of a complex Hermitian
    positive definite matrix A, stored in RFP format. The RFP storage
    format is described in [the F07 Introduction]. The function must be
    preceded by a call to ``zpftrf``, which computes the Cholesky
    factorization of A.

    If `uplo` = 'U', A = U^HU and A^-1 is computed by first inverting U
    and then forming (U^-1)U^-H.

    If `uplo` = 'L', A = LL^H and A^-1 is computed by first inverting L
    and then forming L^-H(L^-1).

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.zpftri'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
    )
    _b_lapacklin.zpftri(
        transr,
        uplo,
        n,
        _ar,
    )
    _ar_py = _ar
    return _ar_py

ZtftriReturnData = _collections.namedtuple(
    'ZtftriReturnData',
    ['ar'],
)

def ztftri(transr, uplo, diag, n, ar):
    u"""
    Inverse of complex triangular matrix, Rectangular Full Packed
    format.

    ``ztftri`` computes the inverse of a complex triangular matrix
    stored in Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07wx

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/f07/f07wxf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ar : complex, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n*n Hermitian matrix A, in either normal or transposed RFP
        format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    ar : complex, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by A^-1, in the same storage format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of A is exactly zero. A is
            singular its inverse cannot be computed.

    Notes
    -----
    ``ztftri`` forms the inverse of a complex triangular matrix A,
    stored using RFP format. The RFP storage format is described in [the
    F07 Introduction]. Note that the inverse of an upper (lower)
    triangular matrix is also upper (lower) triangular.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.ztftri'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
    )
    try:
        _b_lapacklin.ztftri(
            transr,
            uplo,
            diag,
            n,
            _ar,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ar_py = _ar
    _return_data = ZtftriReturnData(
        _ar_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ar_py
